/* A C-program for MT19937B: Real number version */
/* genrand() generate one pseudorandom number with double precision   */
/* which is uniformly distributed on [0,1]-interval for each call.    */
/* sgenrand(seed) set initial values to the working area of 624 words.*/
/* sgenrand(seed) must be called once before calling genrand()        */
/* (seed is any integer except 0).                                    */

/* 
    LICENCE CONDITIONS: 

		Matsumoto and Nishimura consent to GNU General 
		Public Licence

    NOTE: 
		When you use it in your program, please let Matsumoto
                <matumoto@math.keio.ac.jp> know it.

		Because of a machine-trouble, Matsumoto lost emails 
   		which arrived during May 28-29.
*/

/* Modified by S.C.H. (2 Jan 98) to use the getseed() and writeseed() routines */

#include <stdio.h>
#include <stdlib.h>
#include <limits.h>
#include <string.h>

#include "libhdr.h"

#define N              (624)                 
#define M              (397)                 
#define K              (0x9908B0DFU)         
#define hiBit(u)       ((u) & 0x80000000U)   
#define loBit(u)       ((u) & 0x00000001U)   
#define loBits(u)      ((u) & 0x7FFFFFFFU)   
#define mixBits(u, v)  (hiBit(u)|loBits(v))  

static unsigned long   state[N+1];     
static unsigned long   *next;          
static int      left = -1;      

void sgenrand(unsigned long seed)
{ 
	int j;
   unsigned long x,*s;
	
	x = (seed | 1U) & 0xFFFFFFFFU, s = state;
	for(left=0, *s++=x, j=N; --j;
		 *s++ = (x*=69069U) & 0xFFFFFFFFU);
}

static unsigned long reloadMT(void)
{
    unsigned long *p0=state, *p2=state+2, *pM=state+M, s0, s1;
    int j;

    left=N-1, next=state+1;
    for(s0=state[0], s1=state[1], j=N-M+1; --j; s0=s1, s1=*p2++)
        *p0++ = *pM++ ^ (mixBits(s0, s1) >> 1) ^ (loBit(s1) ? K : 0U);
    for(pM=state, j=M; --j; s0=s1, s1=*p2++)
        *p0++ = *pM++ ^ (mixBits(s0, s1) >> 1) ^ (loBit(s1) ? K : 0U);
    s1=state[0], *p0 = *pM ^ (mixBits(s0, s1) >> 1) ^ (loBit(s1) ? K : 0U);
    s1 ^= (s1 >> 11);
    s1 ^= (s1 <<  7) & 0x9D2C5680U;
    s1 ^= (s1 << 15) & 0xEFC60000U;
    return(s1 ^ (s1 >> 18));
}

double genrand(void)
{
	unsigned long y;
	double z;
 	if(--left < 0) return((double)reloadMT()/(double)0xffffffffu);
	y  = *next++;
	y ^= (y >> 11);
	y ^= (y <<  7) & 0x9D2C5680U;
	y ^= (y << 15) & 0xEFC60000U;
	z=(double) (y^(y >> 18))/(double)0xffffffffu;
	return z;
}

double safe_genrand(void)
{
	double x;
	
	do x=genrand(); while(x==0.0 || x==1.0);
	return x;
}

int getseed(const char *fname)
{
	int flag=0,i;
	unsigned long sd;
	char buf[256];
	FILE *fptr;
	
	if(!fname) {
		(void)fprintf(stderr,"getseed(): NULL filename\n");
		return -1;
	}
	fptr=fopen(fname,"r");
	if(fptr)	{
		if(fgets(buf,256,fptr))	{
			flag=sscanf(buf,"mt19937b_seed = %lu\n",&sd);
			if(flag!=1)	{
				if(sscanf(buf,"mt19937b_idx = %d\n",&left)==1) {
					if(left>=0 && left<=N) {
						for(i=0;i<N;i++) if(fscanf(fptr,"%lu",state+i)!=1) break;
						next=state+N-left;
						if(i==N)	{
							(void)printf("getseed(): Using stored state for mt19937b generator\n");
							(void)fclose(fptr);
							return 0;
						}
					}
				}
			}
		}
		(void)fclose(fptr);
		if(flag!=1) {
			(void)fprintf(stderr,"getseed(): Couldn't read seed from file %s\n",fname);
			return -1;
		}
	} else {
		(void)fprintf(stderr,"getseed(): Couldn't open seedfile %s for reading\n",fname);
		return -1;
	}
	(void)printf("getseed(): Using seed from file (%lu)\n",sd);
	sgenrand(sd);
	return 0;
}

int dumpseed(FILE *fptr,const int flag)
{
	int i,err=0;
	unsigned long sd;
	
	if(flag)	{
		err=fprintf(fptr,"mt19937b_idx = %d\n",left);
		if(err>0) for(i=0;i<N;i++)	{
			err=fprintf(fptr,"%lu%c",state[i],(i+1)%16?' ':'\n');
			if(err<0) break;
		}
	} else {
		sd=(unsigned long)(genrand()*(double)0xffffffff);
		err=fprintf(fptr,"mt19937b_seed = %lu\n",sd);
	}
	return err;
}

int bindumpseed(FILE *fptr)
{
	int i,err=0;
	
	if(fprintf(fptr,"%x,%x\n",N,left)<0) err=1;
	for(i=0;!err && i<N;i++) if(fprintf(fptr,"%lx\n",state[i])<0) err=1;
	return err;
}

int binreadseed(FILE *fptr,char *s)
{
	int i,err=0;
	
	if(sscanf(s,"%x,%x\n",&i,&left)!=2 || i!=N) err=1;
	else for(i=0;!err && i<N;i++) if(fscanf(fptr,"%lx\n",state+i)!=1) err=1;
	if(!err) next=state+N-left;
	return err;
}

int writeseed(const char *fname,const int flag)
{
	int err;
	FILE *fptr;
	
	if(!fname) {
		(void)fprintf(stderr,"writeseed(): NULL filename\n");
		return -1;
	}
	err=mkbackup(fname,1);
	fptr=fopen(fname,"w");
	if(fptr) {
		err=dumpseed(fptr,flag);
		(void)fclose(fptr);
	} else {
		(void)fprintf(stderr,"writeseed(): Couldn't open seedfile for writing\n");
		return -1;
	}
	return err;
}
