/****************************************************************************
 *                                                                          *
 *     Loki - Programs for genetic analysis of complex traits using MCMC    *
 *                                                                          *
 *             Simon Heath - University of Washington                       *
 *                                                                          *
 *                      March/April 1997                                    *
 *                                                                          *
 * calc_nrm.c:                                                              *
 *                                                                          *
 * Calculate Inverse of NRM matrix using algorithm of Quaas (1976)          *
 *                                                                          *
 ****************************************************************************/

#include <config.h>

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <sys/wait.h>

#include "utils.h"
#include "scan.h"
#include "sparse.h"

#ifdef FUNC_NAME
#undef FUNC_NAME
#endif
#define FUNC_NAME "Calculate_NRM"
/* Calculate inverse of NRM (G) matrix using Quaas/Henderson method 
 * See Numerical Recipes in C 2nd edition pps. 78-79 for a description 
 * of the sparse matrix storage. Note the matrix is half stored. */
int Calculate_NRM(void)
{
	int id,i,j,k,k1,idd,ids,idd1,ids1,nz,comp,max_comp=0;
	char *fname;
	int *sire_list,*dam_list,*temp,*spouses,*perm;
	FILE *flog,*fptr;
	double d,*u,*v,xx,d2,d4,Detl,TDetl;
	struct SparseMatRec *AIMatrix;
	struct model *model;
	struct model_list *mlist;

	fname=make_file_name("nrm");
	if(unlink(fname)<0) if(errno!=ENOENT) {
		perror("Calculate_NRM");
		fprintf(stderr,"Can't remove file %s.\n",fname);
		free(fname);
		return -1;
	}
	if(!pruned_ped_size)	{
		free(fname);
		return 0;
	}
	/* Do we need an NRM ? */
	model=Models;
	while(model) {
		mlist=model->model_list;
		while(mlist) {
			for(i=0;i<mlist->nvar;i++) if(mlist->element[i]->type&ST_ID) break;
			if(i<mlist->nvar) break;
			mlist=mlist->next;
		}
		if(mlist) break;
		model=model->next;
	}
	if(!model) {
		free(fname);
		return 0;
	}
	/* Find maximum component size */
	for(i=0;i<n_comp;i++) if(comp_size[i]>max_comp) max_comp=comp_size[i];
	if(!(sire_list=calloc((size_t)(pruned_ped_size+max_comp*2),sizeof(int)))) ABT_FUNC(MMsg);
	dam_list=sire_list+max_comp;
	perm=dam_list+max_comp;
	if(!(u=malloc(max_comp*2*sizeof(double)))) ABT_FUNC(MMsg);
	v=u+max_comp;
	for(i=0;i<ped_size;i++)	{
		j=ped_recode1[i];
		if(j) perm[j-1]=i;
	}
	TDetl=0.0;
	if(LogFile) flog=fopen(LogFile,"a");
	else flog=0;
	if(flog)	{
		fprintf(flog,"\n**************** Calculation of NRM inverse *************\n\n");
		fprintf(flog,"     Pedigree size = %d\n\n",pruned_ped_size);
	}
	errno=0;
	if(Filter) {
		i=child_open(WRITE,fname,Filter);
		if(!(fptr=fdopen(i,"w"))) DataFileError(fname);
		if(errno && errno!=ESPIPE) DataFileError(fname);
		errno=0;
	} else if(!(fptr=fopen(fname,"w"))) abt(__FILE__,__LINE__,"%s(): File Error.  Couldn't open '%s' for writing.\n",FUNC_NAME,fname);
	if(fprintf(fptr,"Loki.nrm:%x,%x,%x\n",RunID,pruned_ped_size,n_comp)<0) DataFileError(fname);
	/* Do this 1 component at a time */
	for(id=comp=0;comp<n_comp;comp++) {
		if(!(temp=calloc((size_t)comp_size[comp]+1,sizeof(int)))) ABT_FUNC(MMsg);
		/* Count non-zero (off diagonal) elements in matrix */
		for(j=k=0;j<comp_size[comp];j++)	{
			i=perm[j+id];
			ids=id_array[i].sire;
			idd=id_array[i].dam;
			if(ids) ids=ped_recode1[ids-1]-id;
			if(idd) idd=ped_recode1[idd-1]-id;
			sire_list[j]=ids;
			dam_list[j]=idd;
			if(ids!=idd) {
				k++;
				if(ids>idd) temp[ids-1]++;
				else temp[idd-1]++;
			}
		}
		nz=0;
		spouses=0;
		if(k) {
			if(!(spouses=calloc((size_t)k,sizeof(int)))) ABT_FUNC(MMsg);
			for(j=k=0;j<comp_size[comp];j++)	{
				k1=k;
				k+=temp[j];
				temp[j]=k1;
				ids=sire_list[j];
				idd=dam_list[j];
				if(ids && ids!=(j+1)) nz++;
				if(idd && idd!=ids && idd!=(j+1)) nz++;
				if(ids!=idd) {
					if(ids>idd)	{
						for(k1=temp[ids-1];k1<temp[ids];k1++) {
							if(spouses[k1]==idd) break;
							if(!spouses[k1]) {
								spouses[k1]=idd;
								nz++;
								break;
							}
						}
					} else {
						for(k1=temp[idd-1];k1<temp[idd];k1++) {
							if(spouses[k1]==ids) break;
							if(!spouses[k1]) {
								spouses[k1]=ids;
								nz++;
								break;
							}
						}
					}
				}
			}
		}
		if(!(AIMatrix=malloc((nz+1+comp_size[comp])*sizeof(struct SparseMatRec)))) ABT_FUNC(MMsg);
		/* Set up coordinates for non-zero elements */
		nz=comp_size[comp]+1;
		for(j=0;j<comp_size[comp];j++) {
			u[j]=0.0;
			AIMatrix[j].val=0.0;
			AIMatrix[j].x=nz;
			ids=sire_list[j];
			idd=dam_list[j];
			if(ids && ids!=(j+1)) {
				AIMatrix[nz].x=ids-1;
				AIMatrix[nz++].val=0.0;
			}
			if(idd && idd!=ids && idd!=(j+1)) {
				AIMatrix[nz].x=idd-1;
				AIMatrix[nz++].val=0.0;
			}
			if(!spouses) continue;
			for(k=temp[j];k<temp[j+1];k++) {
				if(!spouses[k]) break;
				AIMatrix[nz].x=spouses[k]-1;
				AIMatrix[nz++].val=0.0;
			}
		}
		AIMatrix[comp_size[comp]].x=nz;
		if(spouses) free(spouses);
		free(temp);
		Detl=0.0;
		printf("Generating inverse of NRM matrix for component %d...\n",comp+1);
		for(i=0;i<comp_size[comp];i++) {
			xx=0.0;
			ids=sire_list[i]-1;
			idd=dam_list[i]-1;
			if(ids>=0) xx=u[ids];
			if(idd>=0) xx+=u[idd];
			xx=1.0-.25*xx;
			d=1.0/xx;
			u[i]+=xx;
			xx=sqrt(xx);
			v[i]=xx;
			Detl+=log(xx);
			for(k=i+1;k<comp_size[comp];k++) {
				xx=0.0;
				ids1=sire_list[k]-1;
				idd1=dam_list[k]-1;
				if(ids1>=i) xx=v[ids1];
				if(idd1>=i) xx+=v[idd1];
				if(xx>0.0) {
					xx=.5*xx;
					u[k]+=xx*xx;
				}
				v[k]=xx;
			}
			AIMatrix[i].val+=d;
			d2= -.5*d;
			d4=.25*d;
			j=AIMatrix[i].x;
			if(ids>=0) {
				AIMatrix[j++].val+=d2;
				AIMatrix[ids].val+=d4;
			}
			if(idd>=0) {
				AIMatrix[j++].val+=d2;
				AIMatrix[idd].val+=d4;
				if(ids>=0) {
					if(ids==idd) AIMatrix[ids].val+=d4;
					else if(ids>idd) {
						for(j=AIMatrix[ids].x;j<AIMatrix[ids+1].x;j++) if(AIMatrix[j].x==idd) {
							AIMatrix[j].val+=d4;
							break;
						}
					} else {
						for(j=AIMatrix[idd].x;j<AIMatrix[idd+1].x;j++) if(AIMatrix[j].x==ids) {
							AIMatrix[j].val+=d4;
							break;
						}
					}
				}
			}
		}
		Detl+=Detl;
		TDetl+=Detl;
		AIMatrix[comp_size[comp]].val=Detl;
		printf("No. non-zero elements = %d, log determinant = %f\n",nz-1,Detl);
		if(fprintf(fptr,"LKNM:%x,%x\n",comp_size[comp],nz)<0) DataFileError(fname);
		for(j=0;j<nz;j++) {
			if(fprintf(fptr,"%x,",AIMatrix[j].x)<0) DataFileError(fname);
			if(txt_print_double(AIMatrix[j].val,fptr)) DataFileError(fname);
			if(fputc('\n',fptr)==EOF) DataFileError(fname);
		}
		free(AIMatrix);
		if(flog)	{
			fprintf(flog,"     Component %d\n",comp+1);
			fprintf(flog,"       No. non-zero elements of NRM = %d\n",nz-1);
			fprintf(flog,"       Log Determinant of NRM matrix = %f\n",Detl);
		}
		id+=comp_size[comp];
	}
	free(sire_list);
	free(u);
	if(fputs("Lnrm.end\n",fptr)==EOF) DataFileError(fname);
	if(fclose(fptr)) DataFileError(fname);
	if(Filter) do i=wait(&j); while(i>0);
	free(fname);
	if(flog)	{
		if(n_comp>1) fprintf(flog,"\n     Log Determinant of combined NRM matrix = %f\n",TDetl);
		fclose(flog);
	}
	return 1;
}
