/****************************************************************************
 *                                                                          *
 *     Loki - Programs for genetic analysis of complex traits using MCMC    *
 *                                                                          *
 *             Simon Heath - University of Washington                       *
 *                                                                          *
 *                        July 1997                                         *
 *                                                                          *
 * gen_elim.c:                                                              *
 *                                                                          *
 * (1) Use pattern of marker data to prune the pedigree and split it into   *
 * components on a marker by marker basis.                                  *
 *                                                                          *
 * (2) Recode unused alleles in each component                              *
 *                                                                          *
 * (3) Perform genotype elimination.                                        *
 *                                                                          *
 * (4) Perform set recoding a la O'Connell & Weeks                          *
 *                                                                          *
 * (5) Determine a peeling sequence                                         *
 *                                                                          *
 * (6) Perform 'logical peeling' to check genotype consistency and          *
 * allow optimizations for later peeling runs                               *
 *                                                                          *
 ****************************************************************************/

#include <config.h>

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#if HAVE_LIMITS_H
#include <limits.h>
#endif
#include <sys/wait.h>
#include <string.h>
#include <errno.h>

#include "utils.h"
#include "scan.h"
#include "control_parse.h"
#include "prep_peel.h"
#include "amdbar.h"
#include "ranlib.h"

extern int catch_sigs,sig_caught;

static int *allele_trans,bad_cnt,rec_gen_flag;

static int *perm,id,*famflag,r_func_size,n_rfuncs,*involved,*prev_inv,*true_involved,*rf_flag;
static int no_peel_flag,n_prev_inv,silent_flag;
static unsigned long *temp_set,*id_set[2],mask,**all_set,*req_set[3];
static struct R_Func *r_func;
static char *marker_name;
static struct Peelseq_Head peelseq_head;

int trace_peel;

void print_orig_allele_id(FILE *fptr,const int i)
{
	if(i>0) {
		print_orig_id(fptr,i,0);
		fputc('m',fptr);
	} else {
		print_orig_id(fptr,-i,0);
		fputc('p',fptr);
	}
}

#ifdef FUNC_NAME
#undef FUNC_NAME
#endif
#define FUNC_NAME "FlagFam"
static void FlagFam(const int fam)
{
	int j,k,l,ids,idd;
	
	ids=family[fam].sire;
	idd=family[fam].dam;
	if(!ids) ABT_FUNC("Internal error - fix me\n");
	j=id_array[ids-1].family;
	if(j) famflag[j-1]|=1;
	for(k=0;k<id_array[ids-1].nfam;k++)	{
		j=id_array[ids-1].famlist[k];
		famflag[j]|=1;
	}
	j=id_array[idd-1].family;
	if(j) famflag[j-1]|=1;
	for(k=0;k<id_array[idd-1].nfam;k++)	{
		j=id_array[idd-1].famlist[k];
		famflag[j]|=1;
	}
	for(k=0;k<family[fam].nkids;k++)	{
		j=family[fam].kids[k];
		for(l=0;l<id_array[j].nfam;l++) {
			famflag[id_array[j].famlist[l]]|=1;
		}
	}
}

void print_orig_alleles(FILE *fptr,const int ind, const int locus,const int linktype)
{
	int i,ch,fg=0;
	
	if(linktype==LINK_Y || (linktype==LINK_X && id_array[ind-1].sex==1)) fg=1;
	if(id_array[ind-1].haplo[0]) {
		for(i=0;i<2;i++) {
			ch=id_array[ind-1].haplo[i][locus];
			if(ch) {
				ch=allele_trans[ch-1];
				if(factor_recode[n_factors+locus][ch]->type==STRING)
				  fputs(factor_recode[n_factors+locus][ch]->data.string,fptr);
				else fprintf(fptr,"%ld",factor_recode[n_factors+locus][ch]->data.value);
			} else fputc('*',fptr);
			if(fg) break;
			if(!i) fputc(',',fptr);
		}
	} else {
		if(fg) fputc('*',fptr);
		else fputs("*,*",fptr);
	}
}

static void print_rec_all(FILE *fptr, int ch,const int locus)
{
	ch=allele_trans[ch];
	if(ch== -1) fputs("__LUMP__",fptr);
	else {
		if(factor_recode[n_factors+locus][ch]->type==STRING)
		  fputs(factor_recode[n_factors+locus][ch]->data.string,fptr);
		else fprintf(fptr,"%ld",factor_recode[n_factors+locus][ch]->data.value);
	}
}

static void print_code_rec_all(FILE *fptr,int ch,lk_ulong a,int locus)
{
	int l,k;
	
	if(a&(1<<ch)) {
		l=k=0;
		while(a) {
			if(a&1) {
				fputc(k++?',':'[',fptr);
				print_rec_all(fptr,l,locus);
			}
			l++;
			a>>=1;
		};
		fputc(']',fptr);
	} else print_rec_all(fptr,ch,locus);
}

#ifdef FUNC_NAME
#undef FUNC_NAME
#endif
#define FUNC_NAME "print_alls"
static void print_alls(FILE *fptr,const int i,const int locus,const int n_all,const int linktype,const int flag)
{
	int j,k,k1,l,fg=0;
	lk_ulong a,*b,m;
	
	if(!(b=calloc(n_all,sizeof(lk_ulong)))) ABT_FUNC(MMsg);
	j=ped_recode1[i-1]-1-id;
	if(linktype==LINK_Y || (linktype==LINK_X && id_array[i-1].sex==1)) fg=1;
	m=(1<<n_all)-1;
	for(k1=k=0;k<n_all;k++)	{
		a=all_set[k][j];
		if(a==m) k1++;
		for(l=0;l<n_all;l++) if(a&(1<<l)) b[l]|=1<<k;
	}
	if(k1==n_all) {
		if(fg) fputs("(*)",fptr);
		else fputs("(*,*)",fptr);
	} else if(linktype==LINK_Y) {
		a=all_set[0][j];
		l=0;
		if(a==m) fputs("(*)",fptr);
		else while(a) {
			if(a&1) {
				fputc('(',fptr);
				print_rec_all(fptr,l,locus);
				fputs(") ",fptr);
			}
			l++;
			a>>=1;
		}
	} else if(linktype==LINK_X && id_array[i-1].sex==1) {
		for(k=0;k<n_all;k++) if(!all_set[k][j]) break;
		if(k==n_all) fputs("(*)",fptr);
		else for(k=0;k<n_all;k++) if(all_set[k][j]) {
			fputc('(',fptr);
			print_rec_all(fptr,k,locus);
			fputs(") ",fptr);
		}
	} else {
		for(k=0;k<n_all;k++) {
			l=0;
			a=all_set[k][j];
			if(a==m) {
				fputs("(*,",fptr);
				if(flag) print_code_rec_all(fptr,k,req_set[X_MAT][j],locus);
				else print_rec_all(fptr,k,locus);
				fputs(") ",fptr);
			} else while(a) {
				if(b[l]!=m && (a&1)) {
					fputc('(',fptr);
					if(flag) print_code_rec_all(fptr,l,req_set[X_PAT][j],locus);
					else print_rec_all(fptr,l,locus);
					fputc(',',fptr);
					if(flag) print_code_rec_all(fptr,k,req_set[X_MAT][j],locus);
					else print_rec_all(fptr,k,locus);
					fputs(") ",fptr);
				}
				l++;
				a>>=1;
			}
		}
		for(l=0;l<n_all;l++) if(b[l]==m) {
			fputc('(',fptr);
			if(flag) print_code_rec_all(fptr,l,req_set[X_PAT][j],locus);
			else print_rec_all(fptr,l,locus);
			fputs(",*) ",fptr);
		}
	}
	fputc('\n',fptr);
	free(b);
}

static void PrintFamily(const int fam,const int locus,const int n_all,const int linktype,const int flag)
{
	int ids,idd,kid,i;
	
	ids=family[fam].sire;
	idd=family[fam].dam;
	if(family_id) {
		fputs("Family ",stderr);
		print_orig_family(stderr,family[fam].kids[0]+1,0);
		fputc('\n',stderr);
	}
	if(ids) {
		fputs("Father: ",stderr);
		print_orig_id1(stderr,ids,0);
		fputs(" [",stderr);
		print_orig_alleles(stderr,ids,locus,linktype);
		fputs("]\n        ",stderr);
		print_alls(stderr,ids,locus,n_all,linktype,flag);
	}
	if(idd && linktype!=LINK_Y) {
		fputs("Mother: ",stderr);
		print_orig_id1(stderr,idd,0);
		fputs(" [",stderr);
		print_orig_alleles(stderr,idd,locus,linktype);
		fputs("]\n        ",stderr);
		print_alls(stderr,idd,locus,n_all,linktype,flag);
	}
	for(i=0;i<family[fam].nkids;i++) {
		kid=family[fam].kids[i];
		if(id_array[kid].flag&2) continue;
		if(linktype==LINK_Y && id_array[kid].sex==2) continue;
		fputs("     -> ",stderr);
		print_orig_id1(stderr,++kid,0);
		fputs(" [",stderr);
		print_orig_alleles(stderr,kid,locus,linktype);
		fputs("]\n        ",stderr);
		print_alls(stderr,kid,locus,n_all,linktype,flag);
	}
}

#ifdef FUNC_NAME
#undef FUNC_NAME
#endif
#define FUNC_NAME "DoFamily"
static int DoFamily(const int fam,const int n_all,const int locus,const int linktype)
{
	int i,j,k,l,m,ids,idd,nc=0,fg,kid,change=0;
	unsigned long a,b,cm,cm1;
	
	ids=family[fam].sire;
	if(!ids)	{
		famflag[fam]&=~1;
		return 0;
	}
	ids=ped_recode1[ids-1]-1-id;
	idd=ped_recode1[family[fam].dam-1]-1-id;
	for(i=0;i<n_all;i++) temp_set[i]=id_set[0][i]=id_set[1][i]=0;
	if(linktype==LINK_AUTO)	{
		for(i=0;i<n_all;i++)	{
			a=all_set[i][ids];
			if(a) for(j=0;j<n_all;j++) if(a&(1<<j)) {
				if(j>i && (all_set[j][ids]&(1<<i))) continue;
				for(k=0;k<n_all;k++) {
					b=all_set[k][idd];
					if(b) for(l=0;l<n_all;l++) if(b&(1<<l)) {
						if(l>k && (all_set[l][idd]&(1<<k))) continue;
						cm=((1<<i)|(1<<j));
						fg=0;
						for(m=0;m<family[fam].nkids;m++)	{
							kid=family[fam].kids[m];
							if(id_array[kid].flag&2) continue;
							kid=ped_recode1[kid]-1-id;
							if(!((all_set[k][kid]&cm)||(all_set[l][kid]&cm))) {
								fg=1;
								break;
							}
						}
						if(fg) continue;
						temp_set[k]|=cm;
						temp_set[l]|=cm;
						id_set[0][i]|=1<<j;
						id_set[0][j]|=1<<i;
						id_set[1][k]|=1<<l;
						id_set[1][l]|=1<<k;
						nc++;
					}
				}
			}
		}
	} else if(linktype==LINK_X) {
		for(i=0;i<n_all;i++) if(all_set[i][ids]) {
			for(k=0;k<n_all;k++)	{
				b=all_set[k][idd];
				if(b) for(l=0;l<n_all;l++) if(b&(1<<l)) {
					if(l>k && (all_set[l][idd]&(1<<k))) continue;
					cm=(1<<i);
					fg=0;
					for(m=0;m<family[fam].nkids;m++)	{
						kid=family[fam].kids[m];
						if(id_array[kid].flag&2) continue;
						cm1=(id_array[kid].sex==1?1:cm);
						kid=ped_recode1[kid]-1-id;
						if(!((all_set[k][kid]&cm1)||(all_set[l][kid]&cm1))) {
							fg=1;
							break;
						}
					}
					if(fg) continue;
					temp_set[k]|=cm;
					temp_set[l]|=cm;
					id_set[0][i]|=1;
					id_set[1][k]|=1<<l;
					id_set[1][l]|=1<<k;
					nc++;
				}
			}
		}
	} else if(linktype==LINK_Y) {
		a=all_set[0][ids];
		for(i=0;i<n_all;i++) if(a&(1<<i)) {
			fg=0;
			cm=1<<i;
			for(m=0;m<family[fam].nkids;m++)	{
				kid=family[fam].kids[m];
				if((id_array[kid].flag&2) || id_array[kid].sex==2) continue;
				kid=ped_recode1[kid]-1-id;
				if(!(all_set[0][kid]&cm)) {
					fg=1;
					break;
				}
			}
			if(!fg) {
				temp_set[0]=cm;
				nc=1;
			}
		}
	}
	if(linktype!=LINK_Y)	{
		for(k=l=j=0;j<n_all;j++) {
			if(all_set[j][ids]&id_set[0][j]) k++;
			if(all_set[j][idd]&id_set[1][j]) l++;
			if(k&&l) break;
		}
	} else j=0;
	if(!nc || j==n_all) {
		if(!silent_flag) {
			fprintf(stderr,"\nDoFamily(): Inconsistent family data for locus %s\n",marker_name);
			PrintFamily(fam,locus,n_all,linktype,0);
		}
		return 1;
	}
	if(linktype!=LINK_Y) for(j=0;j<n_all;j++) {
		a=all_set[j][ids];
		all_set[j][ids]&=id_set[0][j];
		if(a!=all_set[j][ids]) change=1;
		a=all_set[j][idd];
		all_set[j][idd]&=id_set[1][j];
		if(a!=all_set[j][idd]) change=1;
	}
	for(m=0;m<family[fam].nkids;m++)	{
		kid=family[fam].kids[m];
		if(id_array[kid].flag&2 || (linktype==LINK_Y && id_array[kid].sex==2)) continue;
		l=(linktype==LINK_Y?1:n_all);
		k=ped_recode1[kid]-1-id;
		l=(linktype==LINK_Y?1:n_all);
		for(j=0;j<l;j++) {
			if(linktype==LINK_X && id_array[kid].sex==1) cm=temp_set[j]?1:0; 
			else cm=temp_set[j];
			if(all_set[j][k]&cm) break;
		}
		if(j==l)	{
			fprintf(stderr,"\nDoFamily(): (Internal error?) Inconsistent family data for locus %s\n",marker_name);
			PrintFamily(fam,locus,n_all,linktype,0);
			ABT_FUNC("Aborting\n");
		}
		for(j=0;j<l;j++) {
			a=all_set[j][k];
			if(linktype==LINK_X && id_array[kid].sex==1) cm=temp_set[j]?1:0; 
			else cm=temp_set[j];
			all_set[j][k]&=cm;
			if(a!=all_set[j][k]) change=1;
		}
	}
	if(change) FlagFam(fam);
	famflag[fam]&=~1;
	return 0;
}

/* Peel on to both parents */
#ifdef FUNC_NAME
#undef FUNC_NAME
#endif
#define FUNC_NAME "partial_peel"
static void partial_peel(const struct Simple_Element *element,int n_all,int n_bits)
{
	int i,j,k,l,m,n,idx,ids,idd,nc=0,kid,gt[4],nc1;
	unsigned long a,b,cm[2],cm1[2],cma,cmb,c,c1;
	
	ids=element->sire-1;
	idd=element->dam-1;
	nc1=id_array[ids].ngens*id_array[idd].ngens;
	idx=element->out_index;
	if(!(r_func[idx].index=malloc(nc1*sizeof(lk_ulong)))) ABT_FUNC(MMsg);
	ids=ped_recode1[ids]-1-id;
	idd=ped_recode1[idd]-1-id;
	for(i=0;i<n_all;i++)	{
		a=all_set[i][ids];
		if(!a) continue;
		c=1<<i;
		c1=req_set[X_MAT][ids];
		if(c1&c) c=c1;
		cma=c;
		gt[2]=i;
		for(j=0;j<n_all;j++) if(a&((c=(1<<j)))) {
			c1=req_set[X_PAT][ids];
			if(c1&c) c=c1;
			cm[X_PAT]=cma|c;
			gt[3]=j;
			for(k=0;k<n_all;k++)	{
				b=all_set[k][idd];
				if(!b) continue;
				c=1<<k;
				c1=req_set[X_MAT][idd];
				if(c1&c) c=c1;
				cmb=c;
				gt[0]=k;
				for(l=0;l<n_all;l++) if(b&((c=(1<<l)))) {
					c1=req_set[X_PAT][idd];
					if(c1&c) c=c1;
					cm[X_MAT]=cmb|c;
					for(m=0;m<element->n_off;m++)	{
						kid=element->off[m];
						kid=ped_recode1[kid-1]-1-id;
						cm1[X_PAT]=cm[X_PAT];
						c1=req_set[X_PAT][kid];
						if(cm[X_PAT]&c1) cm1[X_PAT]|=c1;
						cm1[X_MAT]=cm[X_MAT];
						c1=req_set[X_MAT][kid];
						if(cm[X_MAT]&c1) cm1[X_MAT]|=c1;
						for(n=0;n<n_all;n++) if(cm1[X_MAT]&(1<<n))
						  if(all_set[n][kid]&cm1[X_PAT]) break;
						if(n==n_all) break;
					}
					if(m==element->n_off) {
						gt[1]=l;
						c=0;
						for(m=3;m>=0;m--)	{
							c<<=n_bits;
							c|=gt[m];
						}
						r_func[idx].index[nc++]=(lk_ulong)c;
					}
				}
			}
		}
	}
	if(!nc) ABT_FUNC("Internal error - inconsistent genotypes\n");
	if(nc!=nc1) if(!(r_func[idx].index=realloc(r_func[idx].index,nc*sizeof(lk_ulong)))) ABT_FUNC(MMsg);
	r_func[idx].n_terms=nc;
	r_func[idx].fill=(double)nc/(double)(n_all*n_all*n_all*n_all);
}

static void calc_order(const int comp,const int n_fam,const int *famlist,const int linktype)
{
	int i,j,fflag,fam,kid,ids,idd;
	
	for(j=0;j<comp_size[comp];j++) {
		i=perm[id+j];
		id_array[i].order=0;
	}
	for(i=0;i<n_fam;i++) {
		fam=famlist[i];
		ids=family[fam].sire;
		idd=family[fam].dam;
		if(ids) {	
			fflag=0;
			if(linktype!=LINK_Y) id_array[idd-1].order++;
			for(j=0;j<family[fam].nkids;j++)	{
				kid=family[fam].kids[j];
				if(id_array[kid].flag&2) continue;
				if(linktype==LINK_Y) {
					if(id_array[kid].sex==1) {
						id_array[kid].order++;
						fflag=1;
					}
				} else if(id_array[kid].sex==2 || linktype==LINK_AUTO) {
					id_array[kid].order++;
					fflag=1;
				}
			}
			id_array[ids-1].order+=fflag;
		} else {
			kid=family[fam].kids[0];
			if(id_array[kid].flag&2) continue;
			if(linktype!=LINK_Y || id_array[kid].sex==1) id_array[kid].order++;
		}
	}
	for(j=0;j<comp_size[comp];j++) {
		i=perm[id+j];
		if(id_array[i].flag&4) id_array[i].order=1;
	} 
}

static void add_to_involved(const int i,int *n,int flag)
{
	int j;
	
	for(j=0;j<(*n);j++) if(i==involved[j]) break;
	if(j==(*n))	{
		rf_flag[j]=0;
		involved[(*n)++]=i;
	}
	if(flag) rf_flag[j]=1;
}

static int assemble_matrix(int size,int *g_perm,int *iw,int *pe,int *len,int *trans,int k2,const int linktype)
{
	int i,x,i1,j,k,l,kid,fam,n_inv,iwlen,ng,sex;

	iwlen=0;
	for(x=0;x<size;x++) {
		i1=g_perm[x];
		i=abs(i1)-1;
		n_inv=0;
		sex=id_array[i].sex;
		if(sex<0 || sex>2) ABT_FUNC("Internal error - illegal sex\n");
		if(linktype!=LINK_AUTO && !sex) ABT_FUNC("Internal error - unknown sex with sex-linked locus\n");
		add_to_involved(i1,&n_inv,0);
		/* How many genes to add for this individual (may depend on sex) */
		if(linktype==LINK_Y) {
			if(sex==1) ng=1;
			else ng=0;
		} else if(sex==2 || linktype==LINK_AUTO) ng=2;
		else ng=1;
		if((id_array[i].flag&HAP_JNT)||(ng==2 &&(id_array[i].flag&HAP_DAT))) add_to_involved(-i1,&n_inv,0);
		if(i1<0) { /* Put in parents */
			if(id_array[i].flag&HAP_P) {
				if(linktype==LINK_AUTO) {
					/* Add in sire's alleles */
					add_to_involved(id_array[i].sire,&n_inv,0);
					add_to_involved(-id_array[i].sire,&n_inv,0);
				} else if(linktype==LINK_X) {
					if(sex==1) ABT_FUNC("Internal error - father-male offspring link for X-linked locus\n");
					/* Add in sire's maternal allele */
					add_to_involved(id_array[i].sire,&n_inv,0);
				} else { /* Y-Linked */
					if(sex==2) ABT_FUNC("Internal error - father-female offspring link for Y-linked locus\n");
					/* Add in sire's paternal allele */
					add_to_involved(-id_array[i].sire,&n_inv,0);
				}
			}
		} else {
			if(id_array[i].flag&HAP_M)	{
				if(linktype==LINK_Y) ABT_FUNC("Internal error - mother-offspring link for Y-linked locus\n");
				else {
					/* Add in dam's alleles */
					add_to_involved(id_array[i].dam,&n_inv,0);
					add_to_involved(-id_array[i].dam,&n_inv,0);
				}
			}
		}
		for(j=0;j<id_array[i].nfam;j++) {
			fam=id_array[i].famlist[j];
			if(!sex) ABT_FUNC("Internal error - unknown sex for parent\n");
			for(k=0;k<family[fam].nkids;k++)	{
				kid=family[fam].kids[k];
				if(sex==1) {
					if(id_array[kid].flag&HAP_P) {
						if(linktype==LINK_X && id_array[kid].sex==1) ABT_FUNC("Internal error - father-male offspring link for X-linked locus\n");
						if(linktype==LINK_Y && id_array[kid].sex==2)  ABT_FUNC("Internal error - father-female offspring link for Y-linked locus\n");
						add_to_involved(-(kid+1),&n_inv,0);
					}
				} else {
					if(id_array[kid].flag&HAP_M) {
						if(linktype==LINK_Y) ABT_FUNC("Internal error - Mother-offspring link for Y-linked locus\n");
						add_to_involved(kid+1,&n_inv,0);
					}
				}
			}
		}
		for(j=0;j<n_rfuncs;j++) if(!r_func[j].flag) {
			for(k=0;k<r_func[j].n_ind;k++) if(r_func[j].id_list[k]==i1) break;
			if(k<r_func[j].n_ind) {
				for(k=0;k<r_func[j].n_ind;k++) {
					l=r_func[j].id_list[k];
					add_to_involved(l,&n_inv,0);
				}
			}
		}
		if(iw) {
			pe[x]=iwlen+1;
			for(k=1;k<n_inv;k++)	{
				iw[iwlen++]=trans[k2+involved[k]];
				if(iw[iwlen-1]<1 || iw[iwlen-1]>size) ABT_FUNC("Internal error - illegal allele entry\n");
			}
			len[x]=n_inv-1;
		} else iwlen+=n_inv-1;
	}
	return iwlen;
}
	
#ifdef FUNC_NAME
#undef FUNC_NAME
#endif
#define FUNC_NAME "min_degree"
static int *min_degree(int *g_perm,int size,const int linktype)
{
	int x,i1,k1,k2,*trans;
	int *pe,*iw,*degree,*nv,*next,*last,*head,*elen,*w,*len;
	int iwlen,pfree,ncmpa,iovflo;
	
	k2=INT_MAX;
	k1=INT_MIN;
	for(x=0;x<size;x++) {
		i1=g_perm[x];
		if(i1<k2) k2=i1;
		if(i1>k1) k1=i1;
	}
	k2= -k2;
	k1+=k2+1;
	if(!(trans=malloc(sizeof(int)*k1))) ABT_FUNC(MMsg);
	for(x=0;x<k1;x++) trans[x]= -1;
	for(x=0;x<size;x++) {
		i1=g_perm[x]+k2;
		trans[i1]=x+1;
	}
	iwlen=assemble_matrix(size,g_perm,0,0,0,0,0,linktype);
	if(iwlen+size>(int)(iwlen*1.2)) iwlen+=size;
	else iwlen=(int)(iwlen*1.2);
	if(!(last=malloc(sizeof(int)*size))) ABT_FUNC(MMsg);
	if(!(pe=malloc(sizeof(int)*(8*size+iwlen)))) ABT_FUNC(MMsg);
	degree=pe+size;
	nv=degree+size;
	next=nv+size;
	head=next+size;
	elen=head+size;
	w=elen+size;
	len=w+size;
	iw=len+size;
	pfree=1+assemble_matrix(size,g_perm,iw,pe,len,trans,k2,linktype);
	iovflo=INT_MAX;
	amdbar_(&size,pe,iw,len,&iwlen,&pfree,nv,next,last,head,elen,degree,&ncmpa,w,&iovflo);
	free(pe);
	free(trans);
	return last;
}

#ifdef FUNC_NAME
#undef FUNC_NAME
#endif
#define FUNC_NAME "count_ops"
static int count_ops(const int i,const int mflag,const int flag,const int linktype)
{
	int i1,j,k,l,m,kid,fam,n_inv=1,sex;
		
	i1=mflag?-(i+1):i+1;
	involved[0]=i1;
	rf_flag[0]=0;
	sex=id_array[i].sex;
	if(sex<0 || sex>2) ABT_FUNC("Internal error - illegal sex\n");
	if(linktype!=LINK_AUTO && !sex) ABT_FUNC("Internal error - unknown sex with sex-linked locus\n");
	if(linktype==LINK_Y) {
		if(sex==1) j=1;
		else j=0;
	} else if(sex==2 || linktype==LINK_AUTO) j=2;
	else j=1;
	if((id_array[i].flag&HAP_JNT)||(j==2&&(id_array[i].flag&HAP_DAT))) {
		rf_flag[n_inv]=0;
		involved[n_inv++]= -i1;
	}
	if(mflag) { /* Put in parents */
		if(id_array[i].flag&HAP_P) {
			if(linktype==LINK_AUTO) {
				add_to_involved(id_array[i].sire,&n_inv,0);
				add_to_involved(-id_array[i].sire,&n_inv,0);
			} else if(linktype==LINK_X) add_to_involved(id_array[i].sire,&n_inv,0);
			else add_to_involved(-id_array[i].sire,&n_inv,0);
			if(flag) id_array[i].flag|=HAD_P;
		}
	} else {
		if(id_array[i].flag&HAP_M)	{
			add_to_involved(id_array[i].dam,&n_inv,0);
			add_to_involved(-id_array[i].dam,&n_inv,0);
			if(flag) id_array[i].flag|=HAD_M;
		}
	}
	for(j=0;j<id_array[i].nfam;j++) { /* And offspring */
		fam=id_array[i].famlist[j];
		if(!sex) ABT_FUNC("Internal Error - unknown sex for parent\n");
		for(k=0;k<family[fam].nkids;k++)	{
			kid=family[fam].kids[k];
			if(sex==1)	{
				if(id_array[kid].flag&HAP_P) {
					add_to_involved(-(kid+1),&n_inv,0);
					if(flag) id_array[kid].flag^=(HAP_P|HAD_P);
				}
			} else {
				if(id_array[kid].flag&HAP_M) {
					add_to_involved(kid+1,&n_inv,0);
					if(flag) id_array[kid].flag^=(HAP_M|HAD_M);
				}
			}
		}
	}
	for(j=0;j<n_rfuncs;j++) if(!r_func[j].flag) {
		for(k=0;k<r_func[j].n_ind;k++) if(r_func[j].id_list[k]==i1) break;
		if(k<r_func[j].n_ind) {
			for(k=0;k<r_func[j].n_ind;k++) {
				l=r_func[j].id_list[k];
				add_to_involved(l,&n_inv,1);
			}
			if(flag) r_func[j].flag=1;
		}
	}
	for(j=0;j<n_rfuncs;j++) if(!r_func[j].flag) {
		for(k=0;k<r_func[j].n_ind;k++) {
			m=r_func[j].id_list[k];
			if(m==i1) break;
			for(l=0;l<n_inv;l++) if(m==involved[l]) break;
			if(l==n_inv) break;
		}
		if(flag && k==r_func[j].n_ind) r_func[j].flag=1;
	}
	if(flag) {
		if(mflag) id_array[i].flag&=~(HAP_P|HAP_PAT|HAP_JNT|HAP_DAT);
		else id_array[i].flag&=~(HAP_M|HAP_MAT|HAP_JNT|HAP_DAT);
	} else if(n_prev_inv) {
		for(j=0;j<n_inv;j++)	{
			l=involved[j];
			for(k=0;k<n_prev_inv;k++) if(l==prev_inv[k]) break;
			if(k==n_prev_inv) break;
		}
		if(j==n_inv) return -1;
	}
	return n_inv-1;
}

static int check_haps(const int i,const int locus,const int linktype)
{
	int fgs=0;
	
	if(id_array[i].flag&PEELED) return fgs;
	if(linktype==LINK_Y) {
		if(id_array[i].sex==1) fgs|=HAP_PAT;
	} else if(id_array[i].sex==2 || linktype==LINK_AUTO) fgs|=HAP_MAT|HAP_PAT;
	else fgs|=HAP_MAT;
	if(id_array[i].flag&WAS_PIVOT) fgs|=HAP_DAT;
	if(locus==n_markers || (markers[locus].element->type&ST_MODEL))
	  if(id_array[i].flag&1) fgs|=HAP_DAT;
	if((fgs&(HAP_DAT|HAP_MAT|HAP_PAT))==(HAP_DAT|HAP_MAT|HAP_PAT)) fgs|=HAP_JNT;
	return fgs;
}

static void set_flags(const int n_fam,const int *famlist,const int comp,const int locus,const int linktype)
{
	int i,j,fm,fp,kid,fam,ids,idd;
	
	for(j=0;j<comp_size[comp];j++) {
		i=perm[id+j];
		id_array[i].flag&=STABLE_FLAGS;
	}
	for(i=0;i<n_fam;i++) {
		fam=famlist[i];
		ids=family[fam].sire;
		if(ids) {
			idd=family[fam].dam;
			fp=check_haps(ids-1,locus,linktype);
			fm=check_haps(idd-1,locus,linktype);
		} else idd=fm=fp=0;
		for(j=0;j<family[fam].nkids;j++)	{
			kid=family[fam].kids[j];
			if(id_array[kid].flag&(2|PEELED)) continue;
			id_array[kid].flag|=check_haps(kid,locus,linktype);
			if(linktype==LINK_Y) { 
				if(id_array[kid].sex==1 && ids) {
					id_array[kid].flag|=HAP_P;
					fp|=HAP_PAT;
				}
			} else if(id_array[kid].sex==2 || linktype==LINK_AUTO) {
				if(ids) {
					id_array[kid].flag|=HAP_P;
					if(linktype==LINK_X) fp|=HAP_MAT;
					else fp|=HAP_MAT|HAP_PAT|HAP_JNT;
				}
				if(idd) {
					id_array[kid].flag|=HAP_M;
					fm|=HAP_MAT|HAP_PAT|HAP_JNT;
				}
			} else id_array[kid].flag|=HAP_M;
		}
		if(ids) id_array[ids-1].flag|=fp;
		if(idd) id_array[idd-1].flag|=fm;
	}
}

#ifdef FUNC_NAME
#undef FUNC_NAME
#endif
#define FUNC_NAME "logical_jntpeel"
static int logical_jntpeel(int n_genes,int *order,int *g_perm,const int n_all,struct Peelseq_Head *pp,const int linktype)
{
	int x,i,j,k,k1= -1,k2,k3,k4,l,pivot,n_ops,n_rf,err=0;
	int n_inv=0,n_peel;
	struct Complex_Element *element;

	if(trace_peel) printf("In logical jnt_peel(%d,%p,%p,%d,%p,%d)\n",n_genes,order,g_perm,n_all,pp,linktype);
	for(x=0;x<=n_genes;x++) {
		if(x==n_genes) pivot=n_ops=0;
		else {
			pivot=order[x];
			i=g_perm[pivot-1];
			if(i>0) n_ops=count_ops(i-1,0,0,linktype);
			else n_ops=count_ops(-1-i,1,0,linktype);
		}
		if(n_inv && n_ops>=0) {
			n_peel=n_inv;
			for(k=0;k<n_prev_inv;k++) {
				j=prev_inv[k];
				true_involved[n_inv++]=j;
			}
			if(!(element=malloc(sizeof(struct Complex_Element)))) ABT_FUNC(MMsg);
			pp->type=PEEL_COMPLEX;
			pp->ptr.complex=element;
			pp= &element->next;
			pp->type=0;
			element->n_peel=n_peel;
			element->n_involved=n_inv;
			for(n_rf=k=0;k<n_rfuncs;k++) if(r_func[k].flag==1) n_rf++;
			element->n_rfuncs=n_rf;
			if(!(element->involved=malloc(sizeof(int)*(n_inv*2+n_rf)))) ABT_FUNC(MMsg);
			element->flags=element->involved+n_inv;
			element->index=element->flags+n_inv;
			if(n_peel==n_inv)	{
				free(r_func[k1].id_list);
				n_rfuncs--;
				k1= -1;
			}
 			element->out_index=k1;
			for(k=0;k<n_inv;k++) element->flags[k]=0;
			for(k=0;k<n_inv;k++) {
				l=true_involved[k];
				if(l>0) {
					element->flags[k]|=id_array[l-1].flag&HAD_M;
					id_array[l-1].flag&=~HAD_M;
				} else {
					element->flags[k]|=id_array[-1-l].flag&HAD_P;
					id_array[-l-1].flag&=~HAD_P;
				}
			}
			for(j=0;j<n_inv;j++) element->involved[j]=true_involved[j];
			for(n_rf=k=0;k<n_rfuncs;k++) if(r_func[k].flag==1) {
				for(k1=0;k1<r_func[k].n_ind;k1++) {
					for(j=0;j<n_inv;j++) if(r_func[k].id_list[k1]==true_involved[j]) break;
					if(j==n_inv) ABT_FUNC("Internal error - allele not found.\n");
					r_func[k].id_list[k1]=j;
					element->flags[j]|=IN_RF;
				}
				element->index[n_rf++]=k;
				r_func[k].flag=2;
			}
			for(k=0;k<n_inv-1;k++) {
				l=true_involved[k];
				for(j=k+1;j<n_inv;j++) if(true_involved[j]== -l) {
					for(k2=0;k2<n_rfuncs;k2++) if(r_func[k2].flag==1) {/* Check if both alleles are in *same* R-Func */
						for(k3=k4=0;k3<r_func[k2].n_ind;k3++) {
							if(l==r_func[k2].id_list[k3]) k4++;
							if(-l==r_func[k2].id_list[k3]) k4++;
							if(k4==2) break;
						}
						if(k4==2) break;
					}
					if(k2==n_rfuncs) {
						element->flags[j]|=HAP_JNT;
						element->flags[k]|=HAP_JNT;
					}
					break;
				}
			}
			if(!no_peel_flag) {
				if(do_peel_op(element,r_func,n_all,id,all_set,req_set)) {
					if(silent_flag) {
						k1=abs(element->involved[0])-1;
						err=id_array[k1].family+1;
					} else ABT_FUNC("Zero probability during peeling operation.\n");
				}
			} else {
				k1=element->out_index;
				if(k1>=0) for(k=0;k<r_func[k1].n_ind;k++)
					  r_func[k1].id_list[k]=true_involved[n_peel+k];
			}
			for(k=0;k<element->n_rfuncs;k++) {
				k1=element->index[k];
				if(r_func[k1].index) free(r_func[k1].index);
			}
			n_inv=0;
			n_prev_inv=0;
		}
		if(!pivot || err) break;
		j=pivot-1;
		pivot=g_perm[j];
		if(pivot>0) n_ops=count_ops(pivot-1,0,1,linktype);
		else n_ops=count_ops(-1-pivot,1,1,linktype);
		if(!n_inv) {
			if(n_rfuncs>=r_func_size) {
				r_func_size<<=1;
				if(!(r_func=realloc(r_func,sizeof(struct R_Func)*r_func_size))) ABT_FUNC(MMsg);
			}
			k1=n_rfuncs++;
			if(!(r_func[k1].id_list=malloc(sizeof(int)*(1+n_ops)))) ABT_FUNC(MMsg);
			r_func[k1].n_ind=n_ops;
			r_func[k1].flag=0;
			r_func[k1].n_terms=0;
			r_func[k1].fill=100.0;
			r_func[k1].index=0;
			for(k=1;k<=n_ops;k++) r_func[k1].id_list[k-1]=involved[k];
		} else if(n_inv) {
			for(k=0;k<r_func[k1].n_ind;k++) if(r_func[k1].id_list[k]==pivot) {
				r_func[k1].id_list[k]=r_func[k1].id_list[--r_func[k1].n_ind];
				break;
			}
			r_func[k1].flag=0;
		}
		for(k=1;k<=n_ops;k++) prev_inv[k-1]=involved[k];
		n_prev_inv=n_ops;	
		true_involved[n_inv++]=pivot;
	}
	if(!no_peel_flag && !silent_flag) printf("No. non-zero terms = %d, non-zero combinations = %d\n",total_terms,total_comb);
	for(k1=i=0;i<n_rfuncs;i++) if(!r_func[i].flag) {
		if(err) {
			if(r_func[i].index) free(r_func[i].index);
		} else {
			fputs("Unused R-Function on: ",stderr);
			for(j=0;j<r_func[i].n_ind;j++) {
				k=r_func[i].id_list[j];
				print_orig_allele_id(stderr,k);
				fputc(' ',stderr);
			}
			fputc('\n',stderr);
			k1=1;
		}
	}
	if(k1) ABT_FUNC("Aborting\n");
	return err;
}

#ifdef FUNC_NAME
#undef FUNC_NAME
#endif
#define FUNC_NAME "find_sequence"
static int find_sequence(const int n_fam,const int *famlist,const int n_all,struct Peelseq_Head *pp,const int locus,const int comp,const int linktype)
{
	int i,j,n_genes=0,old_rfuncs,*bk_flags=0;
	int *g_perm,*order=0,*order1;
	struct Marker *mark;
	
	if(trace_peel) printf("In find_sequence(%d,%p,%d,%p,%d,%d,%d)\n",n_fam,famlist,n_all,pp,locus,comp,linktype);
	old_rfuncs=n_rfuncs;
	if(n_rfuncs) {
		if(!(bk_flags=malloc(sizeof(int)*n_rfuncs))) ABT_FUNC(MMsg);
		for(i=0;i<n_rfuncs;i++) bk_flags[i]=r_func[i].flag;
	}
	set_flags(n_fam,famlist,comp,locus,linktype);
	for(j=0;j<comp_size[comp];j++) {
		i=perm[id+j];
		if(id_array[i].flag&HAP_MAT) n_genes++;
		if(id_array[i].flag&HAP_PAT) n_genes++;
	}
	if(!n_genes) return 0;
	if(trace_peel) printf("\tn_genes = %d\n",n_genes);
	if(!(involved=malloc(sizeof(struct R_Func)*n_genes*6))) ABT_FUNC(MMsg);
	n_prev_inv=0;
	prev_inv=involved+n_genes;
	true_involved=prev_inv+n_genes;
	rf_flag=true_involved+n_genes;
	g_perm=rf_flag+n_genes;
	order1=g_perm+n_genes;
	n_genes=0;
	for(j=0;j<comp_size[comp];j++) {
		i=perm[id+j];
		if(id_array[i].flag&HAP_MAT) g_perm[n_genes++]=i+1;
		if(id_array[i].flag&HAP_PAT) g_perm[n_genes++]= -(i+1);
	}
	if(locus==n_markers) mark=traitlocus;
	else mark=markers+locus;
	order=min_degree(g_perm,n_genes,linktype);
	for(j=old_rfuncs;j<n_rfuncs;j++) free(r_func[j].id_list);
	n_rfuncs=old_rfuncs;
	set_flags(n_fam,famlist,comp,locus,linktype);
	if(n_rfuncs) {
		for(j=0;j<n_rfuncs;j++) r_func[j].flag=bk_flags[j];
		free(bk_flags);
	}
	i=logical_jntpeel(n_genes,order,g_perm,n_all,pp,linktype);
	if(order) free(order);
	free(involved);
	return i;
}

static void fill_rf(int pivot,int n_all,int n_bits)
{
	int k,k1,k2,k3,piv1;
	unsigned long a;
	
	if(n_rfuncs>=r_func_size) {
		r_func_size<<=1;
		if(!(r_func=realloc(r_func,sizeof(struct R_Func)*r_func_size))) ABT_FUNC(MMsg);
	}
	k=n_rfuncs++;
	if(!(r_func[k].id_list=malloc(sizeof(int)*2))) ABT_FUNC(MMsg);
	r_func[k].id_list[0]=pivot;
	r_func[k].id_list[1]= -pivot;
	r_func[k].n_ind=2;
	r_func[k].flag=0;
	r_func[k].mask=0;
	piv1=ped_recode1[pivot-1]-1-id;
	for(k1=k2=0;k2<n_all;k2++) {
		a=all_set[k2][piv1];
		if(a) for(k3=0;k3<n_all;k3++) if(a&(1<<k3)) k1++;
	}
	r_func[k].n_terms=k1;
	r_func[k].fill=(double)k1/(double)(n_all*n_all);
	if(!(r_func[k].index=malloc(k1*sizeof(lk_ulong)))) ABT_FUNC(MMsg);
	for(k1=k2=0;k2<n_all;k2++) {
		a=all_set[k2][piv1];
		if(a) for(k3=0;k3<n_all;k3++) if(a&(1<<k3))
		  r_func[k].index[k1++]=(lk_ulong)((k3<<n_bits)|k2);
	}
}

#ifdef FUNC_NAME
#undef FUNC_NAME
#endif
#define FUNC_NAME "logical_peel"
static int logical_peel(const int n_fam,int *famlist,const int n_all,const int locus,const int comp,const int linktype)
{
	int i,j,k,k1,k2,k3,k4,pivot,fam,kid,ids,idd,nf,n_bits;
	unsigned long a,b;
	struct Peelseq_Head *pp;
	struct Simple_Element *element;
	struct Marker *mark;
	
	if(trace_peel) printf("In logical_peel(%d,%p,%d,%d,%d,%d)\n",n_fam,famlist,n_all,locus,comp,linktype);
	n_bits=num_bits(n_all);
	peelseq_head.type=0;
	pp= &peelseq_head;
	r_func_size=n_fam+1;
	if(!(r_func=malloc(sizeof(struct R_Func)*r_func_size))) ABT_FUNC(MMsg);
	n_rfuncs=0;
	for(k=j=0;j<comp_size[comp];j++) {
		i=perm[id+j];
		if(id_array[i].flag&4) k++;
	}
	if(k) {
		if(trace_peel) printf("\tFound %d singletons\n",k);
		if(!(element=malloc(sizeof(struct Simple_Element)))) ABT_FUNC(MMsg);
		pp->type=PEEL_SIMPLE;
		pp->ptr.simple=element;
		pp= &element->next;
		pp->type=0;
		element->sire=element->dam=element->pivot=0;
		element->out_index= -1;
		if(!(element->off=malloc(sizeof(int)*k))) ABT_FUNC(MMsg);
		for(k=j=0;j<comp_size[comp];j++)	{
			i=perm[id+j];
			if(id_array[i].flag&4) element->off[k++]=i+1;
		}
		element->n_off=k;
	}
	for(j=0;j<comp_size[comp];j++) {
		i=perm[id+j];
		if(id_array[i].flag&6) continue;
		id_array[i].ngens=id_array[i].nhaps[0]=id_array[i].nhaps[1]=0;
		if(linktype==LINK_Y && id_array[i].sex==2) continue;
		k1=k3=k4=0;
		for(b=0,k=0;k<n_all;k++) {
			a=all_set[k][j];
			if(!a) continue;
			k1++;
			b|=a;
			if(a&(1<<k)) k4++;
			while(a) {
				if(a&1) k3++;
				a>>=1;
			}
		}
		k2=0;
		id_array[i].nhaps[X_MAT]=k1;
		req_set[2][j]=b;
		while(b) {
			if(b&1) k2++;
			b>>=1;
		}
		if(k1==1 && k2==1) {
			id_array[i].flag|=4;
		}
		id_array[i].nhaps[X_PAT]=k2;
		id_array[i].ngens=k3;
		id_array[i].ngens1=k4;
	}
	calc_order(comp,n_fam,famlist,linktype);
	nf=n_fam;
	k1=0;
	if(((syst_var[PEEL_OPTION]&1) && locus<n_markers) 
		|| ((syst_var[PEEL_OPTION]&2) && locus==n_markers)) k1=1;
	if(!k1 && trace_peel && nf) printf("\tBeginning primary family peel (nf = %d)\n",nf);
	if(!k1) for(i=0;i<nf;i++) {
		fam=famlist[i];
		ids=family[fam].sire;
		idd=family[fam].dam;
		pivot=0;
		if(ids) {
			if(id_array[ids-1].order>1) pivot=ids;
			if(id_array[idd-1].order>1) {
				if(pivot) pivot= -2;
				else pivot=idd;
			}
		}
		for(j=0;j<family[fam].nkids;j++) {
			kid=family[fam].kids[j];
			if(id_array[kid].flag&2) continue;
			if(pivot<0) {
				if(id_array[kid].ngens>1) break;
			} else if(id_array[kid].order>1)	{
				if(pivot) break;
				pivot=kid+1;
			}
		}
		if(j==family[fam].nkids) {
			if(!(element=malloc(sizeof(struct Simple_Element)))) ABT_FUNC(MMsg);
			pp->type=PEEL_SIMPLE;
			pp->ptr.simple=element;
			pp= &element->next;
			pp->type=0;
			element->sire=ids;
			element->dam=idd;
			if(ids && ids!=pivot && pivot>=0) id_array[ids-1].flag|=PEELED;
			if(idd && idd!=pivot && pivot>=0) id_array[idd-1].flag|=PEELED;
			element->pivot=pivot;
			for(k=0;k<n_rfuncs;k++) if(!r_func[k].flag) {
				k1=0;
				if(r_func[k].id_list[0]==ids) k1=1;
				else if(r_func[k].id_list[0]==idd) k1=1;
				else for(j=0;j<family[fam].nkids;j++) {
					kid=family[fam].kids[j]+1;
					if(r_func[k].id_list[0]==kid)	{
						k1=1;
						break;
					}
				}
				if(k1) {
					r_func[k].flag=2;
					free(r_func[k].index);
				}
			}
			if(pivot>=0 || !(syst_var[PEEL_OPTION]&4)) {
				if(trace_peel) {
					printf("\tPeeling family: ");
					if(family_id) print_orig_family(stdout,family[fam].kids[0]+1,0);
					print_orig_id1(stdout,ids,0);
					fputc(',',stdout);
					print_orig_id1(stdout,idd,0);
					fputc(' ',stdout);
					for(j=k=0;j<family[fam].nkids;j++) {
						kid=family[fam].kids[j];
						if(id_array[kid].flag&2) continue;
						fputc(k++?',':'(',stdout);
						print_orig_id1(stdout,kid+1,0);
					}
					if(k) fputc(')',stdout);
					fputs(" --> ",stdout);
					if(pivot<0) {
						print_orig_id1(stdout,ids,0);
						fputc(',',stdout);
						print_orig_id1(stdout,idd,0);
					} else if(!pivot) fputc('.',stdout);
					else print_orig_id1(stdout,pivot,0);
					fputc('\n',stdout);
				}
				if(pivot) {
					element->out_index=n_rfuncs;
					if(pivot<0) {
						id_array[ids-1].flag|=WAS_PIVOT;
						id_array[idd-1].flag|=WAS_PIVOT;
						id_array[ids-1].order--;
						id_array[idd-1].order--;
						fill_rf(ids,n_all,n_bits);
						fill_rf(idd,n_all,n_bits);
					} else {
						id_array[pivot-1].flag|=WAS_PIVOT;
						id_array[pivot-1].order--;
						fill_rf(pivot,n_all,n_bits);
					}
				} else element->out_index= -1;
				if(!(element->off=malloc(sizeof(int)*family[fam].nkids))) ABT_FUNC(MMsg);
				for(j=k=0;j<family[fam].nkids;j++) {
					kid=family[fam].kids[j];
					if(id_array[kid].flag&2) continue;
					element->off[k++]=kid+1;
					if((kid+1)!=pivot) id_array[kid].flag|=PEELED;
				}
				element->n_off=k;
				famlist[i]=famlist[--nf];
				i= -1;
			}
		}
	}
	k=0;
	if(((syst_var[PEEL_OPTION]&1) && locus<n_markers) 
		|| ((syst_var[PEEL_OPTION]&2) && locus==n_markers)) k=1;
	if(syst_var[PEEL_OPTION]&4) k=1;
	if(!k && trace_peel && nf) printf("\tBeginning secondary family peel (nf = %d)\n",nf);
	if(!k) for(i=0;i<nf;i++) {
		fam=famlist[i];
		for(k2=j=0;j<family[fam].nkids;j++)	{
			kid=family[fam].kids[j];
			if((id_array[kid].flag&2)||id_array[kid].order>1) continue;
			k2++;
		}
		if(k2>1) {
			ids=family[fam].sire;
			idd=family[fam].dam;
			if(!(element=malloc(sizeof(struct Simple_Element)))) ABT_FUNC(MMsg);
			pp->type=PEEL_SIMPLE;
			pp->ptr.simple=element;
			pp= &element->next;
			pp->type=0;
			element->sire=ids;
			element->dam=idd;
			element->pivot= -1;
			for(k=0;k<n_rfuncs;k++) if(!r_func[k].flag) {
				k1=0;
				if(r_func[k].id_list[0]==ids) k1=1;
				else if(r_func[k].id_list[0]==idd) k1=1;
				else for(j=0;j<family[fam].nkids;j++) {
					kid=family[fam].kids[j];
					if((id_array[kid].flag&2)||id_array[kid].order>1) continue;
					if(r_func[k].id_list[0]==kid+1) {
						k1=1;
						break;
					}
				}
				if(k1) {
					r_func[k].flag=2;
					free(r_func[k].index);
				}
			}
			element->out_index=n_rfuncs;
			if(n_rfuncs>=r_func_size) {
				r_func_size<<=1;
				if(!(r_func=realloc(r_func,sizeof(struct R_Func)*r_func_size))) ABT_FUNC(MMsg);
			}
			k=n_rfuncs++;
			if(!(r_func[k].id_list=malloc(sizeof(int)*4))) ABT_FUNC(MMsg);
			r_func[k].id_list[0]=idd;
			r_func[k].id_list[1]= -idd;
			r_func[k].id_list[2]=ids;
			r_func[k].id_list[3]= -ids;
			r_func[k].n_ind=4;
			r_func[k].flag=0;
			r_func[k].mask=0;
			id_array[ids-1].flag|=WAS_PIVOT;
			id_array[idd-1].flag|=WAS_PIVOT;
			if(!(element->off=malloc(sizeof(int)*k2))) ABT_FUNC(MMsg);
			for(j=k=0;j<family[fam].nkids;j++) {
				kid=family[fam].kids[j];
				if((id_array[kid].flag&2)||id_array[kid].order>1) continue;
				element->off[k++]=kid+1;
				id_array[kid].flag|=PEELED;
			}
			element->n_off=k;
			partial_peel(element,n_all,n_bits);
			if(trace_peel) {
				printf("\tPeeling family: ");
				if(family_id) print_orig_family(stdout,family[fam].kids[0]+1,0);
				print_orig_id1(stdout,ids,0);
				fputc(',',stdout);
				print_orig_id1(stdout,idd,0);
				fputc(' ',stdout);
				for(j=k=0;j<family[fam].nkids;j++) {
					kid=family[fam].kids[j];
					if((id_array[kid].flag&2)||id_array[kid].order>1) continue;
					fputc(k++?',':'(',stdout);
					print_orig_id1(stdout,kid+1,0);
				}
				if(k) fputc(')',stdout);
				fputs(" --> ",stdout);
				print_orig_id1(stdout,ids,0);
				fputc(',',stdout);
				print_orig_id1(stdout,idd,0);
				fputc('\n',stdout);
			}
		}
	}
	if(nf) return find_sequence(nf,famlist,n_all,pp,locus,comp,linktype);
	else if(!silent_flag) {
		if(locus<n_markers) mark=markers+locus;
		else mark=traitlocus;
	}
	return 0;
}

#ifdef FUNC_NAME
#undef FUNC_NAME
#endif
#define FUNC_NAME "write_gene_file_header"
static FILE *write_gene_file_header(char **fname)
{
	FILE *fptr;
	int i,j;
	
	*fname=make_file_name("gen");
	for(i=j=0;i<n_markers;i++) if(markers[i].element->n_levels) j++;
	if(!j && !traitlocus) {
		unlink(*fname);
		free(*fname);
		return 0;
	}
	if(Filter) {
		i=child_open(WRITE,*fname,Filter);
		if(!(fptr=fdopen(i,"w"))) DataFileError(*fname);
		if(errno && errno!=ESPIPE) DataFileError(*fname);
		errno=0;
	} else if(!(fptr=fopen(*fname,"w"))) abt(__FILE__,__LINE__,"%s(): File Error.  Couldn't open '%s' for writing\n",FUNC_NAME,*fname);
	if(fprintf(fptr,"Loki.gen:%x,%x,%x,%x\n",RunID,j,n_comp,syst_var[NO_EXTRA_ALLELE])<0) DataFileError(*fname);
	return fptr;
}

#ifdef FUNC_NAME
#undef FUNC_NAME
#endif
#define FUNC_NAME "write_recoding_info"
static void write_recoding_info(FILE *fptr,char *fname,int comp,int n_all,int n_all1,int fg)
{
	int i,j,k;
	
	for(k=j=0;k<comp_size[comp];k++) {
		i=perm[id+k];
		if(!(id_array[i].flag&2)) j++;
	}
	if(fprintf(fptr,"LKCM:%x,%x,%x,%x,%x\n",comp,comp_size[comp],n_all1,fg,j)<0) DataFileError(fname);
	if(n_all1<2 || !j) return;
	if(n_all!=n_all1) {
		for(i=0;i<n_all;i++) {
			if(i) if(fputc(',',fptr)==EOF) DataFileError(fname);
			if(fprintf(fptr,"%d",allele_trans[i])<0) DataFileError(fname);
		}
		if(fputc('\n',fptr)==EOF) DataFileError(fname);
	}
	for(j=0;j<comp_size[comp];j++) {
		i=perm[id+j];
		k=id_array[i].flag&2;
		if(fprintf(fptr,"%x",k)<0) DataFileError(fname);
		if(!(k || no_peel_flag)) {
			for(k=0;k<n_all1;k++) if(fprintf(fptr,",%lx",all_set[k][j])<0) DataFileError(fname);
			if(fg) if(fprintf(fptr,",%lx,%lx",req_set[0][j],req_set[1][j])<0) DataFileError(fname);
		}
		if(fputc('\n',fptr)==EOF) DataFileError(fname);
	}
}

#ifdef FUNC_NAME
#undef FUNC_NAME
#endif
#define FUNC_NAME "write_peeling_info"
static void write_peeling_info(FILE *fptr,char *fname)
{
	int i,j,k;
	struct Simple_Element *simple_em;
	struct Complex_Element *complex_em;
	struct Peelseq_Head *pp;
	
	if(fputs("LKPL\n",fptr)==EOF) DataFileError(fname);
	pp= &peelseq_head;
	while(pp->type) {
		if(pp->type==PEEL_SIMPLE) {
			simple_em=pp->ptr.simple;
			if(simple_em->sire) {
				simple_em->sire=ped_recode1[simple_em->sire-1];
				simple_em->dam=ped_recode1[simple_em->dam-1];
			}
			if(simple_em->pivot>0) simple_em->pivot=ped_recode1[simple_em->pivot-1];
			if(fprintf(fptr,"%x,%x,%x,%x,%d,%d",pp->type,simple_em->sire,simple_em->dam,
						  simple_em->n_off,simple_em->pivot,simple_em->out_index)<0) DataFileError(fname);
			i=simple_em->n_off;
			for(j=0;j<i;j++) {
				simple_em->off[j]=ped_recode1[simple_em->off[j]-1];
				if(fprintf(fptr,",%x",simple_em->off[j])<0) DataFileError(fname);
			}
			pp= &simple_em->next;
		} else {
			complex_em=pp->ptr.complex;
			if(fprintf(fptr,"%x,%x,%x,%d,%x",pp->type,complex_em->n_peel,complex_em->n_involved,
						  complex_em->out_index,complex_em->n_rfuncs)<0) DataFileError(fname);
			i=complex_em->n_involved;
			for(j=0;j<i;j++) {
				k=complex_em->involved[j];
				if(k>0) k=ped_recode1[k-1];
				else k= -ped_recode1[-1-k];
				complex_em->involved[j]=k;
				if(fprintf(fptr,",%d",complex_em->involved[j])<0) DataFileError(fname);			}
			i=complex_em->n_involved*2+complex_em->n_rfuncs;
			for(;j<i;j++) if(fprintf(fptr,",%x",complex_em->involved[j])<0) DataFileError(fname);
			pp= &complex_em->next;
		}
		if(fputc('\n',fptr)==EOF) DataFileError(fname);
	}
	if(fprintf(fptr,"0,%x\n",n_rfuncs)<0) DataFileError(fname);
	for(j=0;j<n_rfuncs;j++)	{
		if(fprintf(fptr,"%x,%x",r_func[j].n_ind,r_func[j].n_terms)<0) DataFileError(fname);
		for(i=0;i<r_func[j].n_ind;i++) if(fprintf(fptr,",%d",r_func[j].id_list[i])<0) DataFileError(fname);	
		if(fputc('\n',fptr)==EOF) DataFileError(fname);
	}
}

#ifdef FUNC_NAME
#undef FUNC_NAME
#endif
#define FUNC_NAME "get_marker_name"
static char *get_marker_name(const int locus)
{
	char *s,tbuf[32];
	size_t i;
	struct Marker *mark;
	if(locus==n_markers) mark=traitlocus;
	else mark=markers+locus;
	
	if(mark->var->vtype&ST_ARRAY)	{
		sprintf(tbuf,"%d",mark->index);
		i=strlen(tbuf)+strlen(mark->var->name)+3;
		if(!(s=malloc(i))) ABT_FUNC(MMsg);
		sprintf(s,"%s(%s)",mark->var->name,tbuf);
	} else {
		i=strlen(mark->var->name)+1;
		if(!(s=malloc(i))) ABT_FUNC(MMsg);
		strcpy(s,mark->var->name);
	}
	return s;
}

#ifdef FUNC_NAME
#undef FUNC_NAME
#endif
#define FUNC_NAME "Check_Locus"
static int Check_Locus(const int locus,FILE *fptr,char *fname,int component,const int linktype)
{
	int i,j,k,k1,km,kp,fg,ids,idd,kid,n_all,ch[2],comp,all_trans[2],sex,n_comp1;
	int *famlist,*nfam,*all_flag,n_all1,n_fam,kid1,err=0,*tmp,*tmp1,cand_flag=0;
	unsigned long m1,a,b;
	struct Marker *mark;
	FILE *ferr=0;
	char *fname1=0;

	if(trace_peel) printf("In Check_Locus(%d,%p,%s,%d,%d) - silent_flag = %d\n",locus,fptr,fname,component,linktype,silent_flag);
	rec_gen_flag=0;
	if(locus==n_markers) {
		if(!traitlocus) return 0;
		n_all=2;
	} else {
		n_all=markers[locus].element->n_levels;
		if(!silent_flag && n_all) {
			if(!(tmp=calloc((size_t)(n_all*2),sizeof(int)))) ABT_FUNC(MMsg);
			tmp1=tmp+n_all;
			for(i=0;i<ped_size;i++) if(id_array[i].haplo[0]) {
				for(k=0;k<2;k++) if((k1=id_array[i].haplo[k][locus])) tmp[k1-1]++;
			}
			for(k1=k=0;k<n_all;k++) if(tmp[k]) k1++;
			if(k1<n_all) {
				for(k1=k=0;k<n_all;k++) {
					if(tmp[k]) {
						factor_recode[n_factors+locus][k1]=factor_recode[n_factors+locus][k];
						tmp1[k]=++k1;
					} else tmp1[k]= -1;
				}
				markers[locus].element->n_levels=n_all=k1;
				for(i=0;i<ped_size;i++) if(id_array[i].haplo[0]) {
					for(k=0;k<2;k++) if((k1=id_array[i].haplo[k][locus])) id_array[i].haplo[k][locus]=tmp1[k1-1];
				}
				printf("No. alleles reduced to %d\n",n_all);
			}
			free(tmp);
		}
		if(!syst_var[NO_EXTRA_ALLELE]) n_all++;/* Add extra allele to allow for possible extra alleles in population */
	}
	if(n_all<2) return 0;
	if(!(all_flag=malloc(sizeof(int)*n_all))) ABT_FUNC(MMsg);
	if(!(temp_set=malloc(sizeof(long)*n_all*3))) ABT_FUNC(MMsg);
	id_set[0]=temp_set+n_all;
	id_set[1]=temp_set+2*n_all;
	if(!(nfam=malloc((n_families*2+ped_size)*sizeof(int)))) ABT_FUNC(MMsg);
	famflag=nfam+ped_size;
	famlist=famflag+n_families;
	if(locus==n_markers)	{
		mark=traitlocus;
	} else {
		mark=markers+locus;
		if(!(mark->allele_trans=malloc(sizeof(void *)*n_comp))) ABT_FUNC(MMsg);
		if(!(mark->allele_trans[0]=malloc(sizeof(int)*n_comp*n_all))) ABT_FUNC(MMsg);
		for(i=1;i<n_comp;i++) mark->allele_trans[i]=mark->allele_trans[i-1]+n_all;
	}
	mark->order=0;
	no_peel_flag=(locus==n_markers);
	if(!silent_flag) if(fprintf(fptr,"LKMK:%x\n",n_all)<0) DataFileError(fname);
	/* First thing to do is to prune pedigree based on data at this marker.
	 * Remove:
	 *   (a) Untyped individuals with no unpruned descendents
	 *   (b) Untyped founders with only 1 unpruned child
	 */
	for(i=0;i<n_families;i++) famflag[i]=0;
	for(j=0;j<pruned_ped_size;j++) id_array[perm[j]].flag=0;
	if(locus<n_markers) for(j=0;j<pruned_ped_size;j++)	{
		i=perm[j];
		if(id_array[i].haplo[0] && (id_array[i].haplo[0][locus] || id_array[i].haplo[1][locus]))
		  id_array[i].flag=1;
		if(id_array[i].group) id_array[i].flag=1;
		nfam[i]=id_array[i].nfam;
	}
 	if(locus==n_markers || (markers[locus].element->type&ST_MODEL)) for(j=0;j<pruned_ped_size;j++) {
		i=perm[j];
		if(id_array[i].group) id_array[i].flag=1;
		if(!id_array[i].flag && id_array[i].data) for(k=0;k<n_id_records;k++) {
			if((id_elements[k]->type&ST_TRAIT) && id_array[i].data[k].flag) {
				id_array[i].flag|=1;
				break;
			}
		}
		if(!id_array[i].flag && id_array[i].data1) for(k1=0;k1<id_array[i].nrec;k1++) {
			if(id_array[i].data1[k1]) for(k=0;k<n_nonid_records;k++)	{
				if((nonid_elements[k]->type&ST_TRAIT) && id_array[i].data1[k1][k].flag)	{
					id_array[i].flag|=1;
					break;
				}
			}
		}
		nfam[i]=id_array[i].nfam;
	}
	if(syst_var[PRUNE_OPTION]==2) do {
		for(fg=i=0;i<n_families;i++) if(!famflag[i])	{
			ids=family[i].sire;
			idd=family[i].dam;
			for(k=j=0;j<family[i].nkids;j++)	{
				kid=family[i].kids[j];
				if(id_array[kid].flag&2) continue;
				if((id_array[kid].flag&1) || nfam[kid]) k++;
				else {
					id_array[kid].flag|=2;
					fg=1;
				}
			}
			if(k==1) {
				if(ids) {
					if(!(id_array[ids-1].flag&1) && nfam[ids-1]<2) {
						k1=id_array[ids-1].family;
						if(!k1) k|=2;
						else if(famflag[k1-1]) k|=2;
					}
					if(!(id_array[idd-1].flag&1) && nfam[idd-1]<2) {
						k1=id_array[idd-1].family;
						if(!k1) k|=4;
						else if(famflag[k1-1]) k|=4;
					}
					if(k==7) k=0;
				}
			}
			if(!k) {
				famflag[i]=fg=1;
				if(ids) {
					nfam[ids-1]--;
					if(!nfam[ids-1] && !(id_array[ids-1].flag&1)) id_array[ids-1].flag|=2;
					nfam[idd-1]--;
					if(!nfam[idd-1] && !(id_array[idd-1].flag&1)) id_array[idd-1].flag|=2;
				}
			}
		}
	} while(fg);
	/* mkflag[locus] indicates whether this individual has been pruned for
	 * this locus */
	for(k=j=0;j<pruned_ped_size;j++) {
		i=perm[j];
		if(id_array[i].flag&2) {
			k++;
			id_array[i].mkflag[locus]=1;
		} else {
			k1=id_array[i].family;
			if(k1 && famflag[k1-1]) k1=0;
			if(!k1 && !nfam[i]) id_array[i].flag|=4;
			id_array[i].mkflag[locus]=0;
			ids=id_array[i].sire;
			idd=id_array[i].dam;
			if(ids && (id_array[ids-1].mkflag[locus]^id_array[idd-1].mkflag[locus])) ABT_FUNC("Internal error - single parent pruned from family\n");
		}
	}
	id=comp=0;
	if(component) {
		for(;comp<component-1;comp++) id+=comp_size[comp];
		n_comp1=component;
		comp=component-1;
	} else n_comp1=n_comp;
	/* Allele recoding */
	for(;comp<n_comp1;comp++) {
		if(locus<n_markers) allele_trans=mark->allele_trans[comp];
		else allele_trans=all_trans;
		if(locus<n_markers && !(markers[locus].element->type&ST_MODEL) && syst_var[RECODE_OPTION]) {
			/* Find out which alleles are used in this component */
			for(i=0;i<n_all;i++) all_flag[i]=0;
			for(j=0;j<comp_size[comp];j++) {
				i=perm[j+id];
				if((id_array[i].flag&3)!=1) continue;
				if(id_array[i].haplo[0]) {
					k=id_array[i].haplo[0][locus];
					if(k) all_flag[k-1]=1;
					k=id_array[i].haplo[1][locus];
					if(k) all_flag[k-1]=1;
				}
			}
			for(i=n_all1=0;i<n_all;i++) n_all1+=all_flag[i];
			n_all1++;
			if(n_all1>n_all) n_all1=n_all;
		} else n_all1=n_all;
		if(n_all1<n_all) {
			for(j=i=0;i<n_all;i++)
			  if(all_flag[i]) allele_trans[j++]=i;
			for(;j<n_all;j++) allele_trans[j]= -1;
		} else for(i=0;i<n_all;i++) allele_trans[i]=i;
		for(i=0;i<n_all;i++) all_flag[i]=n_all1-1;
		if(locus<n_markers && n_all>markers[locus].element->n_levels) allele_trans[n_all-1]= -1;
		/* While all_flag will (temporarily) have the translation table from the
		 * old codes to the new codes */
		for(i=0;i<n_all;i++) if((j=allele_trans[i])>=0) all_flag[j]=i;
		if((size_t)n_all1>LK_LONG_BIT) abt(__FILE__,__LINE__,"%s(): No. segregating alleles for marker %s exceeds %d\n",FUNC_NAME,markers[locus].var->name,LK_LONG_BIT);
		/* Use all_flag to recode haplotypes */
		if(n_all1<n_all) {
			for(j=0;j<comp_size[comp];j++) {
				i=perm[j+id];
				if((id_array[i].flag&3)!=1) continue;
				if(id_array[i].haplo[0]) {
					k=id_array[i].haplo[0][locus];
					if(k) id_array[i].haplo[0][locus]=all_flag[k-1]+1;
					k=id_array[i].haplo[1][locus];
					if(k) id_array[i].haplo[1][locus]=all_flag[k-1]+1;
				}
			}
			rec_gen_flag=1;
		}
		for(k=j=0;k<comp_size[comp];k++)	{
			i=perm[id+k];
			if(!(id_array[i].flag&2)) j++;
		}
		if(!j || n_all1<2) {
			if(!silent_flag) {
				write_recoding_info(fptr,fname,comp,n_all,n_all1,0);
			}
			id+=comp_size[comp];
			continue;
		}
		/* Genotype elimination */
		if(!(all_set=malloc(sizeof(void *)*n_all1))) ABT_FUNC(MMsg);
		if(!(all_set[0]=calloc((size_t)(comp_size[comp]*(n_all1+3)),sizeof(long)))) ABT_FUNC(MMsg);
		for(i=1;i<n_all1;i++) all_set[i]=all_set[i-1]+comp_size[comp];
		req_set[0]=all_set[i-1]+comp_size[comp];
		req_set[1]=req_set[0]+comp_size[comp];
		req_set[2]=req_set[1]+comp_size[comp];
 		mask=(1<<n_all1)-1;
		for(j=0;j<comp_size[comp];j++) {
			i=perm[j+id];
			ids=id_array[i].sire;
			idd=id_array[i].dam;
			if(id_array[i].flag&2) continue;
			if(linktype && !id_array[i].sex) {
				fputs("Individual ",stderr);
				print_orig_id(stderr,i+1,0);
				fputs(" has unknown sex\n",stderr);
				err=-1;
				continue;
			}
			if(locus<n_markers && (id_array[i].flag&1)) {
				if(id_array[i].haplo[0]) {
					ch[0]=id_array[i].haplo[0][locus];
					ch[1]=id_array[i].haplo[1][locus];
				} else ch[0]=ch[1]=0;
				if(!ch[0] && ch[1]) {
					id_array[i].haplo[0][locus]=ch[1];
					id_array[i].haplo[1][locus]=0;
					ch[0]=ch[1];
					ch[1]=0;
				}
				if(linktype==LINK_Y) {
					if(id_array[i].sex==2 || (ch[1] && (ch[0]!=ch[1]))) {
						fputs("(Y-linked marker): Individual ",stderr);
						print_orig_id(stderr,i+1,0);
						if(id_array[i].sex==2) fputs(" is female and has genetic data - deleting\n",stderr);
						else fputs(" has 2 different haplotypes - deleting\n",stderr);
						if(!fname1)	{
							if(!(fname1=malloc(strlen(marker_name)+6))) ABT_FUNC(MMsg);
							sprintf(fname1,"%s.err1",marker_name);
							ferr=fopen(fname1,"w");
						}
						if(ferr)	{
							fprintf(ferr,"%d ",i+1);
							print_orig_id(ferr,i+1,0);
							fputc('\n',ferr);
						}
						bad_cnt++;
						ch[0]=ch[1]=0;
						id_array[i].haplo[0][locus]=id_array[i].haplo[1][locus]=0;
					}
					ch[1]=0;
				} else if(linktype==LINK_X && id_array[i].sex==1) {
					if(ch[1] && ch[0]!=ch[1]) {
						fputs("(X-linked marker): Individual ",stderr);
						print_orig_id(stderr,i+1,0);
						fputs(" is male with 2 different haplotypes - deleting\n",stderr);
						if(!fname1)	{
							if(!(fname1=malloc(strlen(marker_name)+6))) ABT_FUNC(MMsg);
							sprintf(fname1,"%s.err1",marker_name);
							ferr=fopen(fname1,"w");
						}
						if(ferr) {
							fprintf(ferr,"%d ",i+1);
							print_orig_id(ferr,i+1,0);
							fputc('\n',ferr);
						}
						bad_cnt++;
						ch[0]=ch[1]=0;
						id_array[i].haplo[0][locus]=id_array[i].haplo[1][locus]=0;
						for(k=0;k<n_all1;k++) all_set[k][j]=mask;
					}
					ch[1]=0;
				}
				if(ch[1]) {
					for(k=0;k<n_all1;k++) all_set[k][j]=0;
					all_set[ch[0]-1][j]=1<<(ch[1]-1);
					all_set[ch[1]-1][j]=1<<(ch[0]-1);
				} else if(ch[0]) {
					m1=1<<(ch[0]-1);
					if(linktype && id_array[i].sex==1) {
						for(k1=0;k1<n_all1;k1++) all_set[k1][j]=0;
						if(linktype==LINK_X) all_set[ch[0]-1][j]=1;
						else all_set[0][j]=m1;
					} else {
						for(k1=0;k1<n_all1;k1++) all_set[k1][j]=m1;
						all_set[ch[0]-1][j]=mask;
					}
				} else {
					if(linktype==LINK_AUTO) for(k=0;k<n_all1;k++) all_set[k][j]=mask;
					else if(linktype==LINK_Y) {
						for(k=1;k<n_all1;k++) all_set[k][j]=0;
						if(id_array[i].sex==1) all_set[0][j]=mask;
					} else {
						m1=id_array[i].sex==1?1:mask;
						for(k=0;k<n_all1;k++) all_set[k][j]=m1;
					}
				}
			} else {
				if(linktype==LINK_AUTO) for(k=0;k<n_all1;k++) all_set[k][j]=mask;
				else if(linktype==LINK_Y) {
					for(k=1;k<n_all1;k++) all_set[k][j]=0;
					if(id_array[i].sex==1) all_set[0][j]=mask;
				} else {
					m1=id_array[i].sex==1?1:mask;
					for(k=0;k<n_all1;k++) all_set[k][j]=m1;
				}
			}
		}
		if(err) ABT_FUNC("Aborting\n");
		for(n_fam=j=0;j<n_families;j++) if(!famflag[j])	{
			if(id_array[family[j].kids[0]].component==(comp+1)) famlist[n_fam++]=j;
		}
		if(locus<n_markers && n_fam) {
			for(j=0;j<n_fam;j++)	{
				k=famlist[j];
				ids=family[k].sire;
				idd=family[k].dam;
				if(!ids)	{
					kid=family[k].kids[0];
					if(id_array[kid].flag&1) famflag[k]=1;
				} else if((id_array[ids-1].flag&1) || (id_array[idd-1].flag&1)) famflag[k]=1;
				else {
					for(i=0;i<family[k].nkids;i++) {
						kid=family[k].kids[i];
						if(id_array[kid].flag&1) {
							famflag[k]=1;
							break;
						}
					}
				}
			}
			do {
				k=0;
				for(j=0;j<n_fam;j++) {
					i=famlist[j];
					if(!famflag[i]) continue;
					err=DoFamily(i,n_all1,locus,linktype);
					k=1;
					if(err) {
						if(!silent_flag) ABT_FUNC("Pedigree errors, aborting\n");
						err=i+1;
						break;
					}
				}
			} while(k && !err);
		}
		if(err) {
			free(all_set[0]);
			free(all_set);
			break;
		}
		if(locus<n_markers && syst_var[RECODE_OPTION]==2) {
			/* Test if this is a candidate gene */
			if(markers[locus].element->type&ST_MODEL) cand_flag=1;
			/* Find required allele set for each individual (used for set recoding */
			for(j=comp_size[comp]-1;j>=0;j--) {
				i=perm[j+id];
				sex=id_array[i].sex;
				req_set[0][j]=req_set[1][j]=0;
				if(id_array[i].flag&2) continue;
				if(linktype==LINK_Y && sex==2) continue;
				if(id_array[i].flag&1) {
					if(id_array[i].haplo[0]) {
						ch[0]=id_array[i].haplo[0][locus];
						ch[1]=id_array[i].haplo[1][locus];
					} else ch[0]=ch[1]=0;
					if(linktype==LINK_Y || (linktype==LINK_X && sex==1)) {
						if(ch[0]) {
							req_set[X_PAT][j]|=1<<(ch[0]-1);
							continue;
						}
					} else {
						if(ch[0]) {
							for(k=0;k<2;k++) if(ch[k])	{
								m1=1<<(ch[k]-1);
								if(all_set[ch[k]-1][j]) req_set[X_MAT][j]|=m1;
								for(k1=0;k1<n_all1;k1++) if(all_set[k1][j]&m1) req_set[X_PAT][j]|=m1;
							}
						}
						if(ch[1]) continue;
					}
				}
				for(k=0;k<id_array[i].nkids;k++) {
					kid=id_array[i].kids[k];
					if(id_array[kid].flag&2) continue;
					kid1=ped_recode1[kid]-1-id;
					if(linktype==LINK_Y)	{
						if(id_array[kid].sex==2) continue;
						a=req_set[X_PAT][kid1];
						b=all_set[0][j];
						req_set[X_PAT][j]|=(a&b);
					} else if(linktype==LINK_X) {
						if(sex==1 && id_array[kid].sex==1) continue;
						a=req_set[2-sex][kid1];
						for(k1=0;k1<n_all1;k1++) {
							b=all_set[k1][j];
							if(b)	{
								if(id_array[kid].sex==2) req_set[X_PAT][j]|=(a&b);
								if(sex==2) {
									m1=1<<k1;
									if(a&m1) req_set[X_MAT][j]|=m1;
								}
							}
						}
					} else {
						a=req_set[2-sex][kid1];
						for(k1=0;k1<n_all1;k1++) {
							b=all_set[k1][j];
							if(b) {
								req_set[X_PAT][j]|=(a&b);
								m1=1<<k1;
								if(a&m1) req_set[X_MAT][j]|=m1;
							}
						}
					}
				}
			}
			/* Find which individuals we can do set recoding on, lump appropriate alleles together */
			for(j=0;j<comp_size[comp];j++) {
				i=perm[j+id];	
				if(id_array[i].flag&2) continue;
				sex=id_array[i].sex;
				if(linktype==LINK_Y && sex==2) continue;
				/* Don't recode candidate gene if individual has data */
				if(cand_flag && (id_array[i].flag&1)) {
					req_set[0][j]=req_set[1][j]=0;
					continue;
				}
				/* Convert from required set to unrequired set by xoring with possible set */
				for(b=0,k1=0;k1<n_all1;k1++) {
					a=all_set[k1][j];
					if(!a) continue;
					if(linktype!=LINK_Y)	{
						m1=1<<k1;
						req_set[X_MAT][j]^=m1;
					}
					b|=a;
				}
				if(linktype!=LINK_X || sex==2) req_set[X_PAT][j]^=b;
				/* Find if there are >1 possible but unrequired maternal alleles which can
				 * be lumped into allele km */
				k1=km=0;
				if(linktype!=LINK_Y)	{
					a=req_set[X_MAT][j];
					while(a)	{
						k1++;
						if(a&1) km=k1;
						a>>=1;
						if(km) break;
					}
					if(!a) {
						km=0;
						req_set[X_MAT][j]=0;
					}
				}
				/* Ditto for paternal alleles */
				k1=kp=0;
				if(linktype!=LINK_X || sex==2) {
					a=req_set[X_PAT][j];
					while(a)	{
						k1++;
						if(a&1) kp=k1;
						a>>=1;
						if(kp) break;
					}
					if(!a) {
						kp=0;
						req_set[X_PAT][j]=0;
					}
				}
				/* Lump together possible sets for unused alleles, zero remainder */
				if(km || kp) {
					if(km) {
						for(b=0,k1=0;k1<n_all1;k1++) {
							a=all_set[k1][j];
							if(!a) continue;
							m1=1<<k1;
							if(req_set[X_MAT][j]&m1) {
								b|=a;
								all_set[k1][j]=0;
							}
						}
						all_set[km-1][j]=b;
					}
					if(kp) {
						m1=1<<(kp-1);
						for(k1=0;k1<n_all1;k1++) {
							a=(req_set[X_PAT][j]&all_set[k1][j]);
							if(a) {
								all_set[k1][j]^=a;
								all_set[k1][j]|=m1;
							}
						}
					}
				}
			}
			if(!silent_flag) write_recoding_info(fptr,fname,comp,n_all,n_all1,1);
		} else {
			if(!silent_flag) write_recoding_info(fptr,fname,comp,n_all,n_all1,0);
			for(j=comp_size[comp]-1;j>=0;j--) {
				i=perm[j+id];
				req_set[0][j]=req_set[1][j]=0;
			}
		}
		if(silent_flag!=1) {
/*			if(locus<n_markers) for(k=0;k<n_fam;k++) {
				PrintFamily(famlist[k],locus,n_all1,linktype,1);
			} */
			err=logical_peel(n_fam,famlist,n_all1,locus,comp,linktype);
			free_hash_blocks();
			if(!err && !silent_flag) write_peeling_info(fptr,fname);
			for(k=0;k<n_rfuncs;k++) free(r_func[k].id_list);
			free(r_func);
			free_peelseq(&peelseq_head);
		}
		free(all_set[0]);
		free(all_set);
		id+=comp_size[comp];
		if(err) break;
	}
	if(locus<n_markers) mark=markers+locus;
	else mark=traitlocus;
	free(nfam);
	free(temp_set);
	free(all_flag);
	if(silent_flag && locus!=n_markers) {
		free(mark->allele_trans[0]);
		free(mark->allele_trans);
		mark->allele_trans=0;
	}
	if(ferr) fclose(ferr);
	if(fname1) free(fname1);
	return err;
}

static void blank_ind(const int i,int *blank,const int locus,const int fam)
{
	int j,k;
	
	j=ped_recode1[i-1];
	if(j && blank[j-1]>=0) {
		blank[j-1]=fam+1;
		for(k=0;k<2;k++) id_array[i-1].haplo[k][locus]=0;
	}
}

void blank_fam(const int fam,int *blank,const int locus)
{
	int i,k;
	
	i=family[fam-1].sire;
	if(i) blank_ind(i,blank,locus,fam);
	i=family[fam-1].dam;
	if(i) blank_ind(i,blank,locus,fam);
	for(k=0;k<family[fam-1].nkids;k++) {
		i=family[fam-1].kids[k]+1;
		blank_ind(i,blank,locus,fam);
	}
	fputc('.',stdout);
	fflush(stdout);
}

static void add_to_list(const int i,int *j,int *list)
{
	int k;
	
	for(k=0;k<(*j);k++) if(list[k]==i) break;
	if(k==(*j)) list[(*j)++]=i;
}

#ifdef FUNC_NAME
#undef FUNC_NAME
#endif
#define FUNC_NAME "find_id_code"
static int find_id_code(char *buf,int type,int fam)
{
	struct label_node *node=0;
	int i,flag;
	char *p;
	
	flag=fam<0?1:0;
	if(type==INTEGER) {
		i=(int)strtol(buf,&p,10);
		if(*p) fprintf(stderr,"Garbage after id code '%s'\n",buf);
		else node=find_node(&i,type,flag);
	} else node=find_node(buf,type,flag);
	if(node) i=rec_tab[node->index];
	else {
		if(strcmp("*",buf)) {
			fprintf(stderr,"Id code '%s' not found\n",buf);
			i=-1;
		} else i=0;
	}
	if(fam && i>=0) i=rec_tab1[(i-1)*n_orig_families+fam-1];
	else if(flag) i=-i;
	return i;
}

#ifdef FUNC_NAME
#undef FUNC_NAME
#endif
#define FUNC_NAME "Genotype_Elimination"
int Genotype_Elimination(int check_flag)
{
	FILE *fptr,*flog,*ferr;
	char *fname,*fname1;
	char bf[256],bf1[256],bf2[256],bf3[256];
	int i,j,j1,ids,idd,locus,*bk[2],k,k1,k2,k3,k4,sc,fam1,fam,*tlist,*fam_list,linktype,*blank=0;
	struct Link *pl;

	errno=0;
	trace_peel=syst_var[PEEL_TRACE];
	fptr=write_gene_file_header(&fname);
	if(!fptr) return 0;
	if(check_flag) {
		if(!(bk[0]=malloc(sizeof(int)*2*pruned_ped_size))) ABT_FUNC(MMsg);
		bk[1]=bk[0]+pruned_ped_size;
		if(!(blank=malloc(sizeof(int)*pruned_ped_size))) ABT_FUNC(MMsg);
	}
	if(!(perm=malloc(sizeof(int)*(n_families+2*pruned_ped_size)))) ABT_FUNC(MMsg);
	tlist=perm+pruned_ped_size;
	fam_list=tlist+pruned_ped_size;
	for(i=0;i<ped_size;i++) {
		j=ped_recode1[i];
		if(j) perm[j-1]=i;
	}
	for(locus=0;locus<n_markers+(traitlocus?1:0);locus++) {
		bad_cnt=0;
		if(locus<n_markers) {
			if(!markers[locus].element->n_levels) continue;
			pl=links;
			i=markers[locus].link-1;
			while(i && pl) {
				pl=pl->next;
				i--;
			}
			if(!pl) ABT_FUNC("Invalid linkage group\n");
			linktype=pl->type;
		} else linktype=LINK_AUTO;
		ferr=0;
		fname1=0;
		marker_name=get_marker_name(locus);
		if(locus==n_markers) printf("Processing trait locus '%s'\n",marker_name);
		else printf("Processing marker '%s' (no. alleles = %d)\n",marker_name,markers[locus].element->n_levels);
		if(!check_flag || locus==n_markers) {
			silent_flag=0;
			(void)Check_Locus(locus,fptr,fname,0,linktype);
		} else if(check_flag) {
			for(j=0;j<pruned_ped_size;j++) {
				i=perm[j];
				if(id_array[i].haplo[0] && (id_array[i].haplo[0][locus] || id_array[i].haplo[1][locus])) {
					blank[j]=0;
					for(k=0;k<2;k++) bk[k][j]=id_array[i].haplo[k][locus];
				} else blank[j]= -1;
			}
			silent_flag=2;
			fam=Check_Locus(locus,fptr,fname,0,linktype);
			for(j=0;j<pruned_ped_size;j++) if(blank[j]>=0) {
				i=perm[j];
				for(k=0;k<2;k++) id_array[i].haplo[k][locus]=bk[k][j];
			}
			if(fam) {
				if(fam<0) ABT_FUNC("Aborting\n");
				fputs("Genotype inconsistency - searching for errors\n",stderr);
				if(!(fname1=malloc(strlen(marker_name)+5))) ABT_FUNC(MMsg);
				sprintf(fname1,"%s.err",marker_name);
				if((ferr=fopen(fname1,"r"))) {
					k1=0;
					fprintf(stderr,"Reading 'bad' individuals from %s\n",fname1);
					for(;;) {
						if(family_id) {
							i=fscanf(ferr,"%255s %255s %255s %255s",bf3,bf,bf1,bf2);
							if(i!=4) break;
							j=family_recode[0]->type;
							j1=find_id_code(bf3,j,-1);
						} else {
							i=fscanf(ferr,"%255s %255s %255s",bf,bf1,bf2);
							if(i!=3) break;
							j1=0;
						}
						j=ped_recode[0]->type;
						i=find_id_code(bf,j,j1);
						ids=find_id_code(bf1,j,j1);
						idd=find_id_code(bf2,j,j1);
						if(i<1 || id<0 || ids<0) {
							k1=1;
							break;
						}
						j=ped_recode1[i-1]; 
						if(!j || blank[j-1]== -1) {
							k1=1;
							fprintf(stderr,"Individual %s specified with no marker data\n",bf);
							break;
						}
						if(i==ids || i==idd) {
							for(k=0;k<id_array[i-1].nfam;k++) {
								k2=id_array[i-1].famlist[k];
								if(ids==family[k2].sire && idd==family[k2].dam) {
									blank[j-1]=k2+1;
									break;
								}
							}
							if(k==id_array[i-1].nfam) {
								fprintf(stderr,"Can not find family for id %s, parents %s,%s\n",bf,bf1,bf2);
								k1=1;
								break;
							}
						} else {
							blank[j-1]=id_array[i-1].family;
							if(!blank[j-1]) {
								fprintf(stderr,"Marker error file corrupt - individual %s does not match family (%s %s)\n",bf,bf1,bf2);
								k1=1;
								break;
							}
						}
						if(!blank[j-1]) {
							ABT_FUNC("Internal error - non-blank blanked individual...\n");
						}
						if(ids!=family[blank[j-1]-1].sire || idd!=family[blank[j-1]-1].dam) {
							fprintf(stderr,"Marker error file corrupt - individual %s does not match family (%s %s)\n",bf,bf1,bf2);
							k1=1;
							break;
						}
						for(k=0;k<2;k++) id_array[i-1].haplo[k][locus]=0;
					}
					fclose(ferr);
					ferr=0;
					if(!k1) {
						fam=Check_Locus(locus,fptr,fname,0,linktype);
						if(fam) {
							k1=1;
							fprintf(stderr,"Resulting configuration not consistent\n");
						} else for(j=0;j<pruned_ped_size;j++) if(blank[j]>=0) {
							i=perm[j];
							for(k=0;k<2;k++) id_array[i].haplo[k][locus]=blank[j]?0:bk[k][j];
						}
					}
				} else k1=1;
				if(k1) {
					fputs("Searching for 'bad' subset\n - Pass 1: ",stdout);
					fflush(stdout);
					for(j=0;j<pruned_ped_size;j++) if(blank[j]>0) {
						i=perm[j];
						blank[j]=0;
						for(k=0;k<2;k++) id_array[i].haplo[k][locus]=bk[k][j];
					}
					sc=0;
					sig_caught=0;
					catch_sigs=1;
					silent_flag=1;
					k4=0;
					while(fam && !sig_caught) {
						add_to_list(fam,&k4,fam_list);
						k2=0;
						/* Put all (non-pruned) family members into tlist */
						i=family[fam-1].sire;
						if(i) {
							j=ped_recode1[i-1];
							if(j) tlist[k2++]=j-1;
						}
						i=family[fam-1].dam;
						if(i) {
							j=ped_recode1[i-1];
							if(j) tlist[k2++]=j-1;
						}
						for(k1=0;k1<family[fam-1].nkids;k1++) {
							i=family[fam-1].kids[k1];
							j=ped_recode1[i];
							if(j) tlist[k2++]=j-1;
						}
						/* blank genotypes for everyone in the family */
						for(k1=0;k1<k2;k1++)	{
							j=tlist[k1];
							if(blank[j]>=0) {
								i=perm[j];
								blank[j]=fam;
								for(k=0;k<2;k++) id_array[i].haplo[k][locus]=0;
							}
						}
						fam1=Check_Locus(locus,fptr,fname,0,linktype);
						if(rec_gen_flag) for(j=0;j<pruned_ped_size;j++) if(blank[j]>=0)	{
							i=perm[j];
							for(k=0;k<2;k++) id_array[i].haplo[k][locus]=blank[j]?0:bk[k][j];
						}
						/* Is the same family still giving problems? */
						if(fam1==fam) {
							for(k1=0;k1<k2;k1++)	{
								j=tlist[k1];
								i=perm[j];
								k3=id_array[i].family;
								if(k3!=fam && k3) {
									blank_fam(k3,blank,locus);
									add_to_list(k3,&k4,fam_list);
								}
								for(k=0;k<id_array[i].nfam;k++) {
									k3=id_array[i].famlist[k]+1;
									if(k3!=fam) {
										blank_fam(k3,blank,locus);
										add_to_list(k3,&k4,fam_list);
									}
								}
							}
							fam1=Check_Locus(locus,fptr,fname,0,linktype);
							if(rec_gen_flag) for(j=0;j<pruned_ped_size;j++) if(blank[j]>=0) {
								i=perm[j];
								for(k=0;k<2;k++) id_array[i].haplo[k][locus]=blank[j]?0:bk[k][j];
							}
							if(fam1==fam){
								fprintf(stderr,"Sort of bug - I can't cope with this situation\n");
								silent_flag=0;
								(void)Check_Locus(locus,fptr,fname,0,linktype);
							}
						}
						fam=fam1;
						fputc('.',stdout);
						fflush(stdout);
					}
 					silent_flag=1;
					printf("\n - Pass 2 %4d",k4);
					fflush(stdout);
					while(k4 && !sig_caught) {
						k4--;
						fam1=fam_list[k4];
						k2=0;
						i=family[fam1-1].sire;
						if(i)	{
							j=ped_recode1[i-1];
							if(j && blank[j-1]>=0) tlist[k2++]=j-1;
						}
						i=family[fam1-1].dam;
						if(i) {
							j=ped_recode1[i-1];
							if(j && blank[j-1]>=0) tlist[k2++]=j-1;
						}
						for(k1=0;k1<family[fam1-1].nkids;k1++) {
							i=family[fam1-1].kids[k1];
							j=ped_recode1[i];
							if(j && blank[j-1]>=0) tlist[k2++]=j-1;
						}
						for(k1=0;k1<k2;k1++)	{
							j=tlist[k1];
							i=perm[j];
							blank[j]=0;
							for(k=0;k<2;k++) id_array[i].haplo[k][locus]=bk[k][j];
						}
						fam=Check_Locus(locus,fptr,fname,0,linktype);
						if(rec_gen_flag) for(j=0;j<pruned_ped_size;j++) if(blank[j]>=0) {
							i=perm[j];
							for(k=0;k<2;k++) id_array[i].haplo[k][locus]=blank[j]?0:bk[k][j];
						}
						if(fam) {
							for(;k1>0;k1--) {
								j=tlist[k1-1];
								i=perm[j];
								blank[j]=fam1;
								for(k=0;k<2;k++) id_array[i].haplo[k][locus]=0;
								fam=Check_Locus(locus,fptr,fname,0,linktype);
								if(fam) blank[j]=0;
								if(rec_gen_flag) {
									for(j=0;j<pruned_ped_size;j++) if(blank[j]>=0) {
										i=perm[j];
										for(k=0;k<2;k++) id_array[i].haplo[k][locus]=blank[j]?0:bk[k][j];
									}
								} else if(fam) for(k=0;k<2;k++) id_array[i].haplo[k][locus]=bk[k][j];
								if(!fam) break;
							}
							if(!k1) {
								for(k1=0;k1<k2;k1++)	{
									j=tlist[k1];
									i=perm[j];
									blank[j]=fam1;
									for(k=0;k<2;k++) id_array[i].haplo[k][locus]=0;
								}
								for(k1=0;k1<k2;k1++)	{
									j=tlist[k1];
									i=perm[j];
									blank[j]=0;
									for(k=0;k<2;k++) id_array[i].haplo[k][locus]=bk[k][j];
									fam=Check_Locus(locus,fptr,fname,0,linktype);
									if(fam) blank[j]=fam1;
									if(rec_gen_flag) {
										for(j=0;j<pruned_ped_size;j++) if(blank[j]>=0) {
											i=perm[j];
											for(k=0;k<2;k++) id_array[i].haplo[k][locus]=blank[j]?0:bk[k][j];
										}
									} else if(fam) for(k=0;k<2;k++) id_array[i].haplo[k][locus]=0;
								}
							}
						}
						printf("\b\b\b\b%4d",k4);
						fflush(stdout);
					}
					fputc('\n',stdout);
					for(sc=j=0;j<pruned_ped_size;j++) if(blank[j]>0) sc++;
					sig_caught=0;
					catch_sigs=0;
					ferr=fopen(fname1,"w");
				}
				silent_flag=2;
				if(Check_Locus(locus,fptr,fname,0,linktype)) {
					ABT_FUNC("Internal error - OOOK!\n");
				}
				if(rec_gen_flag) for(j=0;j<pruned_ped_size;j++) if(blank[j]>=0) {
					i=perm[j];
					for(k=0;k<2;k++) id_array[i].haplo[k][locus]=blank[j]?0:bk[k][j];
				}
				if(LogFile) flog=fopen(LogFile,"a");
				else flog=0;
				if(flog) {
					fputs("\n**************** Removing genotype errors ***************\n\n",flog);
					fprintf(flog,"Processing marker '%s'\nDeleted subset follows:\n\n",marker_name);
				}
				k1=k2=sc=0;
				for(j=0;j<pruned_ped_size;j++) if(blank[j]>=0) {
					k2++;
					i=perm[j];
					if(blank[j]>0) {
						sc++;
						if(flog) print_orig_id(flog,i+1,1);
						if(ferr) {
							if(family_id) print_orig_family(ferr,i+1,0);
							print_orig_id1(ferr,i+1,1);
							print_orig_id1(ferr,family[blank[j]-1].sire,1);
							print_orig_id1(ferr,family[blank[j]-1].dam,1);
						}
						if(ferr) fputc('\n',ferr);
						if(k1==10) {
							if(flog) fputc('\n',flog);
							k1=0;
						} else k1++;
					}
				}
				sc+=bad_cnt;
				k2+=bad_cnt;
				printf("(%d out of %d = %.3g%%)\n",sc,k2,100.0*(double)sc/(double)k2);
				if(flog) {
					fprintf(flog," (%d out of %d = %.3g%%)\n",sc,k2,100.0*(double)sc/(double)k2);
					fclose(flog);
				}
				if(ferr) fclose(ferr);
				silent_flag=0;
				locus--;
				free(fname1);
			} else {
				silent_flag=0;
				(void)Check_Locus(locus,fptr,fname,0,linktype);
			}
		}
		free(marker_name);
	}
	if(fwrite("Lgen.end",8,1,fptr)!=1) DataFileError(fname);
	if(fclose(fptr)) DataFileError(fname);
	free(fname);
	free(perm);
	if(check_flag) {
		free(bk[0]);
		free(blank);
	}
	if(Filter) do i=wait(&j); while(i>0);
	return locus<(n_markers+(traitlocus?1:0))?1:0;
}
