/****************************************************************************
 *                                                                          *
 *     Loki - Programs for genetic analysis of complex traits using MCMC    *
 *                                                                          *
 *             Simon Heath - University of Washington                       *
 *                                                                          *
 *                       July 1997                                          *
 *                                                                          *
 * loki.c:                                                                  *
 *                                                                          *
 ****************************************************************************/

#include <config.h>

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <sys/wait.h>
#include <unistd.h>
#include <signal.h>
#include <strings.h>

#include "version.h"
#include "ranlib.h"
#include "utils.h"
#include "loki.h"
#include "loki_peel.h"
#include "loki_ibd.h"
#include "sample_rand.h"
#include "loki_tlmoves.h"
#include "sample_effects.h"
#include "sample_nu.h"
#include "calc_var_locus.h"

unsigned int RunID;
int nrm_flag,ranseed_set,catch_sigs,sig_caught,read_dump_flag,append_output_flag;
char *file_prefix="loki",*file_dir;
char *LogFile,*Filter,*Seedfile;
struct remember *RemBlock;
static struct remember *FirstRemBlock;
time_t start_time;
double extra_time,extra_stime,extra_utime;

/* Called when program receives signal */
static void int_handler(int i)
{
	char *sigs[]={"SIGTERM","SIGQUIT","SIGHUP","SIGINT","UNKNOWN"};
	int j,signals[]={SIGTERM,SIGQUIT,SIGHUP,SIGINT};
	
	for(j=0;j<4;j++) if(i==signals[j]) break;
	if(catch_sigs && !sig_caught) {
		sig_caught=i;
		fprintf(stderr,"Received signal %s, please wait while %s closes down.\nRepeat signal to abort immediately.\n",sigs[j],LOKI_NAME);
		return;
	}
	fprintf(stderr,"Exiting on signal %s (%d)\n",sigs[j],i);
	exit(i);
}

static void FreeStuff(void)
{
	int i,j;
	
	if((ranseed_set&3)==2) {
		if(Seedfile) writeseed(Seedfile,1);
		else writeseed("seedfile",1);
	}
	if(n_markers && marker[0].seg[0]) free(marker[0].seg[0]);
	if(Seedfile) free(Seedfile);
	if(LogFile) free(LogFile);
	if(Filter) free(Filter);
	if(Dumpfile) free(Dumpfile);
	if(Freqfile) free(Freqfile);
	if(Outputfile) free(Outputfile);
	if(id_array) free(id_array);
	if(marker) free(marker);
	if(linkage) {
		for(i=0;i<n_links;i++) {
			if(linkage[i].ibd_list) {
				free(linkage[i].ibd_list->pos);
				free(linkage[i].ibd_list);
			}
		}
		free(linkage);
	}
	free_IBD();
	free_rand();
	free_sample_effects();
	if(use_student_t) free_sample_nu();
	seg_dealloc();
	peel_dealloc();
	free_complex_mem();
	meiosis_scan(-1);
	TL_Free();
	if(tlocus) {
		for(i=0;i<n_tloci;i++) {
			if(tlocus[i].seg[0]) free(tlocus[i].seg[0]);
			if(tlocus[i].freq) {
				free(tlocus[i].eff);
				free(tlocus[i].freq);
			}
		}
		free(tlocus);
	}
	if(all_set) {
		if(n_markers && req_set[0][0]) free(req_set[0][0]);
		if(n_markers && allele_trans[0])	{
			if(allele_trans[0][0]) free(allele_trans[0][0]);
			free(allele_trans[0]);
		}
		if(n_markers && all_set[0]) free(all_set[0]);
		if(r_func[0]) {
			for(i=0;i<n_markers+tlocus_flag;i++)
			  for(j=0;j<n_comp;j++)
				 if(r_func[i][j]) free(r_func[i][j]);
			free(r_func[0]);
		}
		if(peelseq_head[0]) {
			for(i=0;i<n_markers+tlocus_flag;i++)
			  for(j=0;j<n_comp;j++)
				 free_peelseq(peelseq_head[i]+j);
			free(peelseq_head[0]);
		}
		free(all_set);
	}
	if(comp_npeel) free(comp_npeel);
	calc_var_locus(n_markers);
	if(AIMatrix) {
		for(i=0;i<n_comp;i++) if(AIMatrix[i]) free(AIMatrix[i]);
		free(AIMatrix);
	}
	if(comp_size) free(comp_size);
	if(model.term) free(model.term);
	if(id_variable) free(id_variable);
	else if(nonid_variable) free(nonid_variable);
	FreeRemem(FirstRemBlock);
	return;
}

static int find_order1(int n1,int n,int i,int *done,int j)
{
	if(done[i]) return j;
	done[i]= ++j;
	if(n1!=i) j=find_order1(n1,i,(n1+i)>>1,done,j);
	if(n!=i) j=find_order1(i,n,(n+i)>>1,done,j);
	return j;
}

#ifdef FUNC_NAME
#undef FUNC_NAME
#endif
#define FUNC_NAME "main"
int main(int argc,char *argv[])
{
	FILE *fptr;
	int i,pid,c;
	struct sigaction s_action;
	char *tfile=0;
	
	start_time=time(0);
	while((c=getopt(argc,argv,"arvd:"))!=-1) switch(c) {
	 case 'd':
		i=strlen(optarg);
		if(tfile) free(tfile);
		if(!(tfile=malloc(i+1))) ABT_FUNC(MMsg);
		strcpy(tfile,optarg);
		read_dump_flag=1;
		break;
	 case 'r':
		if(!tfile) tfile=make_file_name("dump");
		read_dump_flag=1;
		break;
	 case 'a':
		append_output_flag=1;
		break;
	 case 'v':
		printf("%s\n",LOKI_NAME);
		return 0;
	}
	if(optind>=argc) abt(__FILE__,__LINE__,"No parameter file specified\n");
	if(!(FirstRemBlock=malloc(sizeof(struct remember)))) ABT_FUNC(MMsg);
	RemBlock=FirstRemBlock;
	RemBlock->pos=0;
	RemBlock->next=0;
	model.term=0;
	/* If program terminates normally or via a signal I want it to go through
	 * FreeStuff() so that Memory can be cleared so I can check for unfreed blocks
	 */
	atexit(FreeStuff);
	atexit(print_end_time);
	s_action.sa_handler=int_handler;
	s_action.sa_flags=0;
	/* I think these are all the signals I'm interested in...
	 (Can these be set with a single call? */
	sigaction(SIGINT,&s_action,0L);
	sigaction(SIGHUP,&s_action,0L);
	sigaction(SIGQUIT,&s_action,0L);
	sigaction(SIGTERM,&s_action,0L);
	ReadBinFiles();
	print_start_time(LOKI_NAME,"a");
	do pid=wait(&i); while(pid>0); /* Wait for child processes to finish */
	AllocLokiStruct();
 	if((fptr=fopen(argv[optind],"r"))) i=ReadParam(fptr);
	else {
		printf("Couldn't open '%s' for input as parameter file\nAborting...\n",argv[optind]);
		exit(EXIT_FAILURE);
	}
	fclose(fptr);
	if(i) exit(EXIT_FAILURE);
	if(tfile) {
		if(Dumpfile) free(Dumpfile);
		Dumpfile=tfile;
	}
	if(!(ranseed_set&2))	{
		if(!Seedfile) {
			if(getseed("seedfile")) init_ranf(135421);
		} else if(getseed(Seedfile)) init_ranf(135421);
		ranseed_set|=2;
	}
	if(LogFile)	{
		fptr=fopen(LogFile,"a");
		fputs("\n     Starting state of RNG (seedfile):\n\n",fptr);
		(void)dumpseed(fptr,1);
		fputc('\n',fptr);
		fclose(fptr);
	}
	LokiSetup();
	AllocEffects();
	InitValues();
	seg_alloc();
	peel_alloc();
	SampleLoop();
	return sig_caught;
}
