/****************************************************************************
 *                                                                          *
 *     Loki - Programs for genetic analysis of complex traits using MCMC    *
 *                                                                          *
 *             Simon Heath - University of Washington                       *
 *                                                                          *
 *                       July 1997                                          *
 *                                                                          *
 * loki.h:                                                                  *
 *                                                                          *
 ****************************************************************************/

#ifdef USE_DMALLOC
#include <dmalloc.h>
#endif

#include "ranlib.h"
#include "loki_struct.h"

#define LUMPED_ALLELE "__LUMPED__"

#ifndef M_PI
#define M_PI		3.14159265358979323846
#endif

struct Id_Record {
	int sire,dam;
	int comp; /* Component */
	int fam_code; /* Original family designation */
	int *pruned_flag; /* Pruning on a marker x marker basis */
	int *haplo[2];
	int nhaps[2],ngens;
	unsigned long temp[2];
	int rfp;
	int sex,affected,group;
	int nkids,*kids;
	int flag;
	int n_rec;
	int allele[2];
	double tp[4],tpp[2][2];
	double *res,*cens,*vv,bv;
	int lumped[2];
	struct id_data *data,**data1;
};

struct output_gen {
	struct output_gen *next;
	char *file;
	int link_group;
};

union arg_type {
	char *string;
	int value;
};

struct Id_Recode {
	int flag;
	union arg_type *recode;
};

struct IBD_List {
	int idx,size;
	double *pos;
};

struct Marker {
	char *name;
	int index;
	int n_alleles; /* In entire pedigree */
	int *n_all1;   /* In each component */
	int link_group;
	int rec_flag;
	int *genes[2];
	int **group;
	int lumped;
	union arg_type *recode;
	double pos[2];
	double *lk_store;
	int pos_set;
	double **freq;
	double **counts;
	int *gt; /* Genotypes */
	struct Model_Term *mterm;
	int *count_flag;
	signed char **freq_set;
	signed char *seg[2],*m_flag;
	int flag;
};

struct TraitLocus {
	int n_alleles; /* Currently always 2 */
	int link_group;
	double pos[2],variance;
	double **freq;
	double *eff;
	double dom_par;
	int flag;
	signed char *seg[2];
	int *gt;
};

struct Link {
	char *name;
	int n_markers;
	int type;
	int sample_pos;
	int *mk_index;
	int range_set[2];
	struct IBD_List *ibd_list;
	double r1[2],r2[2]; /* Map range */
};

struct Variable {
	char *name;
	int type;
	int n_levels,index,rec_flag;
	union arg_type *recode;
};

struct Model_Var {
	int type;
	int var_index;
};

struct Model_Term {
	int n_vars;
	int df;
	int out_flag;
	double *eff;
	struct Model_Var *vars;
};

struct Model {
	struct Model_Var var;
	int n_terms;
	struct Model_Term *term;
};

extern void sample_segs(void);
extern void ReadBinFiles(void);
extern int ReadParam(FILE *);
extern void AllocLokiStruct(void);
extern void AllocEffects(void);
extern void print_orig_id(FILE *,const int);
extern void print_orig_id1(FILE *,const int);
extern void print_orig_family(FILE *,const int,const int);
extern void print_orig_triple(FILE *,const int);
extern void print_orig_allele_id(FILE *,const int);
extern void LokiSetup(void);
extern void InitValues(void);
extern void SampleLoop(void);
extern void delete_traitlocus(const int);
extern void print_allele_type1(FILE *,const int,const int);
extern void print_allele_type(FILE *,const int,const int,const int);
extern void sample_all_segs(const int);
extern int get_new_traitlocus(const int);
extern void loki_identity(double *,int,int);
extern double phi2(int,int);
extern char *fget_line(FILE *);
extern char *Filter,*Seedfile,*Output_Phen,*Dumpfile,*Outputfile,*Freqfile,*Haplofile;
extern unsigned int RunID;
extern int ped_size,n_markers,n_links,n_id_records,n_nonid_records,polygenic_flag,n_tloci,sex_map,read_dump_flag;
extern int n_genetic_groups,n_comp,*comp_ngenes,*comp_start,*comp_size,*comp_npeel,max_tloci,min_tloci,start_tloci,extra_allele_flag;
extern int catch_sigs,sig_caught,use_student_t,output_haplo,family_id;

extern struct remember *RemBlock;
extern struct Id_Record *id_array;
extern struct Id_Recode id_recode,fam_recode;
extern struct Marker *marker;
extern struct Link *linkage;
extern struct Variable *id_variable,*nonid_variable;
extern struct Model model;
extern struct SparseMatRec **AIMatrix;
extern struct TraitLocus *tlocus;
extern struct output_gen *Output_Gen;
extern double mjrgene_var,residual_var,additive_var,residual_var_limit,additive_var_limit,grand_mean,total_maplength[2];
extern double tau,tau_beta,tloci_mean,lm_ratio,l_prob;
extern int res_var_set,add_var_set,grand_mean_set,num_iter,sample_from[2],sample_freq[2],tlocus_flag,tloci_mean_set;
extern int no_overdominant,ranseed_set,censored_flag,tau_mode,censor_mode,debug_level,sort_sex,l_cycles;
extern int peel_trace,dump_freq,output_type,append_output_flag,analysis;
extern int max_linked,max_linked_set;

struct move_stats {
	int success;
	int n;
};

extern struct move_stats move_stats[];

#define N_MOVE_STATS 8
#define NUM_SYSTEM_VAR 11

extern struct id_data syst_var[];

#define SYST_NO_OVERDOMINANT 0
#define SYST_TAU 1
#define SYST_TAU_MODE 2
#define SYST_CENSOR_MODE 3
#define SYST_DEBUG_LEVEL 4
#define SYST_LM_RATIO 5
#define SYST_L_PROB 6
#define SYST_L_CYCLES 7
#define SYST_PEEL_TRACE 8
#define SYST_BACKUPS 9
#define SYST_SI_MODE 10

#define DEFAULT_ANALYSIS 0
#define AFFECTED_ANALYSIS 1
#define NULL_ANALYSIS 2
#define IBD_ANALYSIS 4
#define IBS_ANALYSIS 8
#define ESTIMATE_IBD 16

#define OUTPUT_TYPE_ORIGINAL 0
#define OUTPUT_VERSION_2_1 1
#define OUTPUT_VERSION_2_2 2
#define OUTPUT_VERSION_2_3 3
#define DEFAULT_OUTPUT_TYPE OUTPUT_VERSION_2_3
