/****************************************************************************
 *                                                                          *
 *     Loki - Programs for genetic analysis of complex traits using MCMC    *
 *                                                                          *
 *             Simon Heath - Rockefeller University                         *
 *                                                                          *
 *                       November 1997                                      *
 *                                                                          *
 * loki_tlmoves.c:                                                          *
 *                                                                          *
 * Lots of the more weird update moves                                      *
 *                                                                          *
 ****************************************************************************/

#include <config.h>
#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#if HAVE_LIMITS_H
#include <limits.h>
#endif
#include <float.h>
#if HAVE_VALUES_H
#include <values.h>
#endif
#ifndef DBL_MAX
#define DBL_MAX MAXDOUBLE
#endif

#include "ranlib.h"
#include "utils.h"
#include "loki.h"
#include "loki_peel.h"
#include "loki_tlmoves.h"
#include "handle_res.h"
#include "sample_cens.h"

static int *perm;
static double *prior,*new_freq;
int sort_sex;
static double marker_move_prob=1.0; /* Do we attempt to move markers? */

static double safe_exp(double x)
{
	static double max_arg;
	static char f=0;
	
	if(!f) {
		max_arg=log(DBL_MAX);
		f=1;
	}
	if(x>max_arg) return DBL_MAX;
	if(x<-max_arg) return 0.0;
	return exp(x);
}

/* Comparison function for qsort(), used to sort loci into position
 * order within a linkage group */
int cmp_loci_pos(const void *s1,const void *s2)
{
	double x1,x2;
	int i;
	
	i=*((const int *)s1);
	if(i>=0) x1=marker[i].pos[sort_sex];
	else x1=tlocus[-1-i].pos[sort_sex];
	i=*((const int *)s2);
	if(i>=0) x2=marker[i].pos[sort_sex];
	else x2=tlocus[-1-i].pos[sort_sex];
	if(x1<x2) return -1;
	if(x1>x2) return 1;
	return 0;
}

/* Get list of loci (marker + trait) in linkage group.  Returns number found
 * in count */
void get_locuslist(int *perm,const int link,int *count)
{
	int i,j;

	for(i=0;i<linkage[link].n_markers;i++) perm[i]=linkage[link].mk_index[i];
	for(j=0;j<n_tloci;j++) if((tlocus[j].flag&TL_LINKED) && tlocus[j].link_group==link) perm[i++]= -(j+1);
	*count=i;
}

static void Adjust_Mean(const double adj)
{
	int i,j,type;

	type=model.var.type;
	if(type&ST_CONSTANT)	{
		for(i=0;i<ped_size;i++) if(id_array[i].res) id_array[i].res[0]+=adj;
	} else for(i=0;i<ped_size;i++) if(id_array[i].res) for(j=0;j<id_array[i].n_rec;j++) id_array[i].res[j]+=adj;
	grand_mean-=adj;
}

static void Adjust_for_TL(const int tl,const double z)
{
	int i,j,k,type;
	double *eff;
	int *gt;
	
	eff=tlocus[tl].eff;
	gt=tlocus[tl].gt;
	type=model.var.type;
	if(type&ST_CONSTANT)	{
		for(i=0;i<ped_size;i++) if(id_array[i].res) {
			k=gt[i]-1;
			if(k) id_array[i].res[0]+=z*eff[k-1];
		}
	} else for(i=0;i<ped_size;i++) if(id_array[i].res) {
		k=gt[i]-1;
		if(k) for(j=0;j<id_array[i].n_rec;j++) id_array[i].res[j]+=z*eff[k-1];
	}
}

static double interval_size(int j,int k,int link,int *pm)
{
	int k2,k1;
	double s=0.0,x;
	
	for(k2=0;k2<=sex_map;k2++)	{
		if(!j) {
			k1=pm[0];
			if(k1<0) x=tlocus[-1-k1].pos[k2];
			else x=marker[k1].pos[k2];
			s+=x-linkage[link].r1[k2];
		} else if(j==k) {
			k1=pm[k-1];
			if(k1<0) x=tlocus[-1-k1].pos[k2];
			else x=marker[k1].pos[k2];
			s+=linkage[link].r2[k2]-x;
		} else {
			k1=pm[j];
			if(k1<0) x=tlocus[-1-k1].pos[k2];
			else x=marker[k1].pos[k2];
			k1=pm[j-1];
			if(k1<0) x-=tlocus[-1-k1].pos[k2];
			else x-=marker[k1].pos[k2];
			s+=x;
		}
	}
	return s;
}

static double check_intervals(int j,int k,int link,int *pm,double *p,double move_p)
{
	int k1;
	double z,z1;
	
	z=(1.0-move_p)*.5;
	p[1]=move_p;
	/* Check if we can move to neighbouring intervals */
	p[0]=j?z:0.0;
	p[2]=j<k?z:0.0;
	/* Are there intervals outside the markers? */
	if(j==1) {
		k1=pm[0];
		if(k1<0) z=tlocus[-1-k1].pos[0];
		else z=marker[k1].pos[0];
		if(z<=linkage[link].r1[0]) p[0]=0.0;
	}
	if(j==(k-1)) {
		k1=pm[k];
		if(k1<0) z=tlocus[-1-k1].pos[0];
		else z=marker[k1].pos[0];
		if(z>=linkage[link].r2[0]) p[2]=0.0;
	}
	if(p[0]>0.0) p[0]*=interval_size(j-1,k,link,pm);
	p[1]*=(z1=interval_size(j,k,link,pm));
	if(p[2]>0.0) p[2]*=interval_size(j+1,k,link,pm);
	z=p[0]+p[1]+p[2];
	for(k1=0;k1<3;k1++) p[k1]/=z;
	return z1;
}

#ifdef FUNC_NAME
#undef FUNC_NAME
#endif
#define FUNC_NAME "Sample_LinkageGroup"
/* Sample genotypes for all marker loci in linkage group - trait loci are sampled during
 * movement step - see Sample_TL_Position() */
void Sample_LinkageGroup(const int link)
{
	int i,j,j1,k1,k2,k3,*perm1;
	double old_p[2],new_p[2],z,z1,z2;
	
	if(debug_level&4) printf("[S:%d]",link);
	fflush(stdout);
	k2=linkage[link].n_markers+n_tloci;
	perm1=perm+k2;
	get_locuslist(perm1,link,&k3);
	for(i=0;i<k3;i++) perm[i]=perm1[i];
	gnu_qsort(perm1,(size_t)k3,(size_t)sizeof(int),cmp_loci_pos);
	gen_perm(perm,k3);
	for(i=0;i<k3;i++) {
		k1=perm[i];
		for(j=0;j<k3;j++) if(perm1[j]==k1) break;
		if(k1>=0) {
			if(marker[k1].pos_set==2 && ranf()<marker_move_prob) {
				z=peel_locus(perm1,j,k3,0);
				if(z== -DBL_MAX) ABT_FUNC("Zero probability at start of step\n");
				for(k2=0;k2<2;k2++) old_p[k2]=marker[k1].pos[k2];
				/* Sample new position within interval */
				for(k2=0;k2<=sex_map;k2++)	{
					if(j) {
						j1=perm1[j-1];
						if(j1<0) z1=tlocus[-1-j1].pos[k2];
						else z1=marker[j1].pos[k2];
					} else z1=linkage[link].r1[k2];
					if(j+1<k3) {
						j1=perm1[j+1];
						if(j1<0) z2=tlocus[-1-k1].pos[k2];
						else z2=marker[j1].pos[k2];
					} else z2=linkage[link].r2[k2];
					new_p[k2]=z1+ranf()*(z2-z1);
				} 
				if(!sex_map) new_p[1]=new_p[0];
				for(k2=0;k2<2;k2++) marker[k1].pos[k2]=new_p[k2];
				z1=peel_locus(perm1,j,k3,0);
				if(z1== -DBL_MAX) z=-1.0;
				else z=exp(z1-z);
				if(ranf()>=z) for(k2=0;k2<2;k2++) marker[k1].pos[k2]=old_p[k2];
			}
			(void)peel_locus(perm1,j,k3,1);
		}
	}
	if(debug_level&4) {
		fputc('*',stdout);
		fflush(stdout);
	}
}

double calc_tl_like(const int tl,int *perm,const int sflag)
{
	int i,k,link;
	double l;
	
	link=tlocus[tl].link_group;
	if(link<0) {
		perm[0]= -(tl+1);
		i=0;
		k=1;
	} else {
		get_locuslist(perm,link,&k);
		gnu_qsort(perm,(size_t)k,(size_t)sizeof(int),cmp_loci_pos);
		for(i=0;i<k;i++) if(perm[i]== -(tl+1)) break;
	}
	l=peel_locus(perm,i,k,sflag);
	return l;
}

#ifdef FUNC_NAME
#undef FUNC_NAME
#endif
#define FUNC_NAME "TL_Alloc"
void TL_Alloc()
{
	int k;
	
	k=n_markers+max_tloci;
	if(!k) k++;
	if(!(perm=malloc(k*2*sizeof(int)))) ABT_FUNC(MMsg);
	if(!(new_freq=malloc(2*(n_genetic_groups+n_links+1)*sizeof(double)))) ABT_FUNC(MMsg);
	prior=new_freq+n_genetic_groups*2;
}

void TL_Free()
{
	if(perm) free(perm);
	if(new_freq) free(new_freq);
}

#ifdef FUNC_NAME
#undef FUNC_NAME
#endif
#define FUNC_NAME "Sample_TL_Position"
/* Sample a new position for trait locus.  Allows moves to be made between
 * linkage groups (including to and from unlinked).
 * BIGMOVE_PROB is the probability of proposing a large move (uniform over entire genome)
 * vs. a small move (current location +/- 1 interval)
 * If unlinked then a move to linked is always attempted.
 */
void Sample_TL_Position(const int tl)
{
	int i,j,j2,k,k1,k2,fg,link,oldlink,step;
	double l,l1,old_pos[2],map_length,z,z1,z2,r,*prob,ttm,p_i[3],p_i1[3],s1,s2;

	/* In the case of no linkage groups, or if this trait locus is not currently linkable, 
	 just sample genotypes for trait locus and return */
	if(!n_links || !(tlocus[tl].flag&TL_LINKABLE)) {
		(void)calc_tl_like(tl,perm,1);
		return;
	}
	prob=prior+n_links+1;
	/* Set up prior probs. for chromosomes (and unlinked region) */
	/* Note that we use the average of sex-specific maps */
	for(i=0;i<=n_links;i++) prior[i]=0.0;
	ttm=0.0;
	for(k=0;k<=sex_map;k++)	{
		z=total_maplength[k];
		ttm+=z;
		for(i=0;i<n_links;i++) {
			map_length=linkage[i].r2[k]-linkage[i].r1[k];
			z-=map_length;
			prior[i]+=map_length;
		}
		prior[i]+=z;
	}
	if(tlocus[tl].flag&TL_LINKED) {
		link=tlocus[tl].link_group;
		old_pos[0]=tlocus[tl].pos[0];
		old_pos[1]=tlocus[tl].pos[1];
	} else {
		link= -1;
		old_pos[0]=old_pos[1]=0.0;
	}
	l=calc_tl_like(tl,perm,0);
	if(l== -DBL_MAX) ABT_FUNC("Internal error - zero probability at start of step\n");
	oldlink=link;
	z1=1.0;
	/* If unlinked always attempt a 'big move', otherwise do so with probability BIGMOVE_PROB  */
	/* A 'big' move can be to anywhere in the genome (except that if unlinked can not jump to */
	/* another unlinked location) */
	if(link== -1 || (BIGMOVE_PROB>0.0 && ranf()<=BIGMOVE_PROB))	{
		step=5;
		z=0.0;
		for(i=0;i<=n_links;i++)	{
			z+=prior[i];
			prob[i]=z;
		}
		if(link== -1) { /* If already unlinked, prevent choosing unlinked again */
			z-=prior[n_links];
			prob[n_links]=z;
			j=n_links;
		} else j=n_links+1;
		if(z<=0.0) ABT_FUNC("Internal error - no possible move\n");
		do {
			r=z*ranf();
			for(i=0;i<j;i++) if(prior[i]>0.0 && r<=prob[i]) break;
		} while(i==j);
		link=i;
		if(link==n_links)	{
			link= -1;
			tlocus[tl].pos[0]=tlocus[tl].pos[1]=0.0;
		} else {
			if(link) r-=prob[link-1]; /* Get position within the chromosome we've landed on relative to left of linkage map */
			/* Get list of loci on linkage group (excluding current locus) */
			fg=tlocus[tl].flag;
			tlocus[tl].flag&=~TL_LINKED;
			get_locuslist(perm,link,&k);
			gnu_qsort(perm,(size_t)k,(size_t)sizeof(int),cmp_loci_pos);
			tlocus[tl].flag=fg;
			if(sex_map)	{ /* Find interval */
				for(j=0;j<k;j++) {
					k1=perm[j];
					if(k1<0)	{
						z=tlocus[-1-k1].pos[0]-linkage[link].r1[0];
						z+=tlocus[-1-k1].pos[1]-linkage[link].r1[1];
					} else {
						z=marker[k1].pos[0]-linkage[link].r1[0];
						z+=marker[k1].pos[1]-linkage[link].r1[1];
					}
					if(z>=r) break;
				}
				/* Sample male and female positions (independently) within chosen interval */
				for(k2=0;k2<2;k2++) {
					if(j<k) {
						k1=perm[j];
						if(k1<0) z=tlocus[-1-k1].pos[k2];
						else z=marker[k1].pos[k2];
					} else z=linkage[link].r2[k2];
					if(j)	{
						k1=perm[j-1];
						if(k1<0) z2=tlocus[-1-k1].pos[k2];
						else z2=marker[k1].pos[k2];
					} else z2=linkage[link].r1[k2];
					tlocus[tl].pos[k2]=z2+(z-z2)*ranf();
				}
			} else tlocus[tl].pos[0]=tlocus[tl].pos[1]=r+linkage[link].r1[0];
		}
		tlocus[tl].link_group=link;
		if(link!=oldlink)	{
			if(link<0 && oldlink>=0) z1=ttm/(BIGMOVE_PROB*(ttm-prior[n_links]));
			else if(link>=0 && oldlink<0) z1=BIGMOVE_PROB*(ttm-prior[n_links])/ttm;
		}
		if(link<0) {
			tlocus[tl].flag&=~TL_LINKED;
			tlocus[tl].flag|=TL_UNLINKED;
		} else {
			tlocus[tl].flag&=~TL_UNLINKED;
			tlocus[tl].flag|=TL_LINKED;
		}
	} else { /* 'Small' move - move at most 1 interval from current location */
		step=4;
		get_locuslist(perm,link,&k);
		gnu_qsort(perm,(size_t)k,(size_t)sizeof(int),cmp_loci_pos);
		for(j=0;j<k;j++) if(-1-perm[j]==tl) break;
		if(j==k) ABT_FUNC("Internal error - couldn't find trait locus in linkage group\n");
		for(j2=j+1;j2<k;j2++) perm[j2-1]=perm[j2];
		k--;
		s1=check_intervals(j,k,link,perm,p_i,SMALLMOVE_P);
		/* Pick an interval */
		do {
			z2=ranf();
			z=0.0;
			for(k2=0;k2<3;k2++) if(p_i[k2]>0.0)	{
				z+=p_i[k2];
				if(z2<=z) break;
			}
		} while(k2==3);
		/* Have we changed intervals? */
		if(k2!=1) {
			j2=j+k2-1;
			/* Compute probs. for new interval */
			s2=check_intervals(j2,k,link,perm,p_i1,SMALLMOVE_P);
			/* Compute proposal probability */
			z1=(s2/s1)*p_i1[2-k2]/p_i[k2];
		} else j2=j;
		/* Sample new positions */
		for(k2=0;k2<=sex_map;k2++)	{
			if(j2) {
				k1=perm[j2-1];
				if(k1<0) z=tlocus[-1-k1].pos[k2];
				else z=marker[k1].pos[k2];
			} else z=linkage[link].r1[k2];
			if(j2<k) {
				k1=perm[j2];
				if(k1<0) z2=tlocus[-1-k1].pos[k2];
				else z2=marker[k1].pos[k2];
			} else z2=linkage[link].r2[k2];
			tlocus[tl].pos[k2]=z+ranf()*(z2-z);
		} 
		if(!sex_map) tlocus[tl].pos[1]=tlocus[tl].pos[0];
	}
	if(debug_level&4) printf("[%d:%d]",step,tl);
	move_stats[step].n++;
	if(link<0) {
		l1=calc_tl_like(tl,perm,0);
		if(l1== -DBL_MAX) z= -1.0;
		else z=exp(l1-l)*z1;
	} else {
		l1=calc_tl_like(tl,perm,0);
		if(l1== -DBL_MAX) z= -1.0;
		else z=exp(l1-l)*z1;
	}
	i=0;
 	if(ranf()>=z) {
		i=1;
		tlocus[tl].pos[0]=old_pos[0];
		tlocus[tl].pos[1]=old_pos[1];
		link=oldlink;
		tlocus[tl].link_group=link;
		if(link<0) {
			tlocus[tl].flag&=~TL_LINKED;
			tlocus[tl].flag|=TL_UNLINKED;
		} else {
			tlocus[tl].flag&=~TL_UNLINKED;
			tlocus[tl].flag|=TL_LINKED;
		}
		if(debug_level&4) fputc('F',stdout);
	} else {	
		if(debug_level&4) fputc('S',stdout);
		move_stats[step].success++;
	}
	/* Always sample genotypes before returning */
	z=calc_tl_like(tl,perm,1);
	if(z== -DBL_MAX) ABT_FUNC("Internal error - zero probability at end of Move Step\n");
	if(debug_level&4) fflush(stdout);
}

#ifdef FUNC_NAME
#undef FUNC_NAME
#endif
#define FUNC_NAME "TL_Birth_Death"
/* Trait locus Birth/Death/Split/Combine step */
void TL_Birth_Death(void)
{
	int i,j,k,k1,k2,x= -1,x1= -1,step,nq,er=0;
	double l,l1,r,z,z1,z2,map_length,u1,u2,u3,jd=0.0,K=0.0,d,a,ua,ud;
	double va_prop,vd_prop,alpha= -1,mean_eff,mean_eff1,pp[4],eff[2],eff1[2],p,p1;
	double newpos[2],old_res_var,q1,(*res_fn)();
	double old_l,old_l1,old_res;
	int newlink,oldflag,cens_flag,mtype;
	static char *bk[3];
	
	/* Count how many QTL currently in model (nq) and no. unused QTL (j) */
	for(j=nq=x=0;x<n_tloci;x++) if(!tlocus[x].flag) j++; else nq++;
	j+=max_tloci-n_tloci;
	if(!(j || nq)) return;
	if(!bk[0]) {
		/* Allocate temporary storage for old genotypes/SI's for locus x */
		if(!(bk[0]=malloc(ped_size*3))) ABT_FUNC(MMsg);
		RemBlock=AddRemem(bk[0],RemBlock);
	}
	pp[0]=j?BIRTH_STEP:0.0; /* Birth step possible if there is space for new QTL */
	pp[1]=nq>min_tloci?DEATH_STEP:0.0; /* Death step possible if no. QTL not at minimum */
	pp[2]=(j&&nq)?SPLIT_STEP:0.0; /* Split step requires at least 1 QTL + space for new QTL */
	pp[3]=(nq>1 && nq>min_tloci)?COMBINE_STEP:0.0; /* Combine step requires >1 QTL and no. QTL not at minimum */
	mean_eff=mean_eff1=0.0;
	for(z=0.0,i=0;i<4;i++) z+=pp[i];
	/* We have a fixed no. loci - can't change anything */
	if(z<=0.0) return;
	do {
		r=ranf()*z;
		for(z1=0.0,step=0;step<4;step++) {
			if(pp[step]>0.0) {
				z1+=pp[step];
				if(r<=z1) break;
			}
		}
	} while(step==4);
	r=pp[step]/z; /* Proposal probability for step */
	if(!(step&1)) { /* Get position for new TL */
		for(i=0;i<=n_links;i++) prior[i]=0.0;
		z1=0.0;
		for(k=0;k<=sex_map;k++) {
			z=total_maplength[k];
			z1+=z;
			for(i=0;i<n_links;i++) {
				map_length=linkage[i].r2[k]-linkage[i].r1[k];
				z-=map_length;
				prior[i]+=map_length;
			}
		}
		prior[i]+=z;
		do {
			z2=z1*ranf();
			z=0.0;
			for(i=0;i<=n_links;i++) {
				if(prior[i]>0.0) {
					z+=prior[i];
					if(z2<=z) {
						z1=z2+prior[i]-z;
						break;
					}
				}
			}
		} while(i==n_links+1);
		newlink=(i==n_links?-1:i);
		if(newlink<0) newpos[0]=newpos[1]=0.0;
		else if(sex_map) {
			/* Get list of loci on linkage group */
			get_locuslist(perm,newlink,&k);
			gnu_qsort(perm,(size_t)k,(size_t)sizeof(int),cmp_loci_pos);
			/* Find interval */
			for(j=0;j<k;j++) {
				k1=perm[j];
				if(k1<0) {
					z=tlocus[-1-k1].pos[0]-linkage[newlink].r1[0];
					z+=tlocus[-1-k1].pos[1]-linkage[newlink].r1[1];
				} else {
					z=marker[k1].pos[0]-linkage[newlink].r1[0];
					z+=marker[k1].pos[1]-linkage[newlink].r1[1];
				}
				if(z>=z1) break;
			}
			/* Sample male and female positions (independently) within chosen interval */
			for(k2=0;k2<2;k2++) {
				if(j<k) {
					k1=perm[j];
					if(k1<0) z=tlocus[-1-k1].pos[k2];
					else z=marker[k1].pos[k2];
				} else z=linkage[newlink].r2[k2];
				if(j) {
					k1=perm[j-1];
					if(k1<0) z2=tlocus[-1-k1].pos[k2];
					else z2=marker[k1].pos[k2];
				} else z2=linkage[newlink].r1[k2];
				newpos[k2]=z2+(z-z2)*ranf();
			}
		} else newpos[0]=newpos[1]=z1+linkage[newlink].r1[0];
	} else {
		newlink=0;
		newpos[0]=newpos[1]=0.0;
	}
	va_prop=residual_var*PROP_RATIO;
	vd_prop=residual_var*PROP_RATIO;
	/* Find proposal distribution for next time if current step was to be accepted */
	if(step&1) { /* Death or Combine steps reduce QTL numbers by 1 */
		pp[0]=BIRTH_STEP;
		pp[1]=((nq-1)>min_tloci)?DEATH_STEP:0.0;
		pp[2]=(nq-1)?SPLIT_STEP:0.0;
		pp[3]=(nq>2 && (nq-1)>min_tloci)?COMBINE_STEP:0.0;
	} else { /* Birth or Split steps increase QTL numbers by 1 */
		pp[0]=(j-1)?BIRTH_STEP:0.0;
		pp[1]=DEATH_STEP;
		pp[2]=(j-1)?SPLIT_STEP:0.0;
		pp[3]=nq?COMBINE_STEP:0.0;
	}
	mtype=model.var.type;
	if(!censor_mode && (mtype&ST_CENSORED)) {
		cens_flag=1;
		res_fn=&Calc_CensResLike;
	} else {
		cens_flag=0;
		res_fn=&Calc_ResLike;
	}
	for(z=0.0,i=0;i<4;i++) z+=pp[i];
	if(debug_level&4) printf("[%d]",step);
	move_stats[step].n++;
	old_res=res_fn();
	if(old_res== -DBL_MAX) ABT_FUNC("Internal error - zero probability at start of Death Step\n");	
	old_res_var=residual_var;
	/* If censoring and integrating over censored values, the sampled censored
	 * values will be invalid here, and will need to be sampled */
	if(cens_flag) Sample_Censored();
	if(!step) { /* Birth Step */
		r=pp[1]/(z*r); /* Proposal ratio q(death)/q(birth) */
 		/* Factor in Poisson prior on number of loci */
		if(tloci_mean_set) r*=tloci_mean/(double)(1+nq);
		u1=genexp(va_prop);
		u2=genexp(vd_prop);
		for(u3=0.0,i=0;i<n_genetic_groups;i++) u3+=(new_freq[i]=ranf());
		u3/=(double)n_genetic_groups;
		d=sqrt(u2)/(2.0*u3*(1.0-u3));
		if(ranf()<0.5) d= -d;
		z=(ranf()<0.5)?-1.0:1.0;
		a=z*sqrt(u1/(2.0*u3*(1.0-u3)))-d*(1.0-2.0*u3);
		if(no_overdominant &&fabs(d)>fabs(a)) {
			if(debug_level&4) {
				fputc('f',stdout);
				fflush(stdout);
			}
			return;
		}
		q1=Calc_Resprop();
		/* Find next available unused QTL */
		x=get_new_traitlocus(2);
		tlocus[x].pos[0]=newpos[0];
		tlocus[x].pos[1]=newpos[1];
		tlocus[x].link_group=newlink;
		tlocus[x].flag=(newlink<0?TL_UNLINKED:TL_LINKED);
		tlocus[x].flag|=TL_LINKABLE;
		/* Get likelihood for QTL not in model */
		l=old_res;
		for(i=0;i<n_genetic_groups;i++) {
			tlocus[x].freq[i][0]=new_freq[i];
			tlocus[x].freq[i][1]=1.0-new_freq[i];
		}
		tlocus[x].eff[0]=a+d;
		tlocus[x].eff[1]=a*2.0;
		mean_eff=(1.0-u3)*(1.0-u3)*tlocus[x].eff[1]+2.0*u3*(1.0-u3)*tlocus[x].eff[0];
		Adjust_Mean(mean_eff);
		/* Calculate likelihood (and sample) for model with QTL (and mu = mu - mean_eff) */
		l1=calc_tl_like(x,perm,2);
		if(l1== -DBL_MAX) er=1;
		else {
			/* We'll need to sample more censored values conditional on the new QTL */
			if(cens_flag) Sample_Censored();
			z=Sample_ResVar();
			q1=q1-z+Calc_Res_Ratio(residual_var,old_res_var)+Calc_Var_Prior(residual_var)-Calc_Var_Prior(old_res_var);
			/* prior prob. for TL effects */
			z1=0.0;
			for(i=0;i<2;i++) z1+=tlocus[x].eff[i]*tlocus[x].eff[i];
			z=-.5*(2.0*log(2.0*M_PI*tau)+z1/tau);
			/* proposal prob. for u1,u2 */
			z1= -u1/va_prop-u2/vd_prop-log(va_prop*vd_prop);
			/* det(jacobian) for transformation from (u1,u2,u3)->(effect[0],effect[1],p) */
			z1+=.5*(log(2.0*u1*u2)+3.0*(log(u3)+log(1.0-u3)));
			/* Acceptance ratio */
			alpha=safe_exp(l1+z-l-z1+q1);
			if(debug_level&8) {printf("<%g,%g,%g,%g,%g>",l,l1,z,z1,q1);}
			if(alpha<DBL_MAX) alpha*=r;
		}
		if(!er) z=ranf();
		if(debug_level&8) {
			if(er) fputs("(er)",stdout);
			else printf("(%g,%g)",alpha,z);
		}
		if(!er && z<alpha)	{
			/* If successful, sample genotypes for new QTL */
			if(debug_level&4) fputc('S',stdout);
			move_stats[step].success++;
		} else {
			residual_var=old_res_var;
			Adjust_for_TL(x,1.0);
			delete_traitlocus(x);
			Adjust_Mean(-mean_eff);
			if(debug_level&1)	{
				l=res_fn();
				z=fabs(l-old_res);
				if(z>1.0e-8) {
					printf("Birth: %g %g %g %d\n",old_res,l,z,er);
					ABT_FUNC("aborting\n");
				}
			}
			if(debug_level&4) fputc('F',stdout);
		}
	} else if(step==1) { /* Death_Step */
		r=pp[0]/(z*r); /* Proposal ratio q(birth)/q(death) */
 		/* Factor in Poisson prior on number of loci */
		if(tloci_mean_set) r*=(double)nq/tloci_mean;
		/* Pick a non-blank QTL at random */
		i=(int)(safe_ranf()*(double)nq);
		for(x=0;x<n_tloci;x++) if(tlocus[x].flag) if(!(i--)) break;
		if(debug_level&1)	{
			old_l=calc_tl_like(x,perm,0);
			if(old_l== -DBL_MAX) ABT_FUNC("Internal error - zero probability at start of Death Step\n");
		} else old_l=0.0;
		q1=Calc_Resprop();
		for(u3=0.0,i=0;i<n_genetic_groups;i++) u3+=tlocus[x].freq[i][0];
		u3/=(double)n_genetic_groups;
		mean_eff=(1.0-u3)*(1.0-u3)*tlocus[x].eff[1]+2.0*u3*(1.0-u3)*tlocus[x].eff[0];
		/* Adjust residuals as if taking QTL out of model */
		Adjust_Mean(-mean_eff);
		Adjust_for_TL(x,1.0);
		/* Get new variance estimate */
		q1-=Sample_ResVar();
		/* Put means back... */
		Adjust_Mean(mean_eff);
		Adjust_for_TL(x,-1.0);
		/* And calculate likelihood with new variance estimate */
		l=calc_tl_like(x,perm,0);
		q1+=Calc_Res_Ratio(residual_var,old_res_var)+Calc_Var_Prior(residual_var)-Calc_Var_Prior(old_res_var);
		/* and take means out again! */
		Adjust_Mean(-mean_eff);
		Adjust_for_TL(x,1.0);
		a=tlocus[x].eff[1]*.5;
		d=tlocus[x].eff[0]-a;
		z=a+d*(1.0-2.0*u3);
		u1=2.0*u3*(1.0-u3)*z*z;
		z=2.0*u3*(1.0-u3)*d;
		u2=z*z;
		/* Calculate likelihood if QTL is out of model (and mean is adjusted) */
		l1=res_fn();
		if(l1== -DBL_MAX) alpha= -1.0;
		else {
			/* Prior for QTL effects */
			z1=0.0;
			for(i=0;i<2;i++) z1+=tlocus[x].eff[i]*tlocus[x].eff[i];
			z=-.5*(2.0*log(2.0*M_PI*tau)+z1/tau);
			/* proposal prob. for u1,u2 */
			z1= -u1/va_prop-u2/vd_prop-log(va_prop*vd_prop);
			/* det(jacobian) for transformation from (u1,u2,u3)->(effect[0],effect[1],p) */
			z1+=.5*(log(2.0*u1*u2)+3.0*(log(u3)+log(1.0-u3)));
			/* Acceptance ratio */
			alpha=safe_exp(l1-l-z+z1+q1);
			if(alpha<DBL_MAX) alpha*=r;
		}
		z=ranf();
		if(debug_level&8) printf("(%g,%g)",alpha,z);
		if(z<alpha) {
			delete_traitlocus(x);
			if(debug_level&4) fputc('S',stdout);
			move_stats[step].success++;
		} else {
			residual_var=old_res_var;
			Adjust_Mean(mean_eff);
			Adjust_for_TL(x,-1.0);
			if(debug_level&1)	{
				l=res_fn();
				z=fabs(l-old_res);
				if(z>1.0e-8) {
					printf("Death 1: %g %g %g %d\n",old_res,l,z,er);
					ABT_FUNC("aborting\n");
				}
				l=calc_tl_like(x,perm,0);
				z=fabs(l-old_l);
				if(z>1.0e-8) {
					printf("Death 2: %g %g %g %d\n",old_l,l,z,er);
					ABT_FUNC("aborting\n");
				}
			}
			if(debug_level&4) fputc('F',stdout);
		}
	} else if(step==2) { /* Split step */
		r=pp[3]/(z*r); /* Proposal ratio q(combine)/q(split) */
 		/* Factor in Poisson prior on number of loci */
		if(tloci_mean_set) r*=tloci_mean/(double)(nq+1);
		/* Pick a non-blank QTL to split at random */
		i=(int)(safe_ranf()*(double)nq);
		for(x=0;x<n_tloci;x++) if(tlocus[x].flag) if(!(i--)) break;
		bk[1]=bk[0]+ped_size;
		bk[2]=bk[1]+ped_size;
		/* Copy current genotypes/SI's for locus x */
		memcpy(bk[0],tlocus[x].seg[0],ped_size);
		memcpy(bk[1],tlocus[x].seg[1],ped_size);
		memcpy(bk[2],tlocus[x].gt,ped_size);
		/* Copy current QTL effect for locus x */
		for(i=0;i<2;i++) eff[i]=tlocus[x].eff[i];
		/* Calculate likelihood for model with current number of QTL and current effect for locus x */
		l=calc_tl_like(x,perm,0);
		if(l== -DBL_MAX) ABT_FUNC("Internal error - zero probability at start of Split Step\n");
		old_l=l;
		/* prior prob. for original QTL effects for locus x */
		z1=0.0;
		for(i=0;i<2;i++) z1+=eff[i]*eff[i];
		z=-.5*(2.0*log(2.0*M_PI*tau)+z1/tau);
		l+=z;
		/* Add current effect of QTL onto residuals */
		Adjust_for_TL(x,1.0);
		/* Estimate variance contribution from locus x (assuming H-W) */
		for(p=0.0,i=0;i<n_genetic_groups;i++) p+=tlocus[x].freq[i][0];
		p/=(double)n_genetic_groups;
		a=eff[1]*.5;
		d=eff[0]-a;
		z2=(d<0.0)?-1.0:1.0;
		z=a+d*(1.0-2.0*p);
		z1=(z<0.0)?-1.0:1.0;
		u1=2*p*(1.0-p)*z*z;
		z=2.0*p*(1.0-p)*d;
		u2=z*z;
		/* Calculate new effect for locus x */
		ua=genbet(BETA_A,BETA_B);
		ud=genbet(BETA_A,BETA_B);
		d=z2*sqrt(u2*ud)/(2.0*p*(1.0-p));
		a=z1*sqrt(u1*ua/(2.0*p*(1.0-p)))-d*(1.0-2.0*p);
		tlocus[x].eff[0]=d+a;
		tlocus[x].eff[1]=a*2.0;
		if(no_overdominant && fabs(d)>fabs(a)) er=1;
		if(!er) {
			/* Estimate change in mean caused by changing effect of locus x*/
			mean_eff=(1.0-p)*(1.0-p)*(tlocus[x].eff[1]-eff[1])+2.0*p*(1.0-p)*(tlocus[x].eff[0]-eff[0]);
			Adjust_Mean(mean_eff);
			/* Sample new genotypes for locus x */
			tlocus[x].flag&=~MARKER_SAMPLED;
			l1=calc_tl_like(x,perm,2);
			if(l1== -DBL_MAX) er=2;
		} else l1=0.0;
		if(!er) {
			/* prior prob. for new QTL effects for locus x */
			z1=0.0;
			for(i=0;i<2;i++) z1+=tlocus[x].eff[i]*tlocus[x].eff[i];
			z=-.5*(2.0*log(2.0*M_PI*tau)+z1/tau);
			l1+=z;
			/* Likelihood ratio of new effect to original effect for locus x */
			K=l1-l;
			/* Find next available unused QTL */
			x1=get_new_traitlocus(2);
			tlocus[x1].pos[0]=newpos[0];
			tlocus[x1].pos[1]=newpos[1];
			tlocus[x1].link_group=newlink;
			tlocus[x1].flag=(newlink<0?TL_UNLINKED:TL_LINKED);
			/* Calculate effect for locus x1 */
			for(p1=0.0,i=0;i<n_genetic_groups;i++) p1+=(new_freq[i]=ranf());
			p1/=(double)n_genetic_groups;
			d=sqrt(u2*(1.0-ud))/(2.0*p1*(1.0-p1));
			if(ranf()<0.5) d= -d;
			z=(ranf()<0.5)?-1.0:1.0;
			a=z*sqrt(u1*(1.0-ua)/(2.0*p1*(1.0-p1)))-d*(1.0-2.0*p1);
			tlocus[x1].eff[0]=d+a;
			tlocus[x1].eff[1]=a*2.0;
			for(i=0;i<n_genetic_groups;i++) {
				tlocus[x1].freq[i][0]=new_freq[i];
				tlocus[x1].freq[i][1]=1.0-new_freq[i];
			}
			if(no_overdominant && fabs(d)>fabs(a)) er=3;
			/* Likelihood for model with nq QTL with new genotypes for locus x */
			if(!er) {
				l=res_fn();
				if(l== -DBL_MAX) er=3;
			}
			/* Estimate change in mean caused by changing effect of locus x1 */
			mean_eff1=(1.0-p1)*(1.0-p1)*tlocus[x1].eff[1]+2.0*p1*(1.0-p1)*tlocus[x1].eff[0];
			Adjust_Mean(mean_eff1);
			/* Likelihood for model with nq+1 QTL */
			if(!er) {
				l1=calc_tl_like(x1,perm,0);
				if(l1== -DBL_MAX) er=4;
			}
			if(!er) {
				/* prior prob. for QTL effects for locus x1 */
				z1=0.0;
				for(i=0;i<2;i++) z1+=tlocus[x1].eff[i]*tlocus[x1].eff[i];
				z=-.5*(2.0*log(2.0*M_PI*tau)+z1/tau);
				l1+=z;
				K+=l1-l;
				/* proposal prob. for ua,ud,p1 */
				z=lgamma(BETA_A+BETA_B)-lgamma(BETA_A)-lgamma(BETA_B)+log(ua)*(BETA_A-1.0)+log(1.0-ua)*(BETA_B-1.0);
				z+=lgamma(BETA_A+BETA_B)-lgamma(BETA_A)-lgamma(BETA_B)+log(ud)*(BETA_A-1.0)+log(1.0-ud)*(BETA_B-1.0);
				K-=z;
				a=eff[1]*.5;
				d=eff[0]-a;
				jd=(1.0/4.0)*(a+d-2.0*d*p)*d*sqrt((1.0-p)*(1.0-p)*(1.0-p)*p*p*p/
															 ((1.0-p1)*(1.0-p1)*(1.0-p1)*p1*p1*p1*ua*ud*(1.0-ua)*(1.0-ud)));
				if(jd<0.0) jd= -jd;
			}
		}
		if(!er) {
			/* Acceptance ratio */
			alpha=safe_exp(K);
			if(alpha<DBL_MAX) alpha*=jd*r/.25;
		}
		if(!er && ranf()<alpha)	{
			l=calc_tl_like(x1,perm,2);
			if(l== -DBL_MAX) ABT_FUNC("Internal error - zero probability after successful Split Step\n");
			if(debug_level&4) fputc('S',stdout);
			move_stats[step].success++;
		} else {
			if(!er || er>2) {
				delete_traitlocus(x1);
				Adjust_for_TL(x,1.0);
			}
			if(er==2) Adjust_Mean(-mean_eff);
			else if(er!=1) Adjust_Mean(-(mean_eff+mean_eff1));
			for(i=0;i<2;i++) tlocus[x].eff[i]=eff[i];
			memcpy(tlocus[x].seg[0],bk[0],ped_size);
			memcpy(tlocus[x].seg[1],bk[1],ped_size);
			memcpy(tlocus[x].gt,bk[2],ped_size);
			tlocus[x].flag|=MARKER_SAMPLED;
			if(er==2) Recalc_Res();
			else Adjust_for_TL(x,-1.0);
			if(debug_level&1)	{
				l=res_fn();
				z=fabs(l-old_res);
				l1=calc_tl_like(x,perm,0);
				z1=fabs(l1-old_l);
				if(z>1.0e-8 || z1>1.0e-8) {
					printf("Split 1: %g %g %g %d\n",old_res,l,z,er);
					printf("Split 2: %g %g %g %d\n",old_l,l1,z1,er);
					ABT_FUNC("aborting\n");
				}
			}
			if(debug_level&4) fputc('F',stdout);
		}
	} else if(step==3) { /* Combine step */
		r=pp[2]/(z*r); /* Proposal ratio q(split)/q(combine) */
 		/* Factor in Poisson prior on number of loci */
		if(tloci_mean_set) r*=(double)nq/tloci_mean;
		/* Pick two non-blank QTL to combine at random */
		i=(int)(safe_ranf()*(double)nq);
		do j=(int)(safe_ranf()*(double)nq); while(i==j);
		for(x=0;x<n_tloci;x++) if(tlocus[x].flag) if(!(i--)) break;
		for(x1=0;x1<n_tloci;x1++) if(tlocus[x1].flag) if(!(j--)) break;
		/* Copy current QTL effect for locus x */
		for(i=0;i<2;i++) eff[i]=tlocus[x].eff[i];
		/* Estimate variances from both loci */
		for(p=0.0,i=0;i<n_genetic_groups;i++) p+=tlocus[x].freq[i][0];
		p/=(double)n_genetic_groups;
		a=eff[1]*.5;
		d=eff[0]-a;
		z2=(d<0.0)?-1.0:1.0;
		z=a+d*(1.0-2.0*p);
		z1=(z<0.0)?-1.0:1.0;
		u1=2*p*(1.0-p)*z*z;
		z=2.0*p*(1.0-p)*d;
		u2=z*z;
		for(p1=0.0,i=0;i<n_genetic_groups;i++) p1+=tlocus[x1].freq[i][0];
		p1/=(double)n_genetic_groups;
		a=tlocus[x1].eff[1]*.5;
		d=tlocus[x1].eff[0]-a;
		z=a+d*(1.0-2.0*p1);
		u3=2*p1*(1.0-p1)*z*z;
		ua=u1/(u1+u3);
		u1+=u3;
		z=2.0*p1*(1.0-p1)*d;
		u3=z*z;
		ud=u2/(u2+u3);
		u2+=u3;
		/* Calculate new effect for locus x (combining variance from both loci) */
		d=z2*sqrt(u2)/(2.0*p*(1.0-p));
		a=z1*sqrt(u1/(2.0*p*(1.0-p)))-d*(1.0-2.0*p);
		eff1[0]=d+a;
		eff1[1]=a*2.0;
		if(no_overdominant && fabs(d)>fabs(a)) er=1;
		l=calc_tl_like(x1,perm,0);
		if(l== -DBL_MAX) ABT_FUNC("Internal error - zero probability at start of Combine Step\n");
		oldflag=tlocus[x1].flag;
		old_l1=l;
		if(debug_level&1)	{
			old_l=calc_tl_like(x,perm,0);
		} else old_l=0.0;
		mean_eff=(1.0-p)*(1.0-p)*(eff1[1]-eff[1])+2.0*p*(1.0-p)*(eff1[0]-eff[0]);
		mean_eff1=(1.0-p1)*(1.0-p1)*tlocus[x1].eff[1]+2.0*p1*(1.0-p1)*tlocus[x1].eff[0];
		/* Add current effect of locus x1 on to residuals */
		Adjust_Mean(-mean_eff1);
		Adjust_for_TL(x1,1.0);
		/* Calculate likelihood if QTL x1 is out of model (and mean is adjusted) */
		K=res_fn();
		tlocus[x1].flag=0;
		l1=0.0;
		if(K== -DBL_MAX) er=1;
		else {
			/* Add on prior for current QTL effects at x and x1 */
			z1=0.0;
			for(i=0;i<2;i++) {
				z1+=eff[i]*eff[i];
				z1+=tlocus[x1].eff[i]*tlocus[x1].eff[i];
			}
			z=-.5*(4.0*log(2.0*M_PI*tau)+z1/tau);
			K-=l+z;
			/* Likelihood under model without x1 of locus x with current effect (mean adj.) */
			l1=calc_tl_like(x,perm,0);
			if(l1== -DBL_MAX) er=1;
		}
		if(!er) {
			K-=l1;
			Adjust_Mean(mean_eff);
			Adjust_for_TL(x,1.0);
			for(i=0;i<2;i++) tlocus[x].eff[i]=eff1[i];
			/* Add on prior for new QTL effect at x */
			z1=0.0;
			for(i=0;i<2;i++) z1+=eff1[i]*eff1[i];
			K+= -.5*(2.0*log(2.0*M_PI*tau)+z1/tau);
			/* Likelihood under model without x1 of locus x with new effect (mean adj.) */
			tlocus[x].flag&=~MARKER_SAMPLED;
			l1=calc_tl_like(x,perm,0);
			if(l1== -DBL_MAX) er=2;
			else {
				K+=l1;
				/* proposal prob. for ua,ub,p1 */
				z=lgamma(BETA_A+BETA_B)-lgamma(BETA_A)-lgamma(BETA_B)+log(ua)*(BETA_A-1.0)+log(1.0-ua)*(BETA_B-1.0);
				z+=lgamma(BETA_A+BETA_B)-lgamma(BETA_A)-lgamma(BETA_B)+log(ud)*(BETA_A-1.0)+log(1.0-ud)*(BETA_B-1.0);
				K+=z;
				/* det(jacobian) for transformation from (loci[x].effect,ua,ub,m,p(x),p(x1),mu)->(eff(x)',eff(x1)',p(x),p(x1),mu') */
				a=eff1[1]*.5;
				d=eff1[0]-a;
				jd=(1.0/4.0)*(a+d-2.0*d*p)*d*
				  sqrt((1.0-p)*(1.0-p)*(1.0-p)*p*p*p/((1.0-p1)*(1.0-p1)*(1.0-p1)*p1*p1*p1*
																  ua*ud*(1.0-ua)*(1.0-ud)));
				if(jd<0.0) jd= -jd;
			}
		}
		if(!er) {
			/* Acceptance ratio */
			alpha=safe_exp(K);
			if(alpha<DBL_MAX) alpha*=.25*r/jd;
		}
		if(!er && ranf()<alpha) {
			delete_traitlocus(x1);
			l=calc_tl_like(x,perm,2);
			if(l== -DBL_MAX) ABT_FUNC("Internal error - zero probability after successful Split Step\n");
			if(debug_level&4) fputc('S',stdout);
			move_stats[step].success++;
		} else {
			tlocus[x1].flag=oldflag;
			for(i=0;i<2;i++) tlocus[x].eff[i]=eff[i];
			if(er!=1) {
				Adjust_for_TL(x,-1.0);
				tlocus[x].flag|=MARKER_SAMPLED;
			}
			Adjust_for_TL(x1,-1.0);
			if(er==1) Adjust_Mean(mean_eff1);
			else Adjust_Mean(mean_eff1-mean_eff);
			if(debug_level&1)	{
				l=res_fn();
				z=fabs(l-old_res);
				if(z>1.0e-8) {
					printf("Combine 1: %g %g %g %d\n",old_res,l,z,er);
					ABT_FUNC("aborting\n");
				}
				l=calc_tl_like(x,perm,0);
				z=fabs(l-old_l);
				if(z>1.0e-8) {
					printf("Combine 2: %g %g %g %d\n",old_l,l,z,er);
					ABT_FUNC("aborting\n");
				}
				l=calc_tl_like(x1,perm,0);
				z=fabs(l-old_l1);
				if(z>1.0e-8) {
					printf("Combine 3: %g %g %g %d\n",old_l1,l,z,er);
					ABT_FUNC("aborting\n");
				}
			}
			if(debug_level&4) fputc('F',stdout);
		}
	}
	if(debug_level&4) fflush(stdout);
}
