/****************************************************************************
 *                                                                          *
 *     Loki - Programs for genetic analysis of complex traits using MCMC    *
 *                                                                          *
 *             Simon Heath - University of Washington                       *
 *                                                                          *
 *                       March 1997                                         *
 *                                                                          *
 * prep.c:                                                                  *
 *                                                                          *
 * Data preperation - reading control file, reading datafiles, recoding     *
 * factorial data, and writing out binary datafiles for loki                *
 *                                                                          *
 ****************************************************************************/

#include <config.h>

#include <stdio.h>
#include <stdlib.h>
#if HAVE_LIMITS_H
#include <limits.h>
#endif
#include <signal.h>

#include "version.h"
#include "ranlib.h"
#include "utils.h"
#include "scan.h"
#include "control_parse.h"

unsigned int RunID;
int nrm_flag,catch_sigs,sig_caught;
struct remember *RemBlock;
static struct remember *FirstRemBlock;
char *file_prefix="loki",*file_dir;
time_t start_time;
double extra_time,extra_stime,extra_utime;

void free_infile(struct InFile *infile)
{
	struct DataBlock *db,*db1;
	
	db=infile->data;
	while(db) {
		if(db->records) free(db->records);
		if(db->type) free(db->type);
		db1=db->next;
		free(db);
		db=db1;
	}
	if(infile->format) {
		free(infile->format->f_atoms);
		free(infile->format);
	}
	if(infile->fformat) {
		if(infile->fformat->rs) free(infile->fformat->rs);
		if(infile->fformat->fs) free(infile->fformat->fs);
		free(infile->fformat);
	}
	free(infile->name);
	if(infile->element) free(infile->element);
	free(infile);
}

static void free_miss(struct Miss *m)
{
	if(m->Missing) free(m->Missing);
	if(m->element) free(m->element);
	free(m);
}

static void free_link(struct Link *l)
{
	if(l->name) free(l->name);
	if(l->element) free(l->element);
	free(l);
}

static void free_model(struct model *m)
{
	struct model_list *ml,*ml1;
	
	ml=m->model_list;
	while(ml) {
		ml1=ml->next;
		if(ml->element) free(ml->element);
		free(ml);
		ml=ml1;
	}
	free(m);
}

void free_op(struct operation *op)
{
	struct operation *op1;
	while(op) {
		if(op->type==STRING) free(op->arg.string);
		op1=op->next;
		free(op);
		op=op1;
	}
}

void free_restrict(struct Restrict *res)
{
	if(res->element) free(res->element);
	free_op(res->Op_List);
	free(res);
}

/*   This is called when the program finishes without calling abt(),
 *   i.e., normal termination or SIGINT etc.
 */
static void FreeStuff(void)
{
	int i;
	struct InFile *infile,*infile1;
	struct Miss *m,*m1;
	struct Link *l,*l1;
	struct model *model,*model1;
	struct Restrict *res,*res1;
	struct Censor *cen,*cen1;
	
	if(id_array) {
		for(i=0;i<ped_size;i++) if(id_array[i].haplo[0]) free(id_array[i].haplo[0]);
		free(id_array[0].mkflag);
		free(id_array);
	}
	if(id_elements) free(id_elements);
	if(ped_recode) free(ped_recode);
	if(family_recode) free(family_recode);
	if(ped_recode1) free(ped_recode1);
	if(rec_tab) free(rec_tab);
	if(rec_tab1) free(rec_tab1);
	if(rsformat) free(rsformat);
	if(fsformat) free(fsformat);
	if(factor_recode) {
		for(i=0;i<(n_factors+n_markers);i++) if(factor_recode[i]) free(factor_recode[i]);
		free(factor_recode);
	}
	if(var_factors) free(var_factors);
	if(comp_size) free(comp_size);
	infile=Infiles;
	while(infile) {
		infile1=infile->next;
		free_infile(infile);
		infile=infile1;
	}
	m=Miss;
	while(m)	{
		m1=m->next;
		free_miss(m);
		m=m1;
	}
	l=links;
	while(l)	{
		l1=l->next;
		free_link(l);
		l=l1;
	}
	model=Models;
	while(model) {
		model1=model->next;
		free_model(model);
		model=model1;
	}
	res=Restrictions;
	while(res) {
		res1=res->next;
		free_restrict(res);
		res=res1;
	}
	cen=Censored;
	while(cen) {
		cen1=cen->next;
		free_op(cen->Op_List);
		free(cen);
		cen=cen1;
	}
	if(Affected) free_op(Affected);
	if(markers) {
		for(i=0;i<n_markers;i++) {
			if(markers[i].order)	{
				if(markers[i].order[0]) free(markers[i].order[0]);
				free(markers[i].order);
			}
			if(markers[i].allele_trans) {
				if(markers[i].allele_trans[0]) free(markers[i].allele_trans[0]);
				free(markers[i].allele_trans);
			}
			if(markers[i].o_size) free(markers[i].o_size);
		}
		free(markers);
	}
	if(traitlocus) {
		if(traitlocus->order) {	
			if(traitlocus->order[0]) free(traitlocus->order[0]);
			free(traitlocus->order);
		}
		if(traitlocus->o_size) free(traitlocus->o_size);
		free(traitlocus);
	}
	if(root_var) free_var(root_var);
	if(Filter) free(Filter);
	if(LogFile) free(LogFile);
	if(OutputFile) free(OutputFile);
	FreeRemem(FirstRemBlock);
}

/* Called when program receives signal */
static void int_handler(int i)
{
	char *sigs[]={"SIGTERM","SIGQUIT","SIGHUP","SIGINT","UNKNOWN"};
	int j,signals[]={SIGTERM,SIGQUIT,SIGHUP,SIGINT};
	
	for(j=0;j<4;j++) if(i==signals[j]) break;
	if(catch_sigs)	{
		sig_caught=i;
		return;
	}
	fprintf(stderr,"Exiting on signal %s (%d)\n",sigs[j],i);
	exit(i);
}

#ifdef FUNC_NAME
#undef FUNC_NAME
#endif
#define FUNC_NAME "main"
int main(int argc,char *argv[])
{
	FILE *fptr;
	int err;
	int ReadControl(FILE *,char *);
	struct sigaction s_action;
	
	start_time=time(0);
	if(!(FirstRemBlock=malloc(sizeof(struct remember)))) ABT_FUNC(MMsg);
	RemBlock=FirstRemBlock;
	RemBlock->pos=0;
	RemBlock->next=0;
	/* If program terminates normally or via a signal I want it to go through
	 * FreeStuff() so that Memory can be cleared so I can check for unfreed blocks
	 */
	atexit(FreeStuff);
	atexit(print_end_time);
	s_action.sa_handler=int_handler;
	s_action.sa_flags=0;
	sigaction(SIGINT,&s_action,0L);
	sigaction(SIGHUP,&s_action,0L);
	sigaction(SIGQUIT,&s_action,0L);
	sigaction(SIGTERM,&s_action,0L);
	if(argc<2) abt(__FILE__,__LINE__,"No control file specified\n");
	if((fptr=fopen(argv[1],"r"))) err=ReadControl(fptr,argv[1]);
	else {
		printf("Couldn't open '%s' for input as control file\nAborting...\n",argv[1]);
		exit(EXIT_FAILURE);
	}
	fclose(fptr);
	if(err) {
		LogFile=0;
		exit(EXIT_FAILURE);
	}
	print_start_time(PREP_NAME,"w");
	if(getseed("seedfile")) init_ranf(135421);
	RunID=(unsigned int)(ranf()*(double)0xffffffff);
	if(!scan_error_n) ReadData();          /* Read in the datafile(s) and recode (where necessary) */
	if(!pruned_ped_size) {
		printf("Zero size pedigree\nAborting...\n");
		exit(EXIT_FAILURE);
	}
	if(!scan_error_n)	{
		InitFamilies();
		count_loops();
		check_inbreeding();
		check_ymark();
	}
	if(!scan_error_n && (traitlocus || n_markers)) err=Genotype_Elimination(syst_var[CORRECT_ERRORS]);
	if(!scan_error_n && !err) nrm_flag=Calculate_NRM(); 
	writeseed("seedfile",1);
	if(family) free(family);
	if(!scan_error_n && !err && nrm_flag>=0) {
		WriteData();
		WriteReport();
	}
	free_nodes();
	if(scan_error_n) fprintf(stderr,"Errors: %d  ",scan_error_n);
	if(scan_warn_n) fprintf(stderr,"Warnings: %d  ",scan_warn_n);
	if(scan_error_n || scan_warn_n) fprintf(stderr,"\n");
	return sig_caught;
}
