/****************************************************************************
 *                                                                          *
 *     Loki - Programs for genetic analysis of complex traits using MCMC    *
 *                                                                          *
 *             Simon Heath - University of Washington                       *
 *                                                                          *
 *                       March 1997                                         *
 *                                                                          *
 * read_data.c:                                                             *
 *                                                                          *
 * Reads in data from all input files and recodes factorial data.           *
 * Also calls setup_pedigree() <setup_ped.c> to initialize pedigree         *
 * data structures.                                                         *
 *                                                                          *
 ****************************************************************************/

#include <config.h>

#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include <errno.h>
#include <string.h>
#include <sys/wait.h>
#if HAVE_REGCOMP
#include <regex.h>
#endif
#include "utils.h"
#include "control_parse.h"
#include "scan.h"

#define INIT_BLOCK_SIZE 128 /* Start allocating memory in blocks of INIT_BLOCK_SIZE records, doubling */
#define MAX_BLOCK_SIZE 512  /* the size if more space required up to MAX_BLOCK_SIZE */
#define S_BLOCK_SIZE 4096   /* For string allocation: should be at least as big as BUFFER_SIZE */
#define BUFFER_SIZE 4096    /* Maximum size of columns for free format reads */
#define LINE_COUNT 5000     /* How often to print 'At line' */

static char *StringData=0;
static char *default_rsformat="\n";
static size_t StringPos;
static struct miss_var_tag *miss_var;
static struct DataBlock *DataBlock;
static struct label_node *node_strings,*node_ints,*node_ped_int,*node_ped_str;
static int num_nodes;
struct label_node **ped_recode,**family_recode,***factor_recode;
int *ped_recode1;
int ped_size;

struct miss_var_tag
{
	int nmiss;
	char **Missing;
};

#ifdef FUNC_NAME
#undef FUNC_NAME
#endif
#define FUNC_NAME "process_missing"
static void process_missing(struct InFile *infile) {
	struct Miss *ms;
	int i,j,col=0,total_miss=0;
	char **p;
	
	if(!(miss_var=malloc(sizeof(struct miss_var_tag)*infile->ncol))) ABT_FUNC(MMsg);
	for(i=0;i<infile->ncol;i++) miss_var[i].nmiss=0;
	ms=Miss;
	while(ms) {
		for(col=j=0;j<infile->nvar;j++) if(infile->element[j]) {
			if(ms->element) {
				for(i=0;i<ms->nvar;i++) if(ms->element[i]==infile->element[j]) {
					miss_var[col].nmiss++;
					total_miss++;
				}
			} else {
				miss_var[col].nmiss++;
				total_miss++;
			}
			col++;
		}
		ms=ms->next;
	}
	if(total_miss) {
		if(!(p=malloc(sizeof(void *)*total_miss))) ABT_FUNC(MMsg);
	} else p=0;
	for(i=0;i<col;i++) {
		if(miss_var[i].nmiss) {
			miss_var[i].Missing=p;
			p+=miss_var[i].nmiss;
			miss_var[i].nmiss=0;
		} else miss_var[i].Missing=0;
	}
	ms=Miss;
	while(ms) {
		for(col=j=0;j<infile->nvar;j++) if(infile->element[j]) {
			if(ms->element) {
				for(i=0;i<ms->nvar;i++) if(ms->element[i]==infile->element[j])
				  miss_var[col].Missing[miss_var[col].nmiss++]=ms->Missing;
			} else miss_var[col].Missing[miss_var[col].nmiss++]=ms->Missing;
			col++;
		}
		ms=ms->next;
	}
}

#ifdef FUNC_NAME
#undef FUNC_NAME
#endif
#define FUNC_NAME "alloc_node"
static struct label_node *alloc_node(const void *s,int type)
{
	struct label_node *node;
	size_t i;
	
	if(!(node=malloc(sizeof(struct label_node)))) ABT_FUNC(MMsg);
	node->left=node->right=(struct label_node *)0;
	node->type=type;
	node->index=num_nodes++;
	node->balance=0;
	switch(type) {
	 case STRING:
		i=strlen((char *)s)+1;
		if(i>(S_BLOCK_SIZE-StringPos)) {
			if(!(StringData=malloc(S_BLOCK_SIZE))) ABT_FUNC(MMsg);
			RemBlock=AddRemem(StringData,RemBlock);
			StringPos=0;
		}
		strcpy(StringData+StringPos,(char *)s);
		node->data.string=StringData+StringPos;
		StringPos+=i;
		break;
	 case INTEGER:
		node->data.value= *(long *)s;
		break;
	 default:
		ABT_FUNC("Internal error - invalid type\n");
	}
	return node;
}

#ifdef FUNC_NAME
#undef FUNC_NAME
#endif
#define FUNC_NAME "find_inode"
static struct label_node *find_inode(struct label_node *node,const long s)
{
	long i;
	
	i=s-node->data.value;
	if(i<0) {
		if(node->left) node=find_inode(node->left,s);
		else node=0;
	} else if(i>0) {
		if(node->right) node=find_inode(node->right,s);
		else node=0;
	}
	return node;
}

#ifdef FUNC_NAME
#undef FUNC_NAME
#endif
#define FUNC_NAME "find_snode"
static struct label_node *find_snode(struct label_node *node,const char *s)
{
	long i;
	
	i=strcasecmp(s,node->data.string);
	if(i<0) {
		if(node->left) node=find_snode(node->left,s);
		else node=0;
	} else if(i>0) {
		if(node->right) node=find_snode(node->right,s);
		else node=0;
	}
	return node;
}

static struct label_node *rotate_left(struct label_node *node)
{
	struct label_node *l,*gc;

	l=node->left;
	if(l->balance==-1) {
		node->left=l->right;
		l->right=node;
		node->balance=l->balance=0;
		node=l;
	} else {
		gc=l->right;
		l->right=gc->left;
		gc->left=l;
		node->left=gc->right;
		gc->right=node;
		switch(gc->balance) {
		 case -1:
			node->balance=1;
			l->balance=0;
			break;
		 case 0:
			node->balance=l->balance=0;
			break;
		 case 1:
			node->balance=0;
			l->balance=-1;
		}
		gc->balance=0;
		node=gc;
	}
	return node;
}

static struct label_node *rotate_right(struct label_node *node)
{
	struct label_node *r,*gc;

	r=node->right;
	if(r->balance==1) {
		node->right=r->left;
		r->left=node;
		node->balance=r->balance=0;
		node=r;
	} else {
		gc=r->left;
		r->left=gc->right;
		gc->right=r;
		node->right=gc->left;
		gc->left=node;
		switch(gc->balance) {
		 case -1:
			node->balance=0;
			r->balance=1;
			break;
		 case 0:
			node->balance=r->balance=0;
			break;
		 case 1:
			node->balance=-1;
			r->balance=0;
		}
		gc->balance=0;
		node=gc;
	}
	return node;
}

#ifdef FUNC_NAME
#undef FUNC_NAME
#endif
#define FUNC_NAME "insert_node"
static struct label_node *insert_node(struct label_node *node,const void *s,struct label_node **node1,int *balanced,int type)
{
	long i;
	
	if(type==INTEGER) i=*((long *)s)-node->data.value;
	else 	i=strcasecmp((char *)s,node->data.string);
	if(i<0) {
		if(node->left) node->left=insert_node(node->left,s,node1,balanced,type);
		else {
			*node1=node->left=alloc_node(s,type);
			*balanced=0;
		}
		if(!(*balanced)) {
			switch(node->balance) {
			 case -1:
				node=rotate_left(node);
				*balanced=1;
				break;
			 case 0:
				node->balance=-1;
				break;
			 case 1:
				node->balance=0;
				*balanced=1;
			}
		}
	} else if(i>0) {
		if(node->right) node->right=insert_node(node->right,s,node1,balanced,type);
		else {
			*node1=node->right=alloc_node(s,type);
			*balanced=0;
		}
		if(!(*balanced)) {
			switch(node->balance) {
			 case -1:
				node->balance=0;
				*balanced=1;
				break;
			 case 0:
				node->balance=1;
				break;
			 case 1:
				node=rotate_right(node);
				*balanced=1;
			}
		}
	} else {
		*node1=node;
		*balanced=1;
	}
	return node;
}

struct label_node *find_node(const void *s,int type,int flag)
{
	struct label_node *node;

	if(type==INTEGER) {
		if(flag) node=node_ints;
		else node=node_ped_int;
		if(node) node=find_inode(node,*((long *)s));
	} else {
		if(flag) node=node_strings;
		else node=node_ped_str;
		if(node) node=find_snode(node,(char *)s);
	}
	return node;
}

static void free_node_tree(struct label_node *node)
{
	if(node->left) free_node_tree(node->left);
	if(node->right) free_node_tree(node->right);
	free(node);
}

void free_nodes(void)
{
	if(node_ints) free_node_tree(node_ints);
	if(node_strings) free_node_tree(node_strings);
	if(node_ped_int) free_node_tree(node_ped_int);
	if(node_ped_str) free_node_tree(node_ped_str);
	node_ints=node_strings=node_ped_int=node_ped_str=0;
}

static void set_missing(int col,int ncol,int rec,struct DataBlock *db)
{
	int i,j;
	
	i=rec*ncol+col;
	db->records[i].value=0;
	j=7-(i&7);
	i>>=3;
	db->type[i]|=(1<<j);
}

int check_missing(int col,int ncol,int rec,struct DataBlock *db)
{
	int i,j;
	i=rec*ncol+col;
	j=7-(i&7);
	i>>=3;
	return db->type[i]&(1<<j);
}

#ifdef FUNC_NAME
#undef FUNC_NAME
#endif
#define FUNC_NAME "handle_string"
static void handle_string(char *string,struct InFile *infile,struct var_element *elem,int col)
{
	int i,miss,ncol;
	char *sptr;
	struct label_node *node;	
	long value;
	
	ncol=infile->ncol;
	miss=0;
	if(miss_var) {
		for(i=0;i<miss_var[col].nmiss;i++) {
			if(!strcmp(string,miss_var[col].Missing[i])) {
				miss=1;
				break;  
			}
		}
	}
	if(!miss) {
		if(elem->type&(ST_ID|ST_SIRE|ST_DAM|ST_FAMILY|ST_FACTOR|ST_HAPLO))	{
			if(elem->type&ST_INTTYPE) {
				value=strtol(string,&sptr,10);
				if(*sptr) {
					print_scan_err("[%s:%d] %s(): Line %d column %d - Malformed integer %s\n",__FILE__,__LINE__,FUNC_NAME,lineno,col+1,string);
					miss=1;
				} else {
					if(elem->type&(ST_ID|ST_SIRE|ST_DAM|ST_RANDOM))	{
 						if(!node_ped_int) node=node_ped_int=alloc_node(&value,INTEGER);
						else node_ped_int=insert_node(node_ped_int,&value,&node,&i,INTEGER);
					} else {
 						if(!node_ints) node=node_ints=alloc_node(&value,INTEGER);
						else node_ints=insert_node(node_ints,&value,&node,&i,INTEGER);
					}
					DataBlock->records[DataBlock->record_ptr*ncol+col].node=node;
				}
			} else {
				if(elem->type&(ST_ID|ST_SIRE|ST_DAM|ST_RANDOM))	{
					if(!node_ped_str) node=node_ped_str=alloc_node(string,STRING);
					else node_ped_str=insert_node(node_ped_str,string,&node,&i,STRING);
				} else {
					if(!node_strings) node=node_strings=alloc_node(string,STRING);
					node_strings=insert_node(node_strings,string,&node,&i,STRING);
				}
				DataBlock->records[DataBlock->record_ptr*ncol+col].node=node;
			}
		} else {
			if(elem->type&ST_INTTYPE) DataBlock->records[DataBlock->record_ptr*ncol+col].value=strtol(string,&sptr,10);
			else DataBlock->records[DataBlock->record_ptr*ncol+col].rvalue=strtod(string,&sptr);
			if(*sptr) {
				print_scan_err("[%s:%d] %s(): Line %d column %d - Malformed %s %s\n",__FILE__,__LINE__,FUNC_NAME,lineno,col+1,(elem->type&ST_INTTYPE)?"integer":"floating point number",string);
				miss=1;
			}
		}
	}
	if(miss) {
		set_missing(col,ncol,DataBlock->record_ptr,DataBlock);
		if(elem->type&ST_ID) print_scan_err("[%s:%d] %s(): Line %d column %d - Missing id variable\n",__FILE__,__LINE__,FUNC_NAME,lineno,col+1);
		if(elem->type&ST_FAMILY) print_scan_err("[%s:%d] %s(): Line %d column %d - Missing family variable\n",__FILE__,__LINE__,FUNC_NAME,lineno,col+1);
	}
}

#ifdef FUNC_NAME
#undef FUNC_NAME
#endif
#define FUNC_NAME "update_datablock"
static void update_datablock(int ncol)
{
	int i;
	
	DataBlock->record_ptr++;
	if(DataBlock->record_ptr==DataBlock->blocksize) {
		if(!(DataBlock->next=malloc(sizeof(struct DataBlock)))) ABT_FUNC(MMsg);
		DataBlock->next->blocksize=DataBlock->blocksize;
		DataBlock=DataBlock->next;
		if(DataBlock->blocksize<MAX_BLOCK_SIZE) DataBlock->blocksize*=2;
		i=DataBlock->blocksize*ncol;
		if(!(DataBlock->records=malloc(sizeof(union DataRec)*i))) ABT_FUNC(MMsg);
		if(!(DataBlock->type=calloc((size_t)(i>>3),sizeof(char)))) ABT_FUNC(MMsg);
		DataBlock->next=0;
		DataBlock->record_ptr=0;
	}
}

#ifdef FUNC_NAME
#undef FUNC_NAME
#endif
#define FUNC_NAME "ReadData"
void ReadData(void)
{
	int i,j,ncol,num_records,col,realcol,eor;
	struct InFile *infile;
	char *buffer,ch,*string,*string_start,oldchar,*rs,*fs;
	int space[256],buffersize,fs_reg=0;
	FILE *fptr,*flog;
	size_t l,string_leng;
	struct recode_table_tag *recode_table=0;
	struct format *format;
	struct var_element *elem;
#ifdef HAVE_REGCOMP	
	regex_t preg;
	regmatch_t pmatch;
#endif
	
	infile=Infiles;
	if(!(StringData=malloc(S_BLOCK_SIZE))) ABT_FUNC(MMsg);
	RemBlock=AddRemem(StringData,RemBlock);
	StringPos=0;
	for(i=0;i<256;i++) space[i]=0;
	space[(int)' ']=1;
	space[(int)'\n']=1;
	space[(int)'\t']=1;
	if(LogFile) flog=fopen(LogFile,"a");
	else flog=0;
	if(flog) fprintf(flog,"\n******************** Reading in data ********************\n\n");
	while(infile) {
		fs_reg=0;
		if(infile->fformat) {
			fs=infile->fformat->fs;
			rs=infile->fformat->rs;
			if(fs) {
				if(strlen(fs)>1) {
#if HAVE_REGCOMP
					if((i=regcomp(&preg,fs,REG_EXTENDED))) fs=0;
					else fs_reg=1;
#endif				  
				} else if(fs[0]==' ') fs=0;
			}
			if(!rs) rs=default_rsformat;
		} else {
			fs=fsformat;
			if(rsformat) rs=rsformat;
			else rs=default_rsformat;
		}
		if(rs[0]=='\0') rs=0;
		if(!scan_error_n) {
			ncol=infile->ncol;
			if(Miss) process_missing(infile);
			else miss_var=0;
			if(infile->shell_flag) {
				if(!(fptr=popen(infile->name,"r"))) {
					fprintf(stderr,"Can't execute '%s'.\n",infile->name);
					perror("read_data()");
					scan_error_n++;
					break;
				}
				printf("Reading in output from shell command '%s'\n",infile->name);
			} else {
				if(!(fptr=fopen(infile->name,"r"))) {
					fprintf(stderr,"Can't open data file '%s' for reading.\n",infile->name);
					perror("read_data()");
					scan_error_n++;
					break;
				}
				printf("Reading in data from file '%s'\n",infile->name);
			}
			num_records=0;
			lineno=1;
			if(!(DataBlock=malloc(sizeof(struct DataBlock)))) ABT_FUNC(MMsg);
			i=INIT_BLOCK_SIZE*ncol;
			if(!(DataBlock->records=malloc(sizeof(union DataRec)*i))) ABT_FUNC(MMsg);
			if(!(DataBlock->type=calloc((size_t)i>>3,sizeof(unsigned int)))) ABT_FUNC(MMsg);
			DataBlock->next=0;
			DataBlock->blocksize=INIT_BLOCK_SIZE;
			DataBlock->record_ptr=0;
			infile->ncol=ncol;
			infile->data=DataBlock;
			format=infile->format;
			if(format) buffersize=2+format->f_atoms[format->n_atoms-1].pos+format->f_atoms[format->n_atoms-1].size;
			else buffersize=BUFFER_SIZE+1;
			if(!(buffer=malloc((size_t)buffersize))) ABT_FUNC(MMsg);
			errno=0;
			col=realcol=0;
			string=buffer;
			for(;;) {
				if(format) {
					if(!fgets(buffer,buffersize,fptr)) {
						if(errno) {
							perror("ReadData()");
							abt(__FILE__,__LINE__,"Aborting\n");
						}
						break;
					}
					l=strlen(buffer);
					for(i=0;i<(int)l;i++) if(!isspace((int)buffer[i])) break;
					if(i<(int)l) {
						for(col=realcol=0;realcol<infile->nvar;realcol++) {
							elem=infile->element[realcol];
							if(elem) {
								i=format->f_atoms[realcol].pos;
								if((size_t)i>=l) set_missing(col,ncol,DataBlock->record_ptr,DataBlock);
								else {
									j=format->f_atoms[realcol].size;
									string_start=buffer+i;
									oldchar=buffer[i+j];
									buffer[i+j]=0;
									if(string_start[0]) qstrip(string_start);
									if(string_start[0])	handle_string(string_start,infile,elem,col);
									else set_missing(col,ncol,DataBlock->record_ptr,DataBlock);
									buffer[i+j]=oldchar;
								}
								col++;
							}
						}
						update_datablock(ncol);
						num_records++;
					}
					lineno++;
					if(!(lineno%LINE_COUNT)) printf("At %d\n",lineno);
					/* read to end of line */
					while(buffer[l-1]!='\n') {
						if(!fgets(buffer,buffersize,fptr)) {
							if(errno) {
								perror("ReadData()");
								abt(__FILE__,__LINE__,"Aborting\n");
							}
							break;
						}
						l=strlen(buffer);
					}
					continue;
				}
				l=BUFFER_SIZE;
				if(l<=(size_t)(string-buffer)) ABT_FUNC("column width exceeds buffer size\n");
				l-=(size_t)(string-buffer);
				if(!(l=fread(string,1,l,fptr))) {
					if(errno) {
						perror("ReadData()");
						abt(__FILE__,__LINE__,"Aborting\n");
					}
					break;
				}
				*(string+l)=(char)0;
				string=buffer;
				for(;;) {
					eor=0;
					if(fs) {
						string_start=string;
#if HAVE_REGCOMP						
						if(fs_reg) {
							if(regexec(&preg,string_start,1,&pmatch,0)) {
								while((ch= *(string++))) {
									if(ch=='\n') lineno++;
									if(rs) if(ch==rs[0])	{
										eor=1;
										break;
									}
								}
							} else {
								string=string_start+pmatch.rm_eo;
								for(i=0;i<(int)pmatch.rm_eo;i++)	{
									ch=string_start[i];
									if(ch=='\n') lineno++;
									if(rs) if(ch==rs[0])	{
										eor=1;
										string=string_start+i+1;
										break;
									}
								}
								if(!eor) string_start[pmatch.rm_so]='\0';
								ch=1;
							}
						} else
#endif						  
						{
							while((ch= *(string++))) {
								if(ch=='\n') lineno++;
								if(ch==fs[0]) break;
								if(rs) if(ch==rs[0])	{
									eor=1;
									break;
								}	
							}
						}
					} else {
						while((ch= *(string++))) {
							if(ch=='\n') lineno++;
							if(rs) if(ch==rs[0]) {
								eor=1;
								break;
							}
							if(!space[(int)ch]) break;
						}
						if(!ch) {
							string=buffer;
							break;
						}
						string_start=string-1;
						if(!eor)	{
							while((ch= *(string++))) {
								if(rs) if(ch==rs[0]) {
									eor=1;
									break;
								}
								if(space[(int)ch]) break;
							}
							if(ch=='\n') lineno++;
						}
					}
					*(string-1)=(char)0;	
					if(string_start[0]) qstrip(string_start);
					string_leng=strlen(string_start);
					if(!ch) {
						if(buffer!=string_start) memmove(buffer,string_start,string_leng);
						string=buffer+string_leng;
						break;
					}
					elem=infile->element[realcol];
					if(elem && (string_start[0] || !eor)) {
						if(string_start[0]) handle_string(string_start,infile,elem,col);
						else set_missing(col,ncol,DataBlock->record_ptr,DataBlock);
						col++;
					}
					realcol++;
					if(realcol==infile->nvar || eor || (rs && col==ncol))	{
						if(col) {
							for(;col<ncol;col++) set_missing(col,ncol,DataBlock->record_ptr,DataBlock);
							num_records++;
							update_datablock(ncol);
						}
						realcol=col=0;
						if(rs) while(!eor) {
							while((ch= *(string++))) {
								if(ch=='\n') lineno++;
								if(ch==rs[0]) {
									eor=1;
									break;
								}
							}
							if(!ch) {
								string=buffer;
								if(!(l=fread(string,1,BUFFER_SIZE,fptr)))	{
									if(errno) {
										perror("ReadData()");
										abt(__FILE__,__LINE__,"Aborting\n");
									}
									eor=1;
									break;
								}
								*(string+l)=(char)0;
							}
						}
						if(!(lineno%LINE_COUNT)) printf("At %d\n",lineno);
					}
				}
			}
			if(miss_var) {
				for(i=0;i<ncol;i++) if(miss_var[i].nmiss) {
					free(miss_var[i].Missing);
					break;
				}
				free(miss_var);
			}
			fclose(fptr);
			if(infile->shell_flag) do i=wait(&j); while(i>0);
#if HAVE_REGCOMP			
			if(fs_reg) regfree(&preg);
#endif			
			infile->n_records=num_records;
			printf("%d records of %d columns read in\n",num_records,ncol);
			if(flog) fprintf(flog,"     Read %d records of %d columns from '%s'\n",num_records,ncol,infile->name);
			free(buffer);
		}
		infile=infile->next;
	}
	if(scan_error_n || !num_nodes) return;
	if(flog && flog!=stdout) fclose(flog);
	if(!(recode_table=calloc((size_t)num_nodes,sizeof(struct recode_table_tag)))) ABT_FUNC(MMsg);
	setup_pedigree(num_nodes,recode_table);
	if(scan_error_n) return;
	for(i=0;i<num_nodes;i++) recode_table[i].index=0;
	recode_factors(num_nodes,recode_table);
	if(scan_error_n) return;
	if(n_markers) recode_marker_data(num_nodes,recode_table);
	if(recode_table) free(recode_table);
	if(scan_error_n) return;
	match_records();
	if(Restrictions) restrict_data();
	if(Censored) censored_data();
	if(Affected) affected_data();
	check_sex();
	handle_groups();
	cleanup_unused();
	for(i=0;i<2;i++) if(sex_exp[i]) {
		if(sex_exp[i]->type==ST_STRING) if(sex_exp[i]->arg.string) free(sex_exp[i]->arg.string);
		free(sex_exp[i]);
	}
	prune_pedigree();
	count_components();
	if(OutputFile) Output_Data();
}
