/****************************************************************************
 *                                                                          *
 *     Loki - Programs for genetic analysis of complex traits using MCMC    *
 *                                                                          *
 *             Simon Heath - University of Washington                       *
 *                                                                          *
 *                       March 1997                                         *
 *                                                                          *
 * utils.c:                                                                 *
 *                                                                          *
 * Small utility routines used by both prep and loki                        *
 *                                                                          *
 ****************************************************************************/

#include <config.h>

#include <math.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#if HAVE_LIMITS_H
#include <limits.h>
#endif
#include <signal.h>
#include <string.h>
#include <sys/times.h>
#include <sys/stat.h>
#include <time.h>
#include <errno.h>
#if HAVE_UNISTD_H
#include <unistd.h>
#endif
#if HAVE_SYS_SYSTEMINFO_H
#include <sys/systeminfo.h>
#endif

#ifdef USE_DMALLOC
#include <dmalloc.h>
#endif

#include "ranlib.h"
#include "utils.h"

#ifndef MAXHOSTNAMELEN
#define MAXHOSTNAMELEN 64
#endif

const char *FMsg="File Error - aborting\n";
const char *IntErr="Fatal Internal Error - aborting\n";
const char *MMsg="Out of Memory Error - aborting\n";

void abt(const char *file, const int line, const char *fmt, ...)
{
	va_list args;

	if(stdout) (void)fflush(stdout);
	(void)fprintf(stderr,"[%s:%d] ",file,line);
	va_start(args,fmt);
	(void)vfprintf(stderr,fmt,args);
	va_end(args);
	_exit(EXIT_FAILURE); 	/* Avoid cleanup routines if aborting */
}

/* performs a p<p1 operation on the strings pointed to be 
p & p1.  Returns 1 if p<p1 and 0 otherwise.  Used for sorting */
int strless(const char *p, const char *p1)
{
	size_t i,l,l1;
	
	if(!p1) return 0;
	if(!p) return 1;
	l=strlen(p);
	l1=strlen(p1);
	if(l<l1) return 1;
	else if(l>l1) return 0;
	for(i=0;i<l;i++) {
		if(p[i]<p1[i]) return 1;
		if(p[i]>p1[i]) return 0;
	}
	return 0;
}

int mystrcmp(const char *p1, const char *p2)
{
	if(!p1) {
		if(!p2) return 0;
		return 1;
	}
	if(!p2) return 1;
	return strcmp(p1,p2);
}

size_t mystrlen(const char *p1)
{
	if(!p1) {
		(void)fprintf(stderr,"mystrlen(): Warning - zero pointer passed\n");
		return 0;
	} else return strlen(p1);
}

/* Case insensitive string comparison */
int ci_strcmp(const char *s1,const char *s2)
{
	size_t l,i;
	
	l=mystrlen(s1);
	if(!(l==mystrlen(s2))) return 1;
	if(!l) return 0;
	for(i=0;i<l;i++) if(toupper((int)s1[i])!=toupper((int)s2[i])) return 1;
	return 0;
}

void qstrip(char *s1)
{
	char *p,*p1;
	
	p=s1;
	p1=s1-1;
	while(*s1) {
		if(!isspace((int)*s1)) break;
		s1++;
	}
	while(*s1) {
		if(!isspace((int)*s1)) p1=p;
		*(p++)= *(s1++);
	}
	*(++p1)='\0';
}

int nstrcmp(const char *s1,char *s2)
{
	qstrip(s2);
	return strcmp(s1,s2);
}

#ifdef FUNC_NAME
#undef FUNC_NAME
#endif
#define FUNC_NAME "copy_string"
char *copy_string(const char *s)
{
	char *s1;
	
	if(!(s1=malloc(strlen(s)+1))) ABT_FUNC(MMsg);
	return strcpy(s1,s);
}

#ifdef FUNC_NAME
#undef FUNC_NAME
#endif
#define FUNC_NAME "make_file_name" 
char *make_file_name(const char *s)
{
	size_t l;
	char *s1,*s2,*s3;
	
	if(!s) ABT_FUNC("Passed zero pointer\n");
	l=strlen(s)+strlen(file_prefix)+2;
	if(file_dir) l+=strlen(file_dir)+1;
	if(!l) ABT_FUNC(IntErr);
	if(!(s1=malloc(l))) ABT_FUNC(MMsg);
	s2=s1-1;
	s3=file_dir;
	if(s3) {
		while((*++s2 = *s3++));
		*s2 = '/';
	}
	s3=file_prefix;
	while((*++s2 = *s3++));
	*s2 = '.';
	while((*++s2 = *s++));
	return s1;
}

void print_start_time(const char *progname,const char *mode)
{
	FILE *flog;
	char *buf;
	struct stat sbuf;
	int i,j;
	
	if(LogFile && mode[0]=='w') {
		if(!stat(LogFile,&sbuf)) {
			i=1;
			j=(int)strlen(LogFile);
			buf=(char *)malloc((size_t)j+2);
			if(buf) {
				strcpy(buf,LogFile);
				buf[j]='~';
				buf[j+1]='\0';
				i=rename(LogFile,buf);
				free(buf);
			}
			if(i) (void)fprintf(stderr,"print_start_time(): Couldn't rename old logfile\n");
		}
	}
	if(LogFile) flog=fopen(LogFile,mode);
	else flog=stdout;
	if(flog) {
		(void)fprintf(flog,"\n********************** Starting *************************\n\n");
		(void)fprintf(flog,"     %s: %s",progname,ctime(&start_time));
		if(LogFile) (void)fclose(flog);
	} else (void)fprintf(stderr,"Couldn't write to log file\n");
}

void print_end_time(void)
{
	FILE *flog;
	time_t end_time;
	double l;
	struct tms tbuf;
	long tps;
	int t=0,flag=0;
	char hname[MAXHOSTNAMELEN+1];
	
	if(LogFile) flog=fopen(LogFile,"a");
	else flog=stdout;
	if(flog)	{
#if HAVE_SYS_SYSTEMINFO_H
		if(sysinfo(SI_HOSTNAME,hname,MAXHOSTNAMELEN)<0)
#else
		if(gethostname(hname,MAXHOSTNAMELEN)<0)
#endif
			 strcpy(hname,"UNKNOWN");
		
		(void)fprintf(flog,"\n*********************** Exiting *************************\n\n");
		(void)fprintf(flog,"     Hostname:     %s\n", hname);
		tps=sysconf (_SC_CLK_TCK);
		errno=0;
		(void)times(&tbuf);
		if(errno) perror("print_end_time():");
		else {
			(void)fprintf (flog,"     System time:  %.4f seconds\n",extra_stime+(double)tbuf.tms_stime/(double)tps);
			(void)fprintf (flog,"     User time:    %.4f seconds\n",extra_utime+(double)tbuf.tms_utime/(double)tps);
		}
		end_time=time(0);
		l=extra_time+difftime(end_time,start_time);
		(void)fputs("     Elapsed time: ",flog);
		if(l>86400.0) {
			t=(int)l/86400.0;
			l-=(double)t*86400.0;
			(void)fprintf(flog,"%d day%s",t,t!=1?"s, ":", ");
			flag=1;
		}
		if(l>3600.0) {
			t=(int)(l/3600.0);
			l-=(double)t*3600.0;
			flag=1;
		}
		if(flag) (void)fprintf(flog,"%d hour%s",t,t!=1?"s, ":", ");
		if(l>60.0) {
			t=(int)(l/60.0);
			l-=(double)t*60.0;
			flag=1;
		}
		if(flag) (void)fprintf(flog,"%d minute%s",t,t!=1?"s, ":", ");
		(void)fprintf(flog,"%d second%c\n",(int)l,(int)l!=1?'s':' ');
		if(LogFile) (void)fclose(flog);
	}
}

void gen_perm(int *x,int n)
{
	int i,j;
	
	while(n) {
		j=(int)(safe_ranf()*(double)(n--));
		i=x[j];
		x[j]=x[n];
		x[n]=i;
	}
}

int txt_print_double(double x,FILE *fptr)
{
	int i,er=0;
	double y,z;
	static char *hexdigits="0123456789abcdef";

	y=frexp(x,&i);
	if(fprintf(fptr,"%d:",i)<0) er=1;
	if(!er && y<0.0) {
		y=-y;
		if(fputc('-',fptr)==EOF) er=1;
	}
	if(!er) {
		if(y) {
			while(y>0.0 && !er) {
				y=frexp(y,&i);
				y=ldexp(y,i+4);
				y=modf(y,&z);
				if(fputc(hexdigits[(int)z],fptr)==EOF) er=1;
			}
		} else if(fputc('0',fptr)==EOF) er=1;
	}
	return er;
}

int txt_get_double(char *p,char **p1,double *x)
{
	int j,e,mf=0;
	char *p2;
	double y;
	
	e=strtol(p,p1,10);
	p2=*p1;
	if(*p2++!=':') return 1;
	if(*p2=='-') {
		mf=1;
		p2++;
	}
	p=p2;
	while(isxdigit((int)*p2++));
	y=0.0;
	*p1=--p2;
	while(p2-->p) {
		j=(int)*p2;
		if(j>='0'&&j<='9') j-='0';
		else if(j>='a'&&j<='f') j-='a'-10;
		else return 1;
		y+=(double)j;
		y=frexp(y,&j);
		y=ldexp(y,j-4);
	}
	if(mf) y=-y;
	*x=ldexp(y,e);
	return 0;
}

