%{
/****************************************************************************
 *                                                                          *
 *     Loki - Programs for genetic analysis of complex traits using MCMC    *
 *                                                                          *
 *             Simon Heath - University of Washington                       *
 *                                                                          *
 *                       March 1997                                         *
 *                                                                          *
 * control_lex.l:                                                           *
 *                                                                          *
 * flex (NOT lex!) source for control file lexer.                           *
 *                                                                          *
 ****************************************************************************/

#include <stdio.h>
	
#include <string.h>
#include <ctype.h>
#include <stdlib.h>

#include "control_parse.h"
#include "scan.h"
#include "scanner.h"
#include "utils.h"

#define YY_DECL int yy_orig_lex YY_PROTO(( void ))
#define MAX_STRING 512

int lineno=1,lineno1=1,at_file,at_model,at_use,begin_comm;
char linebuf[512],linebuf1[512];
static char *get_string(char *s,int size);

static int tokenpos,stringbuf_ptr,ff_flag;
static char stringbuf[MAX_STRING];
static void process_env_var(char *ptr);
static void append_to_sbuf(int *sptr,char c);

%}

%x COMMENT
%x STR 
%s ATFORMAT
%s ISTATE

%%

%{
	int token;
%}

\n?.*(\$[a-zA-Z][a-zA-Z0-9_]*)+.*$ { process_env_var(yytext); }

<INITIAL>.*\n {
    strncpy(linebuf,yytext,512);
    strcpy(linebuf1,linebuf);
    yyless((int)0);
    BEGIN ISTATE;
    }

[\t ] {tokenpos+=yyleng;}

\n.* { 
    lineno++;
    tokenpos=0;
    strncpy(linebuf,yytext+1,512);
    yyless((int)1);
    }

"/*" BEGIN(COMMENT);
<COMMENT>. {tokenpos+=yyleng;}
<COMMENT>\n {lineno++; tokenpos=0;}
<COMMENT>"*/" {BEGIN(ISTATE); tokenpos+=yyleng;}

#.*$ ;

[\"\']   { stringbuf_ptr=0;
       BEGIN(STR);
 	  tokenpos+=yyleng; }

<STR>[\"\'] {
           BEGIN(ISTATE);
	      append_to_sbuf(&stringbuf_ptr,'\0');
		 yylval.string=get_string(stringbuf,stringbuf_ptr);
		 tokenpos+=yyleng;
		 return STRING; 
	   }

<STR>\n { BEGIN(ISTATE);
	     append_to_sbuf(&stringbuf_ptr,'\0');
		print_scan_warn("Line %d: Warning - Unterminanted string\n",lineno);
		tokenpos+=yyleng;
		yylval.string=get_string(stringbuf,stringbuf_ptr);
		return STRING; }

<STR>\\[0-7]{1,3} { unsigned int result;
                    (void) sscanf(yytext+1,"%uo",&result);
				tokenpos+=yyleng;
				if(result>0xff) print_scan_err("Line %d: Error - Octal constant too large\n",lineno);
				append_to_sbuf(&stringbuf_ptr,(char)result); }

<STR>\\x[a-fA-F0-9]{1,2} { unsigned int result;
                    (void) sscanf(yytext+1,"%ux",&result);
				tokenpos+=yyleng;
 				append_to_sbuf(&stringbuf_ptr,(char)result); }

<STR>\\[0-9]+ { tokenpos+=yyleng;
                print_scan_err("Line %d: Error - Syntax error in escape sequence\n",lineno); }

<STR>\\a { tokenpos+=yyleng;
	     append_to_sbuf(&stringbuf_ptr,'\a'); }

<STR>\\b { tokenpos+=yyleng;
 	     append_to_sbuf(&stringbuf_ptr,'\b'); }

<STR>\\f { tokenpos+=yyleng;
	     append_to_sbuf(&stringbuf_ptr,'\f'); }

<STR>\\n { tokenpos+=yyleng;
	     append_to_sbuf(&stringbuf_ptr,'\n'); }

<STR>\\r { tokenpos+=yyleng;
	     append_to_sbuf(&stringbuf_ptr,'\r'); }

<STR>\\t { tokenpos+=yyleng;
	     append_to_sbuf(&stringbuf_ptr,'\t'); }

<STR>\\v { tokenpos+=yyleng;
	     append_to_sbuf(&stringbuf_ptr,'\v'); }

<STR>\\(.|\n) { tokenpos++;
	     append_to_sbuf(&stringbuf_ptr,'\a'); }

<STR>[^\\\n\"\']+ { char *ss=yytext;
                  tokenpos+=yyleng;
			   while(*ss) append_to_sbuf(&stringbuf_ptr,*ss++); }

\[ { if(at_file) BEGIN ATFORMAT;
     tokenpos+=yyleng;
     return '['; }

<ATFORMAT>\] { if(at_file) BEGIN ISTATE;
              tokenpos+=yyleng;
              return ']'; }

<ATFORMAT>[xX] { tokenpos+=yyleng; return('x'); }

<ATFORMAT>[yY] { tokenpos+=yyleng; return('y'); }

"<=" { return LEQSYMBOL; }

">=" { return GEQSYMBOL; }

"!="|"<>" { return NEQSYMBOL; }

"==" { return '='; }

\||[Oo][Rr]|\|\| { return ORSYMBOL; }

&|[Aa][Nn][Dd]|&& { return ANDSYMBOL; }

!|[Nn][Oo][Tt] { return NOTSYMBOL; }

<ATFORMAT>[Rr][Ss][ \t]*= { return RSFORMAT; }
<ISTATE>[Rr][Ss][ \t]*= { if(!ff_flag) {yyless(0); token=BREAK;}
                          else token=RSFORMAT;
	                       ff_flag^=1;
	                       return token; }

<ATFORMAT>[Ff][Ss][ \t]*= { return FSFORMAT; }
<ISTATE>[Ff][Ss][ \t]*= { if(!ff_flag) {yyless(0); token=BREAK;}
                          else token=FSFORMAT;
	                       ff_flag^=1;
	                       return token; }

[a-zA-Z][a-zA-Z0-9_]*/([\t ]*=)|([\t ]*\([^\n\"']*\)[\t ]*=) {
                               token=symbol_lookup(yytext,1);
				           if(token==BREAK) yyless(0);
				           else
						 {
						   tokenpos+=yyleng;
						   if(!at_model && !at_use && (token==VARIABLE || token==ARRAY_VAR)) token=ASSIGN;
						 }
						 if((!at_model && !at_use) || token!=BREAK) return token;
					    }

[a-zA-Z][a-zA-Z0-9_]* { token=symbol_lookup(yytext,0);
				    if(token==BREAK) yyless(0);
				    else tokenpos+=yyleng; 
				    return token; }

([0-9]*\.[0-9]+)|(([0-9]+|([0-9]*\.[0-9]+))E[-+]?[0-9]+) {
         tokenpos+=yyleng; yylval.rvalue=atof(yytext); return REAL; }

[0-9]+ { tokenpos+=yyleng; yylval.value=atoi(yytext); return INTEGER; }

. { tokenpos+=yyleng; return yytext[0]; }

%%

void yyerror(char *s)
{
     int i;
	
	if(loop_record== -1)
	{
		print_scan_err("Line %d: %s\n",lineno,s);
		loop_record= -2;
	}
	else if(loop_record>=0)
	{
		print_scan_err("Line %d: %s\n%s\n",lineno,s,linebuf);
		if(scan_error_n<=max_scan_errors)
		{
			for(i=1;i<tokenpos;i++) putc('-',stderr);
			fputs("^\n",stderr);
		}
	}
	scan_error|=SCAN_ERR;
}

void yyerror1(char *s)
{
	if(loop_record== -1)
	{
		print_scan_err("Line %d: %s\n",lineno1,s);
		loop_record= -2;
	}
	else if(loop_record>=0) print_scan_err("Line %d: %s\n%s\n",lineno1,s,linebuf1);
	scan_error|=SCAN_ERR;
}

static void append_to_sbuf(int *sptr,char c)
{
	if(*sptr>=MAX_STRING) print_scan_err("Line %d: Maximum string size exceeded",lineno);
	stringbuf[(*sptr)++]=c;
}

#ifdef FUNC_NAME
#undef FUNC_NAME
#endif
#define FUNC_NAME "get_string"
static char *get_string(char *s,int size)
{
	char *s1;
	if(!(s1=malloc(size))) ABT_FUNC(MMsg);
	memcpy(s1,s,size);
	return s1;
}

#ifdef FUNC_NAME
#undef FUNC_NAME
#endif
#define FUNC_NAME "yylex"
int yylex(void)
{
	int tok,i;
	
	if(loop_record==1)
	{
		tok=yy_orig_lex();
		if(!loop_stack)
		{
			if(!(loop_stack=malloc(sizeof(struct token_store)*loop_stack_size))) ABT_FUNC(MMsg);
			loop_main_ptr=0;
		}
		else if(loop_main_ptr==loop_stack_size)
		{
			loop_stack_size*=2;
			if(!(loop_stack=realloc(loop_stack,sizeof(struct token_store)*loop_stack_size))) ABT_FUNC(MMsg);
		}
		loop_stack[loop_main_ptr].line=lineno;
		loop_stack[loop_main_ptr].line1=lineno1;
		loop_stack[loop_main_ptr].yylval=yylval;
		if(tok==STRING) yylval.string=copy_string(yylval.string);
		loop_stack[loop_main_ptr++].token=tok;
		return tok;	   
	}
	else if(loop_record<0)
	{
		if(in_loopclause== -1)
		{
			in_loopclause= -2;
			return LOOP_CLAUSE_START;
		}
		yylval=loop_stack[loop_main_ptr].yylval;
		lineno=loop_stack[loop_main_ptr].line;
		lineno1=loop_stack[loop_main_ptr].line1;
		tok=loop_stack[loop_main_ptr++].token;
		if(tok==STRING) yylval.string=copy_string(yylval.string);
		if(in_loopclause)
		{
			if(tok==',' && loop_stack[loop_main_ptr].token==BREAK)
			{
				loop_clause_element->arg.value+=loop_clause_step;
				if(loop_clause_element->arg.value==loop_clause_end) i=0;
				else
				{
					i=(loop_clause_element->arg.value>loop_clause_end);
					if(loop_clause_step<0) i=1-i;
				}
				if(i)
				{
					loop_record=loop_stat[--loop_level];
					if(!loop_record) loop_main_ptr=loop_level?loop_ptr[loop_level-1]:0;
					else loop_main_ptr=loop_clause_ptr;
					in_loopclause=0;
					return LOOP_CLAUSE_END;
				}
				else loop_main_ptr=loop_ptr[loop_level-1];
			}
		}
		return tok;
	}
	return yy_orig_lex();
}

void yy_cleanup(void)
{
	yy_delete_buffer(yy_current_buffer);
	if(loop_stack) free(loop_stack);
}

#ifdef FUNC_NAME
#undef FUNC_NAME
#endif
#define FUNC_NAME "process_env_var"
static void process_env_var(char *ptr)
{
	int i=0,j,k,size=512,brack;
	char *s,old,*stringptr;
	
	if(!(stringptr=malloc(size))) ABT_FUNC(MMsg);
	for(;;)
	{
		while(*ptr && *ptr!='$')
		{
			if(i>=size)
			{
				size+=size;
				if(!(stringptr=realloc(stringptr,size))) ABT_FUNC(MMsg);
			}
			stringptr[i++] = *(ptr++);
		}
		if(!(*ptr)) break;
		ptr++;
		j=0;
		brack=0;
		if(*ptr=='(')
		{
			ptr++;
			while(ptr[j] && ptr[j]!=')') j++;
			if(ptr[j]==')') brack=1;
		}
		else
		{
			while(isalnum((int)ptr[j]) || ptr[j]=='_') j++;
		}
		old=ptr[j];
		ptr[j]=0;
		if((s=getenv(ptr)))
		{
			if((k=strlen(s)))
			{
				if(k+i>=size)
				{
					while(k+i>size) size+=size;
					if(!(stringptr=realloc(stringptr,size))) ABT_FUNC(MMsg);
				}
				memcpy(stringptr+i,s,k);
				i+=k;
			}
		}
		ptr[j]=old;
		ptr+=j+brack;
	}
	while(i) unput(stringptr[--i]);
	free(stringptr);
}
